//--------------------------------------------------------------------------//
// xdata3.cc                                                                //
// Evaluation of string expressions                                         //
// Latest revision: 08-01-2023                                              //
//--------------------------------------------------------------------------//

#include "xdata.h"
extern Globals     g;
extern Spreadsheet s;
extern int need_refresh;
char *text=NULL;
double answer;
int child_status;
FILE *fp, *fp2;

//-------------------------------------------------------------------------//
// formula_error                                                           //
//-------------------------------------------------------------------------//
void formula_error(char *error_message)
{
    if(strcmp(error_message, "parse error")) message(error_message);
    g.getout=1; 
}


//-------------------------------------------------------------------------//
// set_to_string                                                           //
//-------------------------------------------------------------------------//
void set_to_string(int rr, int cc)
{

}


//-------------------------------------------------------------------------//
// set_to_formula                                                          //
//-------------------------------------------------------------------------//
void set_to_formula(void)
{
}

 
//-------------------------------------------------------------------------//
// eval - entry point into evaluating a string                             //
// 1st one needs col and row in case string is a formula or label          //
// use the following to prevent repeated error messages if in a callback   //
//   g.message_count = 0;  // hack to see if eval has an error             //
//   s.x[k][row] = eval(s.text[k][row], k, row);                           //
//   if(g.message_count) { fclose(fp); g.message_count=0; return; }        //
// hex numbers must be mangled first because repairing yacc output is a    //
//   nightmare. it is far too much trouble to edit the spaghetti code in   //
//   lex.yy.c                                                              //
//-------------------------------------------------------------------------//
double eval(char* string, int col, int row)
{   
   double value;
   if(s.string[col][row]) return 0;
   if(s.hex[col][row]) return 0;  //// don't evaluate hex numbers
   g.cc = col; 
   g.rr = row;
   s.number[col][row] = 1;
   s.string[col][row] = 0;
   s.hex[col][row] = 0;
   value = eval(string);
   return value;
}
double eval(char* string)
{ 
   int k;
   // get rid of any weird characters in string
   for(k=0;k<(int)strlen(string);k++) if(string[k]==13) string[k]=0;
   if(strlen(string)==0 || strlen(string)>1023) return 0;
   if(!strcmp(string,"=")) return 0;
   if(!strcmp(string,"-")) return 0; //skip
   text = (char*)malloc(strlen(string)+8);
   sprintf(text,"%s;\n\255",string);
   yyparse();
   if((int)strlen(text))free(text);
   return answer;
}    


//-------------------------------------------------------------------------//
// graph                                                                   //
//-------------------------------------------------------------------------//
void graph(int updating, char **arg, int noofargs)
{
#define SCREEN 1
#define POSTSCRIPT 2  
  arg=arg; noofargs=noofargs;
  static int in_graph=0;
  static Dialog dialog;
  static double xmin=0.0, ymin=0.0;
  static double xmax=1.0, ymax=1.0;
  static int xtic=2, ytic=2;
  static int mxtic=10, mytic=10;
  static double pointsize=1.5;
  static int mx=s.current_col, my=1, graphtype=1;
  static int output=POSTSCRIPT;   //1=screen 2=postscript file
  static int xlog=0, ylog=0;
  static char title[128]="Graph";
  static char xlabel[128]="X";
  static char ylabel[128]="Y";
  static char filename[128]="graph.ps";
  static char hexcolor[128]= "0x4477aa";
  static int hit=0;
  int j, k;

  if(in_graph)return;
  in_graph = 1;
 
  ////------------------------////
  //// Start of menu section  ////
  ////------------------------////

   if(hit)
   {    graphtype = dialog.radioset[0];
        output = dialog.radioset[1];

        //// 0
        strcpy(title, dialog.answer[1]);
        strcpy(xlabel, dialog.answer[2]);
        strcpy(ylabel, dialog.answer[3]);
        strcpy(filename, dialog.answer[4]);
        strcpy(g.gvfilename, dialog.answer[5]);
        //// 6
        mx = atoi(dialog.answer[7]);
        my = atoi(dialog.answer[8]);
        xlog = dialog.boxset[9];
        ylog = dialog.boxset[10];
        xmin = atof(dialog.answer[11]);
        ymin = atof(dialog.answer[12]);
        xmax = atof(dialog.answer[13]);
        ymax = atof(dialog.answer[14]);
        xtic = atof(dialog.answer[15]);
        ytic = atof(dialog.answer[16]);
        pointsize = atof(dialog.answer[17]);
   }
   hit=1;

  if(!updating || !g.graph_active)
  { 
     dialog.needrepeat = 0;  
     strcpy(dialog.title,"Graph");
     strcpy(dialog.radio[0][0],"Graph type");
     strcpy(dialog.radio[0][1],"Points + lines");
     strcpy(dialog.radio[0][2],"Points only");
     strcpy(dialog.radio[0][3],"Lines only");
     strcpy(dialog.radio[0][4],"Bar graph");
     strcpy(dialog.radio[1][0],"Output");
     strcpy(dialog.radio[1][1],"Screen (manual size)");
     strcpy(dialog.radio[1][2],"Postscript file");
     strcpy(dialog.radio[2][0],"Auto scale");
     strcpy(dialog.radio[2][1],"On");
     strcpy(dialog.radio[2][2],"Off");
     strcpy(dialog.radio[3][0],"Auto update");
     strcpy(dialog.radio[3][1],"On");
     strcpy(dialog.radio[3][2],"Off");

     dialog.radiono[0] = 5;
     dialog.radiono[1] = 3;
     dialog.radiono[2] = 3;
     dialog.radiono[3] = 3;
     dialog.radioset[0] = graphtype;
     dialog.radioset[1] = output;
     if(g.autoscale) dialog.radioset[2] = 1; else dialog.radioset[2] = 2;
     if(g.autoupdate) dialog.radioset[3] = 1; else dialog.radioset[3] = 2;
     for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;

     strcpy(dialog.boxes[0],"Graph parameters");
     strcpy(dialog.boxes[1],"Title");
     strcpy(dialog.boxes[2],"X label");
     strcpy(dialog.boxes[3],"Y label");
     strcpy(dialog.boxes[4],"PS filename");
     strcpy(dialog.boxes[5],"Location of gv");
     strcpy(dialog.boxes[6],"Data");
     strcpy(dialog.boxes[7],"Column for X");
     strcpy(dialog.boxes[8],"Column for Y");
     strcpy(dialog.boxes[9],"Log X axis");
     strcpy(dialog.boxes[10],"Log Y axis");
     strcpy(dialog.boxes[11],"X min");
     strcpy(dialog.boxes[12],"Y min");
     strcpy(dialog.boxes[13],"X max");
     strcpy(dialog.boxes[14],"Y max");
     strcpy(dialog.boxes[15],"X tic label digits");
     strcpy(dialog.boxes[16],"Y tic label digits");
     strcpy(dialog.boxes[17],"Data point size");
     strcpy(dialog.boxes[18],"No. of x tic intervals");
     strcpy(dialog.boxes[19],"No. of y tic intervals");
     strcpy(dialog.boxes[20],"Color (in hex)");
     strcpy(dialog.boxes[21],"UPDATE GRAPH");

     dialog.boxtype[0]=LABEL;

     dialog.boxtype[1]=STRING;
     sprintf(dialog.answer[1],"%s",title);
     dialog.boxtype[2]=STRING;
     sprintf(dialog.answer[2],"%s",xlabel);
     dialog.boxtype[3]=STRING;
     sprintf(dialog.answer[3],"%s",ylabel);
     dialog.boxtype[4]=STRING;
     sprintf(dialog.answer[4],"%s",filename);
     dialog.boxtype[5]=STRING;
     sprintf(dialog.answer[5],"%s",g.gvfilename);

     dialog.boxtype[6]=LABEL;

     dialog.boxtype[7]=STRING;
     sprintf(dialog.answer[7],"%d",mx);
     dialog.boxtype[8]=STRING;
     sprintf(dialog.answer[8],"%d",my);
     dialog.boxtype[9]=TOGGLE;
     dialog.boxset[9]=xlog;
     dialog.boxtype[10]=TOGGLE;
     dialog.boxset[10]=ylog;

     dialog.boxtype[11]=STRING;
     sprintf(dialog.answer[11],"%g",xmin);
     dialog.boxtype[12]=STRING;
     sprintf(dialog.answer[12],"%g",ymin);
     dialog.boxtype[13]=STRING;
     sprintf(dialog.answer[13],"%g",xmax);
     dialog.boxtype[14]=STRING;
     sprintf(dialog.answer[14],"%g",ymax);
     dialog.boxtype[15]=STRING;
     sprintf(dialog.answer[15],"%d",xtic);
     dialog.boxtype[16]=STRING;
     sprintf(dialog.answer[16],"%d",ytic);
     dialog.boxtype[17]=STRING;
     sprintf(dialog.answer[17],"%g",pointsize);
     dialog.boxtype[18]=STRING;
     sprintf(dialog.answer[18],"%d",mxtic);
     dialog.boxtype[19]=STRING;
     sprintf(dialog.answer[19],"%d",mytic);
     dialog.boxtype[20]=STRING;
     sprintf(dialog.answer[20],"%s",hexcolor);

     dialog.boxtype[21]=PUSHBUTTON;
     sprintf(dialog.answer[21],"%d",k);

     for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
     for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
     for(j=0;j<20;j++) dialog.boxcomeback[j]=0;

     dialog.noofradios=4;
     dialog.noofboxes=22;
     dialog.helptopic=0;
     dialog.keep=0;  //// 1 = keep dialog visible until cancel
     dialog.needrepeat=0;
     dialog.want_changecicb = 0;
     dialog.f1 = graph_update;
     dialog.transient = 1;
     strcpy(dialog.path, ".");

     ///////---///////
     dialogbox(dialog);
     ///////---///////

     in_graph = 0;
     if(g.getout){ g.getout=0; return; }
     graph_draw(dialog);
  }
}


//-------------------------------------------------------------------------//
// graph_draw                                                              //
//-------------------------------------------------------------------------//
void graph_draw(dialoginfo a)
{
  double xmin=0.0, ymin=0.0;
  double xmax=1.0, ymax=1.0;
  double xlowest=0.0, ylowest=0.0;
  double xhighest=0.0, yhighest=1.0;
  int xtic=2, ytic=2;
  int mxtic=2, mytic=2;
  double pointsize=1.5;
  int mx=s.current_col, my=1, graphtype=1;
  int output=POSTSCRIPT;   //1=screen 2=postscript file
  int xlog=0, ylog=0;
  char title[128]="Graph";
  char xlabel[128]="X";
  char ylabel[128]="Y";
  char filename[128]="graph.ps";
  static int hit = 0;
  char tempstring[4*FILENAMELENGTH];
  char program[2*FILENAMELENGTH];
  char hexcolor[128]= "0x000077";
  int row;
  char with_type[32];
  FILE *fp3;

  graphtype = a.radioset[0];
  output = a.radioset[1];
  if(a.radioset[2] == 1) g.autoscale=1 ; else  g.autoscale=2;
  if(a.radioset[3] == 1) g.autoupdate=1 ; else  g.autoupdate=2;
  //// 0
  strcpy(title, a.answer[1]);
  strcpy(xlabel, a.answer[2]);
  strcpy(ylabel, a.answer[3]);
  strcpy(filename, a.answer[4]);
  strcpy(g.gvfilename, a.answer[5]);
  //// 6
  mx = atoi(a.answer[7]);
  my = atoi(a.answer[8]);
  xlog = a.boxset[9];
  ylog = a.boxset[10];
  xmin = atof(a.answer[11]);
  ymin = atof(a.answer[12]);
  xmax = atof(a.answer[13]);
  ymax = atof(a.answer[14]);
  xtic = atoi(a.answer[15]);
  ytic = atoi(a.answer[16]);
  pointsize = atof(a.answer[17]);
  mxtic = atoi(a.answer[18]);
  mytic = atoi(a.answer[19]);
  strcpy(hexcolor, a.answer[20]);


  //// If just starting or if they clicked 'graph off', then
  //// assume no gv is running. If gv is running, too bad, the
  //// data go to both of them.

  if(!g.graph_active || (output==POSTSCRIPT && hit==0))
  {    
       if(g.autoupdate==1)
       {   sprintf(program, "%s -watch %s &", g.gvfilename, filename);
           if(access(g.gvfilename, X_OK)) 
           {    sprintf(tempstring, "Cannot execute %s", program);
                message(tempstring);
                g.getout = 1;
                return;
           }
           printf("Executing %s \n",program);
           system(program);
       }else
       {   sprintf(program, "%s -watch %s &", g.gvfilename, filename);
           sprintf(tempstring, (char*)"To view graph, type the command\ngv -watch %s\nThen click Reload if you change parameters",filename);
           message(tempstring, INFO);
       }
       hit++;
       g.graph_active = 1;
  } 

  if((fp3=fopen("~temp.dat","wt")) == NULL)
      error_message(filename, errno);
  else
      for(row=1; row<=max(s.maxrow[mx],s.maxrow[my]); row++)
          fprintf(fp3, "%g   %g\n",s.x[mx][row],s.x[my][row]);
  fclose(fp3);

  ////  popen gnuplot, tell it to read from fifo
  ////  we are now in child process

  if((fp=(FILE*)popen("gnuplot","w")) == NULL)
      error_message((char*)"gnuplot", errno);
  else
  {
       fprintf(fp,"# Gnuplot graph file produced by xdata\n");
       if(output==POSTSCRIPT)  //postscript
       {
            fprintf(fp, "set output %c%s%c\n", 34,filename,34);
            fprintf(fp, "set terminal postscript portrait color \"Helvetica\" 14\n");
            fprintf(fp, "set size 1.0, 0.75\n");
       }else
       {
            fprintf(fp, "set term x11 persist\n");
       }
       switch(graphtype)
       {   case 1: strcpy(with_type, "linespoints"); break;
           case 2: strcpy(with_type, "points"); break;
           case 3: strcpy(with_type, "lines"); break;
           case 4: strcpy(with_type, "boxes"); break;
       }
       //// pt 6 = open circles
       //// pt 7 = closed circles
       //// lw = linewidth
       //// lt = 2 = dotted 1=solid
       fprintf(fp,"set style line 1 lc rgb \'#0060ad\' lt 1 lw 2 pt 6\n");
       fprintf(fp,"set pointsize 2\n");
       fprintf(fp,"set clip points\n");  // no effect
       fprintf(fp,"set title %c%s%c\n",'"',title,'"');
       fprintf(fp,"set xlabel %c%s%c\n",'"',xlabel,'"');
       fprintf(fp,"set ylabel %c%s%c\n",'"',ylabel,'"');

       if(g.autoscale==2)    //// 1=on 2=0ff
       {    if(graphtype==4) //// bar graph
            {
                 fprintf(fp,"set xrange [%g:%g]\n",xmin, xmax);
                 fprintf(fp,"set yrange [%g:%g]\n",ymin, ymax);
                 fprintf(fp,"set x2range [ * : *]\n");
                 fprintf(fp,"set y2range [ * : *]\n");
                 fprintf(fp,"set zrange [ * : *]\n");
                 fprintf(fp,"set cbrange [ * : *]\n");
                 fprintf(fp,"set rrange [ * : *]\n");
                 fprintf(fp,"set style fill solid border lt -1\n");
                 fprintf(fp,"set xtics scale 0 ()\n");
                 fprintf(fp,"set ytics \n");
                 fprintf(fp,"set mxtics %d\n", mxtic);
                 fprintf(fp,"set mytics %d\n", mytic);
                 fprintf(fp,"set grid y\n");
                 fprintf(fp,"ClusterSize=15\n");
                 fprintf(fp,"xcoord(i) = i * ClusterSize + column(1)\n");                 
            }else            //// line or line+dots
            {
                 fprintf(fp,"set xrange [%g:%g]\n",xmin, xmax);
                 fprintf(fp,"set yrange [%g:%g]\n",ymin, ymax);

                 //// draw tics and minor tics with 10 sub-intervals
                 fprintf(fp,"set xtics\n");
                 fprintf(fp,"set ytics\n");
                 fprintf(fp,"set mxtics %d\n", mxtic);
                 fprintf(fp,"set mytics %d\n", mytic);

                 //// set y axis tic label format
                 //// 12.345
          //     fprintf(fp,"set format y %c%c12.3f%c\n",34,37,34);
                 fprintf(fp,"set format x %c%c5.%df%c\n",34,37,xtic,34);
                 fprintf(fp,"set format y %c%c5.%df%c\n",34,37,ytic,34);

                 //// data point size
                 fprintf(fp,"set pointsize %g\n",pointsize);

                 if(xlog) fprintf(fp,"set logscale x\n");
                 if(ylog) fprintf(fp,"set logscale y\n");
            }
       }else  //// autoscale
       { 
            range(mx, xlowest, xhighest);
            range(my, ylowest, yhighest);
            fprintf(fp,"set xrange [%g:%g]\n",xlowest-1, xhighest+1);
            fprintf(fp,"set yrange [%g:%g]\n",ylowest-1, yhighest+1);           
       }
       fprintf(fp,"unset key\n");                 
       fprintf(fp,"set boxwidth 0.85 abs\n");                      
       fprintf(fp,"plot  \"~temp.dat\" with %s ls 1 lc rgb %s \n",with_type, hexcolor);
       fflush(fp);    
       fclose(fp);
   }
}

//-------------------------------------------------------------------------//
// graph_update                                                            //
//-------------------------------------------------------------------------//
void graph_update(dialoginfo *a, int radio, int radiobutton, int box, int boxbutton)
{
   if(box==21) graph_draw(*a);
}  


//-------------------------------------------------------------------------//
// graph_close                                                             //
//-------------------------------------------------------------------------//
void graph_close(void)
{
   //// All this does is tell the program it will need to call gv again.
   g.graph_active=0;
}

//-------------------------------------------------------------------------//
// reorder - put columns in a different order                              //
//-------------------------------------------------------------------------//
void reorder(void)
{
   int row, col, kk, noofargs, jj;
   double temp[s.cols];
   char column_numbers[s.cols*10];
   column_numbers[0] = 0;
   message((char*)"Enter new column numbers separated by spaces", 
       column_numbers, PROMPT, s.cols*10);
   char **newcol, **newtitle, **newcollabel;

   //// array of strings "1" "2" "4" "3" etc
   newcol = new char*[s.cols];
   for(kk=0; kk<s.cols; kk++){ newcol[kk] = new char[128]; newcol[kk][0]=0;}

   //// copy of column titles
   newtitle = new char*[s.cols];
   for(kk=0; kk<s.cols; kk++){ newtitle[kk] = new char[128]; newtitle[kk][0]=0;}

   //// copy of column labels
   newcollabel = new char*[s.cols];
   for(kk=0; kk<s.cols; kk++){ newcollabel[kk] = new char[128]; newcollabel[kk][0]=0;}

   parse_string(column_numbers, newcol, NULL, noofargs, 0, 32, 99, ' ',0);

   //// re-order the data one line at a time
   for(row=1; row<=s.maxrow[1]; row++)
   {   for(col=1; col<s.cols-1; col++)
       {
            temp[col] = s.x[col][row];
       }    
       for(col=1; col<=noofargs; col++)
       {
            jj = atoi(newcol[col-1]);
            s.x[col][row] = temp[jj];
       }    
   }

   //// re-order the titles and column labels
   for(col=1; col<s.cols; col++)
   {
        strcpy(newtitle[col], s.title[col]);
        strcpy(newcollabel[col], s.collabel[col]);
   }    
   for(col=1; col<=noofargs; col++)
   {
        jj = atoi(newcol[col-1]);
        strcpy(s.title[col], newtitle[jj]);
        strcpy(s.collabel[col], newcollabel[jj]);
        setcoltitle(col, s.title[col]);
   }    

   for(kk=0; kk<s.cols; kk++) delete[] newcollabel[kk];
   delete[] newcollabel;
   for(kk=0; kk<s.cols; kk++) delete[] newtitle[kk];
   delete[] newtitle;
   for(kk=0; kk<s.cols; kk++) delete[] newcol[kk];
   delete[] newcol;
   setallcoltitles(s.title);
   refreshspreadsheet();
}
