//--------------------------------------------------------------------------//
// xdata4.cc                                                                //
// Latest revision: 05/23/2022                                              //
// Statistics                                                               //
//--------------------------------------------------------------------------//

#include "xdata.h"

extern Globals g;
extern Spreadsheet s;
static void quicksort(double  *a, int aSize, int *index); 
static int partition(double  *a, int first, int last, int *index);
static void quicksort(double  *a, int first, int last, int *index); 
static int cmpstringp(const void *p1, const void *p2);
double  u(int m, int n, double  t);
double  ppp(int n, int m, double  t, int M);

static void quicksort_string(char **a, int aSize, int *index); 
static void quicksort_string(char **a, int first, int last, int *index); 
static int partition_string(char **a, int first, int last, int *index); 
void sswap(char *a, char *b);
int alphabetical_before(char *a, char *b);


//--------------------------------------------------------------------------//
//  statistics                                                              //
//  basic statistics                                                        //
//  for geometric mean see EAE Habib, IJRAAS 11(3), 2012                    //
//      modified to include numbers <=0                                     //
//  www.arpapress.com/Volumes/Vol11Issue3/IJRRAS_11_3_08.pdf                //
//--------------------------------------------------------------------------//
void statistics()
{
  int j, noofcols;
  double  x=0.0, mean=0.0, sum=0.0, se=0.0, sd=0.0, gmean=0.0, product=1.0,
        p=0.0, xx=0.0, adev=0.0, curt=0.0, skew=0.0, variance=0.0;
  double  productplus=1.0, productminus=1.0, gmean_habib=1.0, aa,bb;
  static char tempstring[1024];
  char tempstring2[1024];
  SelectStruct data;
  data.s = NULL;
  SelectCountStruct sel;
  sel.start = NULL;
  sel.end = NULL;

  int count=0, countplus=0, countminus=0;
  count = count_selected_data(&sel, 0);
  if(count<2){ message((char*)"Insufficient cells selected",ERROR); return; }

  data.x = new double [count];
  if(data.x==NULL) return;  
  data.isnumber = new uchar[count];
  if(data.isnumber==NULL) { delete_clipboard(&data, count); return; }
  get_selected_data(&data, 0, count, noofcols);
  if(count==0) return;

  sum = 0.0;
  product = 1.0;
  productplus = 1.0;
  productminus = 1.0;
  for(j=0;j<count;j++) sum += data.x[j];
  mean = sum / count;
  for(j=0;j<count;j++) 
  {     x = data.x[j];
        adev += fabs((xx=x-mean));          
        variance += (p=xx*xx);
        skew += (p*=xx);
        curt += (p*=xx);
        product *= x;
        if (x<0){ countminus++; productminus *= -x;}
        if (x>0){ countplus++;  productplus  *=  x;}
  }

  adev /= count;
  variance /= (count-1);
  sd = sqrt(variance);
  se = sd / sqrt(count);
  aa = (double )countplus / (double )count;
  bb = (double )countminus/ (double )count;
  gmean_habib = aa * pow(productplus,  1/(double )countplus) 
              - bb * pow(productminus, 1/(double )countminus);
//  gmean = pow(product, 1.0/(double )count);
  gmean = gmean_habib;
  if(variance!=0)
  {   skew /= (count * variance * sd);
      curt = curt / (count * variance * variance) - 3;
  }

  sprintf(tempstring,"Statistics on selected cells\n\
   n           \t%d\n\
   sum         \t%g\n\
   mean        \t%g\n\
   variance    \t%g\n\
   std. dev.   \t%g\n\
   average dev.\t%g\n\
   std. err.   \t%g\n\
   geom. mean  \t%g",
       count,sum,mean,variance,sd,adev,se,gmean);
  if(variance!=0)
  {              
     sprintf(tempstring2,"\n\
   skew        \t%g\n\
   kurtosis    \t%g",skew,curt);
     strcat(tempstring, tempstring2);
  }
  edit((char*)"Statistics",(char*)"Statistics",tempstring,10,10,1024,0);
  delete_clipboard(&data, count);
}


//--------------------------------------------------------------------------//
//  statistics2                                                             //
//  basic statistics of a column, same as above but returns values          //
//  if n=0 uses entire column                                               //
//--------------------------------------------------------------------------//
void statistics2(int set, int n, double  &mean, double  &sum, double  &sd, 
     double  &se, double  &variance)
{
  int j;
  double  x=0.0,  product=1.0, p=0.0, xx=0.0, adev=0.0, curt=0.0, skew=0.0;
  SelectStruct data;
  data.s = NULL;

  n = 0; // not implemented yet
  variance=0.0;
  mean=0.0;
  sum=0.0;
  se=0.0;
  sd=0.0;

  int count=0;
  for(j=0;j<s.maxrow[set];j++) if(s.number[set][j]) count++;

  data.x = new double [count];
  if(data.x==NULL) return;  
  data.isnumber = new uchar[count];
  if(data.isnumber==NULL) { delete_clipboard(&data, count); return; }
  get_column(&data, set);

  sum = 0.0;
  product = 1.0;
  for(j=0;j<count;j++) if(s.number[set][j]) sum += data.x[j];
  mean = sum / count;
  for(j=0;j<count;j++) 
  {     
        if(!s.number[set][j]) continue;
        x = data.x[j];
        adev += fabs((xx=x-mean));          
        variance += (p=xx*xx);
        skew += (p*=xx);
        curt += (p*=xx);
        product *= x;
  }

  adev /= count;
  variance /= (count-1);
  sd = sqrt(variance);
  se = sd / sqrt(count);
//  gmean = pow(product, 1.0/(double )count);
  if(variance!=0)
  {   skew /= (count * variance * sd);
      curt = curt / (count * variance * variance) - 3;
  }
}


//--------------------------------------------------------------------------//
//  ttest1 - students t test                                                //
//  Assumes 1 data set has same variance but possibly different means.      //
//--------------------------------------------------------------------------//
void ttest1(void)
{
   double  ave0, var0;
   double  svar;      // pooled variance of 2 data sets
   double  sed;       // s.e. of difference of means
   double  sd;        // s.d. of difference of means
   double  t;         // Students t statistic
   double  f;         // F value
   double  prob;      // Probability of a greater t
   double  testval;   // value to test if different from
   int df;           // degrees of freedom
   static char tempstring[1024]= "";
   static char mustring[1024]= "";

   ////  All this extra code just to get around compiler bugs in egcs
   ////  Cant return memory allocated in function anymore
   
   double  *data[1];
   int n[1];
   int sets[1];  
   sets[0]=0;
   n[0]=0;
   if(count_one_dataset(sets,n)!=OK) return;; 
   data[0] = new double [n[0]];  if(data[0]==NULL) return; 
   get_one_dataset(data, sets);
   message((char *)"Test value (mu0)", mustring, PROMPT, 64);
   testval = atof(mustring);
 
   if(n[0]<1){message((char*)"Insufficient data points"); return; }
   avevar(data[0], n[0], &ave0, &var0);
   df = n[0] - 1;
   svar = ((double )(n[0]-1)*var0 )/df;
   sd = sqrt(svar);
   sed = sqrt(svar*(1.0/(double )n[0] ));
   if(sed==0){ message((char*)"Std. err is 0", ERROR); return; }
   t = fabs(ave0-testval) / sed;
   prob = betai(0.5*df, 0.5, df/(df+t*t));
   f = fvalue1(data[0], n[0], testval);

   sprintf(tempstring,"\
   average       \t%g\n\
   variance      \t%g\n\
   test value    \t%g\n\
   deg. freedom  \t%d\n\
   Std dev       \t%g\n\
   Std err       \t%g\n\
   F value       \t%g\n\
   t value       \t%g\n\
   p(>t)         \t%g", ave0,var0,testval,df,sd,sed,f,t,prob);
   
   edit((char*)"t test results",(char*)"t test results",tempstring,8,30,1024,0);
   delete[] data[0];
}


//--------------------------------------------------------------------------//
//  ttest2 - students t test                                                //
//  Assumes 2 data sets have same variance but possibly different means.    //
//--------------------------------------------------------------------------//
void ttest2(void)
{
   double  ave0, ave1, var0, var1;
   double  svar;      // pooled variance of 2 data sets
   double  sed;       // s.e. of difference of means
   double  t;         // Students t statistic
   double  f;         // F value
   double  prob2tail; // Probability of a greater t
   double  prob1tail; // Probability of a greater t
   int df;           // degrees of freedom
   static char tempstring[1024];

   ////  All this extra code just to get around compiler bugs in egcs
   ////  Cant return memory allocated in function anymore
   
   double  *data[2];
   int n[2];
   int sets[2];  
   sets[0]=0;
   sets[1]=0;
   n[0]=0;
   n[1]=0;
   if(count_two_datasets(sets,n)!=OK) return;; 
   data[0] = new double [n[0]];  if(data[0]==NULL) return; 
   data[1] = new double [n[1]];  if(data[1]==NULL) return; 
   get_two_datasets(data, sets);
 
   if(n[0]<2 || n[1]<2){message((char*)"Insufficient data points"); return; }
   avevar(data[0], n[0], &ave0, &var0);
   avevar(data[1], n[1], &ave1, &var1);
   df = n[0] + n[1] - 2;
   svar = ((double )(n[0]-1)*var0 + (double )(n[1]-1)*var1)/df;
   sed = sqrt(svar*(1.0/(double )n[0] + 1.0/(double )n[1]));
   if(sed==0){ message((char*)"Std. err is 0", ERROR); return; }
   t = fabs(ave0-ave1) / sed;
   prob2tail = betai(0.5*df, 0.5, df/(df+t*t));
   prob1tail = prob2tail / 2.0;
   f = fvalue2(data[0], n[0], data[1], n[1]);

   sprintf(tempstring,"\
   average #1    \t%g\n\
   average #2    \t%g\n\
   variance #1   \t%g\n\
   variance #2   \t%g\n\
   deg. freedom  \t%d\n\
   S.E. of diff. \t%g\n\
   F value       \t%g\n\
   t value       \t%g\n\
   \n\
   p(>t>   (2-tailed) \t%g\n\
   p(>t)   (1-tailed) \t%g", ave0, ave1, var0, var1, df, sed, f,
                             t, prob2tail, prob1tail);
   
   edit((char*)"t test results - Equal variances",(char*)"t test results - Equal variances",tempstring,9,32,1024,0);
   delete[] data[1];  
   delete[] data[0];
}



//--------------------------------------------------------------------------//
//  ttest_unequal - students t test, unequal variances, 2 data sets         //
//  Welch's t test                                                          //
//--------------------------------------------------------------------------//
void ttest_unequal(void)
{
   double  ave0, ave1, var0, var1;
   double  s1,s2,s12; // 
   double  t;         // Students t statistic
   double  f;         // F value
   double  prob;      // Probability of a greater t
   double  df;        // degrees of freedom - not an integer here
   static char tempstring[1024];

   ////  All this extra code just to get around compiler bugs in egcs
   ////  Cant return memory allocated in function anymore
   
   double  *data[2];
   int n[2];
   int sets[2];  
   sets[0]=0;
   sets[1]=0;
   n[0]=0;
   n[1]=0;
   if(count_two_datasets(sets,n)!=OK) return;; 
   data[0] = new double [n[0]];  if(data[0]==NULL) return; 
   data[1] = new double [n[1]];  if(data[1]==NULL) return; 
   get_two_datasets(data, sets);
 
   if(n[0]<2 || n[1]<2){message((char*)"Insufficient data points"); return; }
   avevar(data[0], n[0], &ave0, &var0);
   avevar(data[1], n[1], &ave1, &var1);

   s1 = sqrt(var0);
   
   s2 = sqrt(var1);

   //// welch-satterthwaite eq
   df = (  (s1*s1/(double )n[0] + s2*s2/(double )n[1]) * (s1*s1/(double )n[0] + s2*s2/(double )n[1]) ) /
        (  ((s1*s1/(double )n[0] * s1*s1/(double )n[0]) / ((double )n[0]-1.0)) + ((s2*s2/(double )n[1] * s2*s2/n[1]) / ((double )n[1]-1.0)) );

   s12 = sqrt( s1*s1/(double )n[0] + s2*s2/(double )n[1] );

   t = fabs(ave0 - ave1) / s12;

   prob = betai(0.5*df, 0.5, df/(df+t*t));
   f = fvalue2(data[0], n[0], data[1], n[1]);

   sprintf(tempstring,"\
   average #1    \t%g\n\
   average #2    \t%g\n\
   variance #1   \t%g\n\
   variance #2   \t%g\n\
   deg. freedom  \t%g\n\
   s12           \t%g\n\
   F value       \t%g\n\
   t value       \t%g\n\
   \n\
   p(>t)   (2-tailed) \t%g\n\
   p(>t)   (1-tailed) \t%g", ave0,ave1,var0,var1,df,s12,f,t,prob,prob/2.0);
   
   edit((char*)"t test results - Unequal variances",(char*)"t test results - Unequal variances",tempstring,10,32,1024,0);
   delete[] data[1];  
   delete[] data[0];
}


//--------------------------------------------------------------------------//
//  ttestfrommean1 - calculate students t value given mean and SD           //
//--------------------------------------------------------------------------//
void ttestfrommean1(void)
{
   Dialog dialog;
   int j,k;
   double  var1;
   double  svar;      // pooled variance of 2 data sets
   double  sed;       // s.e. of difference of means
   double  t;         // Students t statistic
   double  prob;      // Probability of a greater t
   double  testval = 0.0;
   int df;           // degrees of freedom
   static char tempstring[1024];
   static double  mean1=0.0, sd1=0.0;
   static int n1=0;

   dialog.needrepeat = 0;  
   strcpy(dialog.title,"T test 1 var from mean and SD");
   strcpy(dialog.boxes[0],"Group 1");

   strcpy(dialog.boxes[1],"Mean");
   strcpy(dialog.boxes[2],"Std Dev.");
   strcpy(dialog.boxes[3],"N");
   strcpy(dialog.boxes[4],"Test value");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=STRING; sprintf(dialog.answer[1], "%g", mean1);
   dialog.boxtype[2]=STRING; sprintf(dialog.answer[2], "%g", sd1);
   dialog.boxtype[3]=STRING; sprintf(dialog.answer[3], "%d", n1);
   dialog.boxtype[4]=STRING; sprintf(dialog.answer[4], "%g", testval);

   dialog.radiono[0]=0;
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=0;
   dialog.noofboxes=5;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   mean1   = atof(dialog.answer[1]);
   sd1     = atof(dialog.answer[2]);
   n1      = atoi(dialog.answer[3]);
   testval = atof(dialog.answer[4]);

   tempstring[0] = 0;
 
   if(n1<2){message((char*)"Insufficient data points"); return; }
   df = n1 - 1;
   var1 = sd1 * sd1;
   svar = ((double )(n1-1)*var1)/df;
   sed = sqrt(svar*(1.0/(double )n1));
   if(sed==0){ message((char*)"Std. err is 0", ERROR); return; }
   t = fabs(mean1  -testval) / sed;
   prob = betai(0.5*df, 0.5, df/(df+t*t));

   sprintf(tempstring,"\
   average       \t%g\n\
   variance      \t%g\n\
   test value    \t%g\n\
   deg. freedom  \t%d\n\
   S.D. of diff. \t%g\n\
   S.E. of diff. \t%g\n\
   t value       \t%g\n\
   p(>t)         \t%g", mean1,var1,testval,df,sd1,sed,t,prob);
   
   edit((char*)"t test results",(char*)"t test results",tempstring,8,30,1024,0);
}


//--------------------------------------------------------------------------//
//  ttestfrommeanse1 - calculate students t value given mean and SE         //
//--------------------------------------------------------------------------//
void ttestfrommeanse1(void)
{
   Dialog dialog;
   int j,k;
   double  var1;
   double  svar;      // pooled variance of 2 data sets
   double  sed;       // s.e. of difference of means
   double  sd;        // s.d.
   double  t;         // Students t statistic
   double  prob;      // Probability of a greater t
   double  testval = 0.0;
   int df;           // degrees of freedom
   static char tempstring[1024];
   static double  mean1=0.0, se1=0.0;
   static int n1=0;

   dialog.needrepeat = 0;  
   strcpy(dialog.title,"T test 1 var from mean and SE");
   strcpy(dialog.boxes[0],"Group 1");

   strcpy(dialog.boxes[1],"Mean");
   strcpy(dialog.boxes[2],"Std Err.");
   strcpy(dialog.boxes[3],"N");
   strcpy(dialog.boxes[4],"Test value");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=STRING;  sprintf(dialog.answer[1], "%g", mean1);
   dialog.boxtype[2]=STRING;  sprintf(dialog.answer[2], "%g", se1);
   dialog.boxtype[3]=STRING;  sprintf(dialog.answer[3], "%d", n1);
   dialog.boxtype[4]=STRING;  sprintf(dialog.answer[4], "%g", testval);

   dialog.radiono[0]=0;
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=0;
   dialog.noofboxes=5;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   mean1 = atof(dialog.answer[1]);
   se1   = atof(dialog.answer[2]);
   n1    = atoi(dialog.answer[3]);
   testval = atof(dialog.answer[4]);

   tempstring[0] = 0;
 
   sd = se1 * sqrt(n1);
   if(n1<2){message((char*)"Insufficient number of data points"); return; }
   df = n1 - 1;
   var1 = sd * sd;
   svar = ((double )(n1-1)*var1)/df;
   sed = sqrt(svar*(1.0/(double )n1));  //// same as se1
   if(sed==0){ message((char*)"Std. err is 0", ERROR); return; }
   t = fabs(mean1 - testval) / sed;
   prob = betai(0.5*df, 0.5, df/(df+t*t));

   sprintf(tempstring,"\
   average       \t%g\n\
   variance      \t%g\n\
   test value    \t%g\n\
   deg. freedom  \t%d\n\
   S.D. of diff. \t%g\n\
   S.E. of diff. \t%g\n\
   t value       \t%g\n\
   p(>t)         \t%g", mean1,var1,testval,df,sd,sed,t,prob);
   
   edit((char*)"t test results",(char*)"t test results",tempstring,8,30,1024,0);
}

//--------------------------------------------------------------------------//
//  ttestfrommean2 - calculate students t value given mean and SD           //
//--------------------------------------------------------------------------//
void ttestfrommean2(void)
{
   Dialog dialog;
   int j,k;
   double  var1, var2;
   double  svar;      // pooled variance of 2 data sets
   double  sed;       // s.e. of difference of means
   double  t;         // Students t statistic
   double  prob;      // Probability of a greater t
   int df;           // degrees of freedom
   static char tempstring[1024];
   static double  mean1=0.0, mean2=0.0, sd1=0.0, sd2=0.0;
   static int n1=0, n2=0;

   dialog.needrepeat = 0;  
   strcpy(dialog.title,"Two-sample T test from mean and SD");
   strcpy(dialog.boxes[0],"Group 1");
   strcpy(dialog.boxes[1],"Mean");
   strcpy(dialog.boxes[2],"Std Dev.");
   strcpy(dialog.boxes[3],"N");
   strcpy(dialog.boxes[4],"Group 2");
   strcpy(dialog.boxes[5],"Mean");
   strcpy(dialog.boxes[6],"Std Dev.");
   strcpy(dialog.boxes[7],"N");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=STRING; sprintf(dialog.answer[1], "%g", mean1);
   dialog.boxtype[2]=STRING; sprintf(dialog.answer[2], "%g", sd1);
   dialog.boxtype[3]=STRING; sprintf(dialog.answer[3], "%d", n1);
   dialog.boxtype[4]=LABEL;
   dialog.boxtype[5]=STRING; sprintf(dialog.answer[5], "%g", mean2);
   dialog.boxtype[6]=STRING; sprintf(dialog.answer[6], "%g", sd2);
   dialog.boxtype[7]=STRING; sprintf(dialog.answer[7], "%d", n2);

   dialog.radiono[0]=0;
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=0;
   dialog.noofboxes=8;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   mean1 = atof(dialog.answer[1]);
   sd1   = atof(dialog.answer[2]);
   n1    = atoi(dialog.answer[3]);
   mean2 = atof(dialog.answer[5]);
   sd2   = atof(dialog.answer[6]);
   n2    = atoi(dialog.answer[7]);

   tempstring[0] = 0;
 
   if(n1<2 || n2<2){message((char*)"Insufficient data points"); return; }
   df = n1 + n2 - 2;
   var1 = sd1 * sd1;
   var2 = sd2 * sd2;
   svar = ((double )(n1-1)*var1 + (double )(n2-1)*var2)/df;
   sed = sqrt(svar*(1.0/(double )n1 + 1.0/(double )n2));
   if(sed==0){ message((char*)"Std. err is 0", ERROR); return; }
   t = fabs(mean1-mean2) / sed;
   prob = betai(0.5*df, 0.5, df/(df+t*t));

   sprintf(tempstring,"\
   average #1    \t%g\n\
   average #2    \t%g\n\
   variance #1   \t%g\n\
   variance #2   \t%g\n\
   deg. freedom  \t%d\n\
   S.E. of diff. \t%g\n\
   t value       \t%g\n\
   p(>t)         \t%g", mean1,mean2,var1,var2,df,sed,t,prob);
   
   edit((char*)"t test results",(char*)"t test results",tempstring,8,30,1024,0);
}


//--------------------------------------------------------------------------//
//  ttestfrommeanse2 - calculate students t value given mean and SE         //
//--------------------------------------------------------------------------//
void ttestfrommeanse2(void)
{
   Dialog dialog;
   int j,k;
   double  var1, var2;
   double  svar;      // pooled variance of 2 data sets
   double  sed;       // s.e. of difference of means
   double  t;         // Students t statistic
   double  prob;      // Probability of a greater t
   int df;           // degrees of freedom
   static char tempstring[1024];
   static double  mean1=0.0, mean2=0.0, se1=0.0, se2=0.0, sd1=0.0, sd2=0.0;
   static int n1=0, n2=0;

   dialog.needrepeat = 0;  
   strcpy(dialog.title,"Two sample T test from mean and SE");
   strcpy(dialog.boxes[0],"Group 1");
   strcpy(dialog.boxes[1],"Mean");
   strcpy(dialog.boxes[2],"Std Error.");
   strcpy(dialog.boxes[3],"N");
   strcpy(dialog.boxes[4],"Group 2");
   strcpy(dialog.boxes[5],"Mean");
   strcpy(dialog.boxes[6],"Std Error.");
   strcpy(dialog.boxes[7],"N");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=STRING;  sprintf(dialog.answer[1], "%g", mean1);
   dialog.boxtype[2]=STRING;  sprintf(dialog.answer[2], "%g", se1);
   dialog.boxtype[3]=STRING;  sprintf(dialog.answer[3], "%d", n1);
   dialog.boxtype[4]=LABEL;
   dialog.boxtype[5]=STRING;  sprintf(dialog.answer[5], "%g", mean2);
   dialog.boxtype[6]=STRING;  sprintf(dialog.answer[6], "%g", se2);
   dialog.boxtype[7]=STRING;  sprintf(dialog.answer[7], "%d", n2);

   dialog.radiono[0]=0;
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=0;
   dialog.noofboxes=8;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   mean1 = atof(dialog.answer[1]);
   se1   = atof(dialog.answer[2]);
   n1    = atoi(dialog.answer[3]);
   mean2 = atof(dialog.answer[5]);
   se2   = atof(dialog.answer[6]);
   n2    = atoi(dialog.answer[7]);

   tempstring[0] = 0;
 
   sd1 = se1 * sqrt(n1);
   sd2 = se2 * sqrt(n2);
   if(n1<2 || n2<2){message((char*)"Insufficient number of data points"); return; }
   df = n1 + n2 - 2;
   var1 = sd1 * sd1;
   var2 = sd2 * sd2;
   svar = ((double )(n1-1)*var1 + (double )(n2-1)*var2)/df;
   sed = sqrt(svar*(1.0/(double )n1 + 1.0/(double )n2));
   if(sed==0){ message((char*)"Std. err is 0", ERROR); return; }
   t = fabs(mean1-mean2) / sed;
   prob = betai(0.5*df, 0.5, df/(df+t*t));

   sprintf(tempstring,"\
   average #1    \t%g\n\
   average #2    \t%g\n\
   variance #1   \t%g\n\
   variance #2   \t%g\n\
   deg. freedom  \t%d\n\
   S.E. of diff. \t%g\n\
   t value       \t%g\n\
   p(>t)         \t%g", mean1,mean2,var1,var2,df,sed,t,prob);
   
   edit((char*)"t test results",(char*)"t test results",tempstring,8,30,1024,0);
}


//--------------------------------------------------------------------------//
//  tptest                                                                  //
//  students t test for paired samples                                      //
//--------------------------------------------------------------------------//
void tptest(void)
{
   int j;
   double  cov = 0.0;
   double  ave0, ave1, var0, var1;
   double  sd;  
   double  t;         // Students t statistic
   double  prob;      // Probability of a greater t
   int df;           // degrees of freedom
   static char tempstring[1024];

   double  *data[2];
   int n[2];
   int sets[2];
   sets[0]=0;
   sets[1]=0;
   n[0]=0;
   n[1]=0;
   if(count_two_datasets(sets,n)!=OK) return;; 
   data[0] = new double [n[0]];  if(data[0]==NULL) return; 
   data[1] = new double [n[1]];  if(data[0]==NULL) return; 
   get_two_datasets(data, sets);

   if(n[0]<2 || n[1]<2){message((char*)"Insufficient number of data points"); return; }
   if(n[0] != n[1]){message((char*)"Data sets have different length",ERROR); return; }
   avevar(data[0], n[0], &ave0, &var0);
   avevar(data[1], n[1], &ave1, &var1);
   
   for(j=0;j<n[0];j++)
       cov += (data[0][j] - ave0) * (data[1][j] - ave1);
   cov /= df = n[0]-1;
   sd = sqrt((var0 + var1 - 2.0*cov)/n[0]);
   t = (ave0 - ave1)/sd;
   prob = betai(0.5*df, 0.5, df/(df+t*t));
   
   sprintf(tempstring,"\
   average #1   \t%g\n\
   average #2   \t%g\n\
   variance #1  \t%g\n\
   variance #2  \t%g\n\
   deg. freedom \t%d\n\
   sd           \t%g\n\
   t value      \t%g\n\
   p(>t)        \t%g", ave0,ave1,var0,var1,df,sd,t,prob);  
   edit((char*)"t test results", (char*)"t test results for paired samples",tempstring,8,30,1024,0);
   delete[] data[1];  
   delete[] data[0];
}


//--------------------------------------------------------------------------//
//  ftest                                                                   //
//--------------------------------------------------------------------------//
void ftest(void)
{
   double  ave0, ave1, var0, var1;
   double  ff;        // f ratio statistic
   double  prob;      // Probability of a greater t
   int df1, df2;     // degrees of freedom
   static char tempstring[1024];

   double  *data[2];
   int n[2];
   int sets[2];
   sets[0]=0;
   sets[1]=0;
   n[0]=0;
   n[1]=0;
   if(count_two_datasets(sets,n)!=OK) return;; 
   data[0] = new double [n[0]];  if(data[0]==NULL) return; 
   data[1] = new double [n[1]];  if(data[0]==NULL) return; 
   get_two_datasets(data, sets);

  
   if(n[0]<2 || n[1]<2){message((char*)"Insufficient number of data points"); return; }
   avevar(data[0], n[0], &ave0, &var0);
   avevar(data[1], n[1], &ave1, &var1);

   if(var0 > var1)
   {  ff  = var0/var1;
      df1 = n[0]-1;   
      df2 = n[1]-1;   
   }else
   {  ff = var1/var0;
      df1 = n[1]-1;   
      df2 = n[0]-1;   
   }
   prob = 2.0*betai(0.5*df2, 0.5*df1, df2/(df2+df1*ff));
   if(prob > 1.0) prob = 2.0 - prob;
   
   sprintf(tempstring,"\
   no. of data points  \t%d\n\
   average #1   \t%g\n\
   average #2   \t%g\n\
   variance #1  \t%g\n\
   variance #2  \t%g\n\
   deg. freedom \t%d + %d\n\
   f ratio      \t%g\n\
   probability  \t%g", n[0]+n[1],ave0,ave1,var0,var1,df1,df2,ff,prob);  
   edit((char*)"f statistics",(char*)"f statistics",tempstring,8,30,1024,0);
   delete[] data[1];  
   delete[] data[0];
}



//--------------------------------------------------------------------------//
//  avevar                                                                  //
//  compute mean and variance of a data set                                 //
//--------------------------------------------------------------------------//
void avevar(double  *data, int count, double  *ave, double  *var)
{
     int j;
     double  s,ep;
     *ave=0.0;
     for(j=0;j<count;j++) *ave += data[j];
     *ave /= count;
     *var = ep = 0.0;
     for(j=0;j<count;j++)
     {   s = data[j] - *ave;
         ep += s;
         *var  += s*s;
     }
     *var = (*var - ep*ep/count) / (count-1);
}


//--------------------------------------------------------------------------//
//  fvalue1                                                                 //
//--------------------------------------------------------------------------//
double  fvalue1(double  *data0, int n0, double  testval)
{
     int j,n,df;
     double  f,sb,ss,ss0=0.0,sw,dev,ave,ave0,ave1,var;

     //// Between-sample variation
     ave = ave0 = ave1 = 0.0;
     for(j=0; j<n0; j++) ave0 += data0[j];
     ave1 = testval;
     ave0 /= n0;    
     df = n0;
     n = n0;
     ave = ave0 / n;           // Mean of the sample means

     var = ss = 0.0;
     var = var;
     for(j=0; j<n0; j++)
     {   dev = data0[j] - ave0;
         ss0 += dev * dev;
     }
     sw = ss0 / (2 * df);

     //// Within-sample variation

     ss = (n0*(ave0-ave)*(ave0-ave));  // 1 = groups-1
     sb = ss / (n0 - 1);    // N - r

     //// F is ratio of between-sample to within-sample variation
     f = sb / sw; 
     f = f * double (n0) / double (n0-1);
     return f;
}


//--------------------------------------------------------------------------//
//  fvalue2                                                                 //
//--------------------------------------------------------------------------//
double  fvalue2(double  *data0, int n0, double  *data1, int n1)
{
     int j,n,df;
     double  f,sb,ss,ss0=0.0,ss1=0.0,sw,dev,ave,ave0,ave1,var;

     //// Between-sample variation
     ave = ave0 = ave1 = 0.0;
     for(j=0; j<n0; j++) ave0 += data0[j];
     for(j=0; j<n1; j++) ave1 += data1[j];
     ave0 /= n0;    
     ave1 /= n1;    
     df = n0 + n1;
     n = (n0 + n1) / 2;
     ave = (ave0 + ave1) / n;           // Mean of the sample means

     var = ss = 0.0;
     var = var;
     for(j=0; j<n0; j++)
     {   dev = data0[j] - ave0;
         ss0 += dev * dev;
     }
     for(j=0; j<n1; j++)
     {   dev = data1[j] - ave1;
         ss1 += dev * dev;
     }
     sw = (ss0 + ss1) / (2 * df);

     //// Within-sample variation

     ss = (n0*(ave0-ave)*(ave0-ave) +
           n1*(ave1-ave)*(ave1-ave)) / 1;  // 1 = groups-1
     sb = ss / (n0 + n1 - 2);    // N - r

     //// F is ratio of between-sample to within-sample variation
     f = sb / sw; 
     f = f * double (n0+n1) / double (n0+n1-1);
     return f;
}



//--------------------------------------------------------------------------//
//  betai                                                                   //
//  Incomplete beta function I$_x$(a,b)                                     //
//--------------------------------------------------------------------------//
double  betai(double  a, double  b, double  x)
{
    double  bt;
    if(x<0.0 || x>1.0){ message((char*)"Bad x in betai"); return 0.0; }
    if(x==0.0 || x==1.0) 
        bt=0.0;
    else
        bt = exp(gammln(a+b) - gammln(a) - gammln(b) + a*log(x) + b*log(1.0-x));
    if(x < (a+1.0)/(a+b+1.0))          // Use continued fraction
        return bt*betacf(a,b,x)/a;
    else                               // Make symmetry transformation
        return 1.0 - bt*betacf(b,a,1.0-x)/b;        
}

//--------------------------------------------------------------------------//
//  betacf                                                                  //
//  Continued fraction evaluation for incomplete beta function using        //
//  modified Lentz's method                                                 //
//--------------------------------------------------------------------------//
double  betacf(double  a, double  b, double  x)
{
    const int MAXIT = 100;
    const double  EPS = 3.0e-7;
    int m,m2;
    double  aa,c,d,del, h,qab,qam,qap;
    
    qab = a+b;
    qap = a+1.0;
    qam = a-1.0;
    c = 1.0;
    d = notzero(1.0 - qab*x/qap);
    h = d = 1.0/d;
    for(m=1;m<=MAXIT;m++)    
    {   m2 = 2*m;
        aa = m*(b-m)*x / ((qam+m2)*(a+m2));
        d = notzero(1.0 + aa*d);
        c = notzero(1.0 + aa/c);
        d = 1.0/d;
        h *= d*c;
        aa = -(a+m)*(qab+m)*x / ((a+m2)*(qap+m2));
        d = notzero(1.0 + aa*d);
        c = notzero(1.0 + aa/c);
        d = 1.0/d;
        del = d*c;
        h *= del;
        if(fabs(del - 1.0) < EPS) break;  // termination condition
    }
    if(m>MAXIT) message((char*)"Too many iterations in betacf",ERROR);    
    return h;
}

//--------------------------------------------------------------------------//
//  notzero - prevent a number from being 0                                 //
//--------------------------------------------------------------------------//
double  notzero(double  a)
{
    const double  FPMIN = 1.0e-30;
    if(fabs(a)<FPMIN) a = FPMIN;
    return a;
}



//--------------------------------------------------------------------------//
//  chisq1 - chi square test for one set of observations                    //
//  data[0] is observed numbers of bins                                     //
//  data[1] is expected numbers of bins                                     //
//--------------------------------------------------------------------------//
void chisq1(void)
{
   static char tempstring[1024];
   int j;
   int helptopic = 0;
   int df;
   double  temp, chisq=0.0, prob=0.0;
   static int knstrn = 1;   // No. of constraints
   knstrn = clickbox((char*)"No. of constraints", knstrn, 0, 2, null, helptopic);

   double  *data[2];
   int n[2];
   int sets[2];
   sets[0]=0;
   sets[1]=0;
   n[0]=0;
   n[1]=0;
   if(count_two_datasets(sets, n)!=OK) return;; 
   data[0] = new double [n[0]];  if(data[0]==NULL) return; 
   data[1] = new double [n[1]];  if(data[0]==NULL) return; 
   get_two_datasets(data, sets);


   if(n[0] != n[1]){message((char*)"Data sets have different length",ERROR); return;}

   df = n[0] - knstrn;  
   for(j=0;j<n[0];j++)
   {   if(data[1][j] < 0)
       {  message((char*)"Cannot have negative no. of events in chi square",ERROR); return; }
       temp = data[0][j] - data[1][j];
       chisq += temp*temp/data[1][j];
   }
   prob = gammq(0.5*df, 0.5*chisq);
   sprintf(tempstring,(char*)"Chi square results - observed vs. expected\n\
   df          \t%d\n\
   Chi Square  \t%g\n\
   Probability \t%g", df,chisq,prob);
   edit((char*)"Chi square",(char*)"Chi square",tempstring,3,30,1024,0);
   delete[] data[1];  
   delete[] data[0];
}


//--------------------------------------------------------------------------//
//  chisq2 - chi square test comparing 2 sets of observations               //
//--------------------------------------------------------------------------//
void chisq2(void)
{
   static char tempstring[1024];
   int j;
   int df;
   int helptopic = 0;
   double  temp, chisq=0.0, prob=0.0;
   static int knstrn = 1;   // No. of constraints
   knstrn = clickbox((char*)"No. of constraints", knstrn, 0, 2, null, helptopic);
 
   double  *data[2];
   int n[2];
   int sets[2];
   sets[0]=0;
   sets[1]=0;
   n[0]=0;
   n[1]=0;
   if(count_two_datasets(sets,n)!=OK) return;; 
   data[0] = new double [n[0]];  if(data[0]==NULL) return; 
   data[1] = new double [n[1]];  if(data[0]==NULL) return; 
   get_two_datasets(data, sets);

   if(n[0] != n[1]){message((char*)"Data sets have different length",ERROR); return;}

   df = n[0] - knstrn;  
   for(j=0;j<n[0];j++)
   {   if(data[0][j] == 0.0 && data[1][j] == 0.0) df--;
       else
       {    temp = data[0][j] - data[1][j];
            chisq += temp*temp/(data[0][j] + data[1][j]);
       }
   }
   prob = gammq(0.5*df, 0.5*chisq);
   sprintf(tempstring,"Chi square results - comparing 2 sets of observations\n\
   df          \t%d\n\
   Chi Square  \t%g\n\
   Probability \t%g", df,chisq,prob);
   edit((char*)"Chi square",(char*)"Chi square",tempstring,3,30,1024,0);
   delete[] data[1];  
   delete[] data[0];
}


//--------------------------------------------------------------------------//
//  chisq3 - chi square test comparing 2 sets of observations               //
//  georgetown website method                                               //
//          correct incorrect                                               //
//  group 1    10     7                                                     //
//  group 2    10     0                                                     //
//    column 1  = 10 out of 17 correct                                      //
//    column 2  = 10 out of 10 correct                                      //
//    chisq = 5.55882, p=0.0183878, p<0.025                                 //
//--------------------------------------------------------------------------//
void chisq3(void)
{
   static char tempstring[1024];

   int df;

   double  e, chisq=0.0, prob=0.0, chitot=0.0;
   double  a00, a10, a01, a11, total;

   // column row     
   a00 = s.x[1][1]; // row 0 col 0
   a01 = s.x[1][2]; // row 0 col 1
   a10 = s.x[2][1]; // row 1 col 0
   a11 = s.x[2][2]; // row 1 col 1
   total = a00+a01+a10+a11;
   df = 1;
   
   //row 0 col 0
   // e = expected value = row total x col total / grand total
   // (chisq = obs.erved value - e)^2 / e
   e = (a00+a10)*(a00+a01) / total;
   chisq = (a00 - e) * (a00 - e) / e;
   chitot += chisq;

   //row 0 col 1
   e = (a00+a10)*(a10+a11) / total;
   chisq = (a10 - e) * (a10 - e) / e;
   chitot += chisq;

   //row 1 col 0
   e = (a01+a11)*(a00+a01) / total;
   chisq = (a01 - e) * (a01 - e) / e;
   chitot += chisq;

   //row 1 col 1
   e = (a01+a11)*(a10+a11) / total;
   chisq = (a11 - e) * (a11 - e) / e;
   chitot += chisq;

   prob = gammq(0.5*df, 0.5*chitot);
   sprintf(tempstring,"Chi square results - comparing 2 observations\n\
Numbers are taken from first two rows of first two columns\n\
\t\tcorrect\tincorrect  \n\
   Group 1  \t%g\t %g \n\
   Group 2  \t%g\t %g  \n\
   DF          \t%d\n\
   Chi Square  \t%g\n\
   Probability \t%g", a00, a10, a01, a11, df,chitot,prob);
   edit((char*)"Chi square",(char*)"Chi square",tempstring,6,30,1024,0);
}


//--------------------------------------------------------------------------//
//  gammln - gamma function using logs                                      //
//--------------------------------------------------------------------------//
double  gammln(double  x)
{
    double  y,tmp,ser;
    const double  cof[6] = { 76.18009172947146,  -86.50532032941677,
                            24.01409824083091,   -1.231739572450155,
                          0.1208650973866179e-2, -0.5395239384953e-5 };

    ////  x should already be > 0 but check just in case.
    if(x<0){ message((char*)"Bad parameters in gammln",ERROR); return 0.0; }
    int j;
    y = x;
    tmp = x + 5.5;
    tmp -= (x+0.5) * log(tmp);
    ser = 1.000000000190015;
    for(j=0;j<=5;j++){ ser += cof[j]/++y; }
    return -tmp + log(2.5066282746310005*ser/x);   
}


//--------------------------------------------------------------------------//
//  gammq - incomplete gamma function Q(a,x) = 1- P(a,x)                   //
//--------------------------------------------------------------------------//
double  gammq(double  a, double  x)
{
    double  gamser, gammcf, gln;
    gamser = 0.0;
    if(x < 0.0 || a<= 0.0){ message((char*)"Bad parameters in gammq",ERROR); return 0.0; }
    if(x < a+1.0)
    {    gser(&gamser, a, x, &gln);
         return 1.0 - gamser;
    }else
    {    gcf(&gammcf, a, x, &gln);
         return gammcf;
    }
}


//--------------------------------------------------------------------------//
//  gammp - incomplete gamma function P(a,x)                                //
//--------------------------------------------------------------------------//
double  gammp(double  a, double  x)
{
    return 1.00 - gammq(a,x);
}


//--------------------------------------------------------------------------//
//  gser - incomplete gamma function P(a,x) evaluated by its series         //
//  representation as gamser. Also returns ln Gamma(a) as gln.              //
//--------------------------------------------------------------------------//
void gser(double  *gamser, double  a, double  x, double  *gln)
{
    int n;
    const int ITMAX = 200;
    const double  EPS = 3.0e-7;
    double  sum,del,ap;
 
    *gln = gammln(a);
    if(x<0.0){ message((char*)"x less than 0 in gser",ERROR); *gamser = 0.0; return; }   
    if(x==0.0) return;

    ap = a;
    del = sum = 1.0/a;
    for(n=0; n<ITMAX; n++)
    {   ++ap;
        del *= x/ap;
        sum += del;
        if(fabs(del) < fabs(sum)*EPS)
        {    *gamser = sum*exp(-x+a*log(x)-(*gln));
             return;
        }
    }
    message((char*)"a too large in gser",ERROR); 
    return;
}


//--------------------------------------------------------------------------//
//  gcf - incomplete gamma function P(a,x) evaluated by its continued       //
//  fraction representation representation as gammcf.  Also returns         //
//  ln Gamma(a) as gln.                                                     //
//--------------------------------------------------------------------------//
void gcf(double  *gammcf, double  a, double  x, double  *gln)
{
   int i;
   const int ITMAX = 100;
   const double  FPMIN = 1.0e-30;
   const double  EPS = 3.0e-7;
   double  an,b,c,d,del,h;
   *gln = gammln(a);
   b = x+1.0-a;
   c = 1.0/FPMIN;   
   d = 1.0/b;
   h=d;
   for(i=1;i<=ITMAX;i++)
   {    an = -i*(i-a);
        b += 2.0;
        d = notzero(an*d+b);
        c = notzero(b+an/c);
        d = 1.0/d;
        del = d*c;
        h *= del;
        if(fabs(del-1.0) < EPS) break;
   }
   if(i>=ITMAX) message((char*)"a too large in gcg",ERROR); 
   *gammcf = exp(-x+a*log(x) - (*gln))*h;
}


//--------------------------------------------------------------------//
//  sort                                                              //
//--------------------------------------------------------------------//
void sort(char **arg, int noofargs)
{
  int k, col;
  char tempstring[128];
  int found_number = 0, found_string = 0;
  arg=arg; noofargs=noofargs;
  itoa(s.current_col, tempstring, 10);
  if(noofargs<2) 
  {     message((char*)"Column no.", tempstring, PROMPT, 1023);
        col = atoi(tempstring);
  }else col = atoi(arg[1]);

  //// qsort is part of stdlib, starts at 0 but our arrays start
  //// at 1.

  for(k=1; k<=s.maxrow[col]; k++)
          if(s.number[col][k]) found_number = 1; else found_string = 1;
  if(found_string && found_number) 
          message((char*)"Mixed numeric and strings\nsorting by text");
  if(found_string) 
          qsort(s.text[col]+1, s.maxrow[col], sizeof(char *), cmpstringp);
  else if(found_number)
          qsort(s.x[col]+1, s.maxrow[col], sizeof(double ), compare);
  else
          message((char*)"No sortable entries found");
  g.touched = 1;
  refreshspreadsheet();
}


//--------------------------------------------------------------------------//
//  compare - for qsort                                                     //
//--------------------------------------------------------------------------//
int compare(const void *p, const void *q)
{
   return sgn(*(double *)p - *(double *)q);
}


//--------------------------------------------------------------------//
//  cmpstringp                                                        //
//--------------------------------------------------------------------//
static int cmpstringp(const void *p1, const void *p2)
{
   return strcmp(* (char * const *) p1, * (char * const *) p2);
}


//--------------------------------------------------------------------//
//  linkedsort                                                        //
//--------------------------------------------------------------------//
void linkedsort(char **arg, int noofargs)
{
  static int col2=0, col3=0;
  int found_number = 0, found_string = 0;
  int col,i,k,j; 
  char tempstring[1024];
  int *is_string;     //// for each column, 0=numbers 1=strings 
  int *index;        //// list of how it should be sorted
  char **temp;       //// temporary list of strings
  double  *xtemp;     //// temporary list of numbers
  arg=arg; noofargs=noofargs;
  itoa(s.current_col, tempstring, 10);
  if(noofargs<=1) 
  {     message((char*)"Column no. to sort", tempstring, PROMPT, 1023);
        col = atoi(tempstring);
  }else col = atoi(arg[1]);

  if(noofargs<=2) 
  {     if(col2==0) col2 = col;
        sprintf(tempstring, "%d", col2);
        message((char*)"Starting Linked Column no.", tempstring, PROMPT, 1023);
        col2 = atoi(tempstring);
  }else col2 = atoi(arg[2]);

  if(noofargs<=3) 
  {     if(col3==0) col3 = s.maxcol;
        sprintf(tempstring, "%d", col3);
        message((char*)"Ending Linked Column no.", tempstring, PROMPT, 1023);
        col3 = atoi(tempstring);
  }else col3 = atoi(arg[3]);


  index = new int[s.maxrow[col]+1];
  is_string = new int[s.maxrow[col]+1];
  xtemp = new double [s.maxrow[col]+1];
  for(k=0; k<s.maxrow[col]; k++){ index[k] = k; xtemp[k] = 0; }
  temp = new char *[s.rows];
  for(i=0; i<s.rows; i++) temp[i] = new char[1024];
  
  //// Sort the column of interest and save its indexes
  //// quicksort starts at 0, our vector starts at 1
  //// xtemp, index start at 0

  //// make sure no columns have both strings and numbers,
  //// cant figure out how to sort that

  for(j=col2; j<=col3; j++)
  {    found_string = found_number = 0;
       for(k=1; k<=s.maxrow[j]; k++)
            if(s.string[j][k]) found_string++; else found_number++;
       
       if(found_string && found_number)
       {    sprintf(tempstring, (char*)"Mixed strings and numbers found in column %d",j);
            message(tempstring, ERROR); 
            goto the_end; 
       }
       is_string[j] = found_string;  //// flag column as all strings or no strings
  }
         

  //// create an index of how to sort
  //// create index sorted numerically if col is all numbers
  //// or alphabetically is col is all strings

  if(!is_string[col])
  { 
       for(k=1; k<=s.maxrow[col]; k++)
       {   for(k=0; k<=s.maxrow[col]; k++) xtemp[k] = s.x[col][k+1];
           quicksort(xtemp, s.maxrow[col], index);
       }
  }else
  {     
       for(k=1; k<=s.maxrow[col]; k++)
       {   for(k=0; k<=s.maxrow[col]; k++) strcpy(temp[k], s.text[col][k+1]);
           quicksort_string(temp, s.maxrow[col], index);
       }
  }

  //// sort according to the index
  for(j=col2; j<=col3; j++)
  {    if(!is_string[j])
       {
           for(k=0; k<s.maxrow[j]; k++)
                 xtemp[k] = s.x[j][index[k]+1];
           for(k=0; k<s.maxrow[j]; k++)
                 s.x[j][k+1] = xtemp[k]; 
       }else
       { 
           for(k=0; k<s.maxrow[j]; k++)
                strcpy(temp[k], s.text[j][index[k]+1]);
           for(k=0; k<s.maxrow[j]; k++)
                strcpy(s.text[j][k+1], temp[k]);
       }
  }

  refreshspreadsheet();
the_end:
  for(i=0;i<s.rows;i++) delete[] temp[i];
  delete[] is_string;
  delete[] temp;
  delete[] index;
  delete[] xtemp;

  g.touched = 1;
}

//--------------------------------------------------------------------//
//  custom quicksort for double s                                      //
//  index must start out as sequence of integers                      //
//  index and a are both changed                                      //
//--------------------------------------------------------------------//
static void quicksort(double  *a, int aSize, int *index) 
{
    quicksort(a, 0, aSize - 1, index);
}

static void quicksort(double  *a, int first, int last, int *index) 
{
  if (first < last) 
  {    int pivotIndex = partition(a, first, last, index);
       quicksort(a, first, pivotIndex - 1, index);
       quicksort(a, pivotIndex + 1, last, index);
  }
}

static int partition(double  *a, int first, int last, int *index) 
{
  double  pivot = a[first];
  int lastS1 = first;
  int firstUnknown = first + 1;
  while (firstUnknown <= last) 
  {   if (a[firstUnknown] < pivot) 
      {    lastS1++;
           fswap(a[firstUnknown], a[lastS1]);
           swap(index[firstUnknown], index[lastS1]);
      }
      firstUnknown++;
  }
  fswap(a[first], a[lastS1]);
  swap(index[first], index[lastS1]);
  return lastS1;
}


//--------------------------------------------------------------------//
//  custom quicksort for strings                                      //
//--------------------------------------------------------------------//
static void quicksort_string(char **a, int aSize, int *index) 
{
    quicksort_string(a, 0, aSize - 1, index);
}

static void quicksort_string(char **a, int first, int last, int *index) 
{
  if (first < last) 
  {    int pivotIndex = partition_string(a, first, last, index);
       quicksort_string(a, first, pivotIndex - 1, index);
       quicksort_string(a, pivotIndex + 1, last, index);
  }
}

static int partition_string(char **a, int first, int last, int *index) 
{
  char *pivot = a[first];
  int lastS1 = first;
  int firstUnknown = first + 1;
  while (firstUnknown <= last) 
  {   if (alphabetical_before(a[firstUnknown], pivot)) 
      {    lastS1++;
           sswap(a[firstUnknown], a[lastS1]);
           swap(index[firstUnknown], index[lastS1]);
      }
      firstUnknown++;
  }
  sswap(a[first], a[lastS1]);
  swap(index[first], index[lastS1]);
  return lastS1;
}

void sswap(char *a, char *b)
{
   static char temp[CELLBYTES];
   strcpy(temp, a);
   strcpy(a, b);
   strcpy(b, temp);
}

int alphabetical_before(char *a, char *b)
{
   int k, answer=1;
   for(k=0; k<(int)strlen(a) && k<(int)strlen(b); k++)
   {
       answer = ((uint)a[k] - (uint)b[k]); 
       if(answer) break;
   }
   return answer<0;
}


//--------------------------------------------------------------------//
//  poisson                                                           //
//--------------------------------------------------------------------//
void poisson(char **arg, int noofargs)
{
  static int incol, outcol=2, cumcol=3;
  static double  xx=1.0;
  int row;
  char tempstring[128];
  int k;
  int sel_r1=-1, sel_r2=-1;
  double  p,pcum;
  double  lnnfactorial;
  double  n;

  arg=arg; noofargs=noofargs;
  itoa(s.current_col, tempstring, 10);
  if(noofargs<=1) 
  {     message((char*)"Column no. for data", tempstring, PROMPT, 1023);
        incol = atoi(tempstring);
  }else incol = atoi(arg[1]);

  if(outcol == 0) outcol = incol;
  if(noofargs<=2) 
  {     sprintf(tempstring, "%d", outcol); 
        message((char*)"Column no. for Poisson results", tempstring, PROMPT, 1023);
        outcol = atoi(tempstring);
  }else outcol = atoi(arg[2]);

  if(cumcol == 0) cumcol = outcol;
  if(noofargs<=3) 
  {     sprintf(tempstring, "%d", cumcol); 
        message((char*)"Column no. for cumulative Poisson results", tempstring, PROMPT, 1023);
        cumcol = atoi(tempstring);
  }else cumcol = atoi(arg[3]);

  if(noofargs<=4) 
  {     sprintf(tempstring, "%g", xx); 
        message((char*)"Poisson test value", tempstring, PROMPT, 1023);
        xx = atof(tempstring);
  }else xx = atof(arg[4]);
 
  //// If new column, initialize it
  s.maxcol = max(s.maxcol, 1+outcol);
  s.maxrow[outcol] = max(s.maxrow[incol], s.maxrow[outcol]);
  for(k=1; k<=s.maxrow[outcol]; k++) s.number[outcol][k]=1;
  sel_r1 = 0; sel_r2 = s.maxrow[incol];

  //// If new column, initialize it
  s.maxcol = max(s.maxcol, 1+cumcol);
  s.maxrow[cumcol] = max(s.maxrow[incol], s.maxrow[cumcol]);
  for(k=1; k<=s.maxrow[cumcol]; k++) s.number[cumcol][k]=1;


  //// Calculate probability of nn counts when mean is xx
  
  for(row=sel_r1; row<=sel_r2; row++)
  {    s.string[outcol][row] = 0;
       s.number[outcol][row] = 1; 
       s.hex[outcol][row] = 0; 
       n = s.x[incol][row];

       //// lnnfactorial = ln(n!)
       //// ramanujan approximation
       lnnfactorial = n * log(n) - n + log(n*(1+4*n*(1+2*n)))/6 + log(PI)/2;

       //// p = poisson probability
       //// can't use  p = pow(xx, nn) * exp(-xx) / factorial(nn);
       p = exp( n * log(xx) - xx - lnnfactorial );
       s.x[outcol][row] = p;

       //// cumulative probability
       pcum = gammq(n+1, xx);
       s.x[cumcol][row] = pcum;

  }
  refreshspreadsheet();
}


//--------------------------------------------------------------------//
//  poisson2                                                          //
//--------------------------------------------------------------------//
void poisson2(char **arg, int noofargs)
{
  static int expcol=2, obscol=0, outcol=2, cumcol=3;
  static double  xx=1;
  int row;
  char tempstring[128];
  int k;
  int sel_r1=-1, sel_r2=-1;
  double  n, lnnfactorial, p, pcum;

  arg=arg; noofargs=noofargs;
  if(noofargs<=1) 
  {     itoa(s.current_col, tempstring, 10);
        message((char*)"Column no. for observed number of events", tempstring, PROMPT, 1023);
        obscol = atoi(tempstring);
  }else obscol = atoi(arg[1]);

  if(expcol == 0) expcol = obscol + 1;
  if(noofargs<=2) 
  {     sprintf(tempstring, "%d", expcol); 
        message((char*)"Column no. for expected number (e.g. average)", tempstring, PROMPT, 1023);
        expcol = atoi(tempstring);
  }else expcol = atoi(arg[1]+1);

  if(outcol == 0) outcol = expcol + 1;
  if(noofargs<=3) 
  {     sprintf(tempstring, "%d", outcol); 
        message((char*)"Column no. for Poisson results", tempstring, PROMPT, 1023);
        outcol = atoi(tempstring);
  }else outcol = atoi(arg[3]+1);

  if(cumcol == 0) cumcol = outcol + 1;
  if(noofargs<=4) 
  {     sprintf(tempstring, "%d", cumcol); 
        message((char*)"Column no. for cumulative Poisson results", tempstring, PROMPT, 1023);
        cumcol = atoi(tempstring);
  }else cumcol = atoi(arg[4]+1);

  //// If new column, initialize it
  s.maxcol = max(s.maxcol, 1+outcol);
  s.maxrow[outcol] = max(s.maxrow[expcol], s.maxrow[outcol]);
  for(k=1; k<=s.maxrow[outcol]; k++) s.number[outcol][k]=1;

  //// If new column, initialize it
  s.maxcol = max(s.maxcol, 1+cumcol);
  s.maxrow[cumcol] = max(s.maxrow[expcol], s.maxrow[cumcol]);
  for(k=1; k<=s.maxrow[cumcol]; k++) s.number[cumcol][k]=1;

  sel_r1 = 0; sel_r2 = s.maxrow[expcol];

  //// Calculate probability of nn counts when mean is xx
  for(row=sel_r1; row<=sel_r2; row++)
  {    s.string[outcol][row] = 0;
       s.hex[outcol][row] = 1; 
       s.number[outcol][row] = 1; 
       n = s.x[expcol][row];
       xx = s.x[obscol][row];

       //// lnnfactorial = ln(n!)
       //// ramanujan approximation
       lnnfactorial = n * log(n) - n + log(n*(1+4*n*(1+2*n)))/6 + log(PI)/2;

       //// p = poisson probability
       //// can't use  p = pow(xx, nn) * exp(-xx) / factorial(nn);
       p = exp( n * log(xx) - xx - lnnfactorial );
       s.x[outcol][row] = p;

       //// cumulative probability
       pcum = gammq(n+1, xx);
       s.x[cumcol][row] = pcum;

  }
  refreshspreadsheet();
}


//--------------------------------------------------------------------//
//  poisson3                                                          //
//  calculate raw statistic   p(k) = e^m m^x / x!                     //
//--------------------------------------------------------------------//
void poisson3(char **arg, int noofargs)
{
   Dialog dialog;
   static int operation = 1;
   static double  mean=1.0;
   static double  observed=1.0;
   static double  biggest = 143.0; // biggest integer can do factorial
   double  prob;
   double  probless = 0.0;
   double  probmore = 0.0;
  // double  rangemin = 0.0, rangemax = 0.0;
   char tempstring[1024];   
   int j,k;

   dialog.needrepeat = 0;  
   strcpy(dialog.title,"Poisson statistics");
 
   strcpy(dialog.radio[0][0],"Operator"); 
   strcpy(dialog.radio[0][1],"Poisson statistic");
   dialog.radioset[0] = operation;

   strcpy(dialog.boxes[0],"Group 1");
   strcpy(dialog.boxes[1],"Expected Rate");
   strcpy(dialog.boxes[2],"Observed Number of Events (integer)");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=DOUBLECLICKBOX; sprintf(dialog.answer[1], "%g", mean);
   dialog.boxtype[2]=DOUBLECLICKBOX; sprintf(dialog.answer[2], "%g", observed);

   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
   dialog.boxmin[1] = 0.0; 
   dialog.boxmax[1] = biggest;
   dialog.boxmin[2] = 0.0; 
   dialog.boxmax[2] = biggest;
 
   dialog.radiono[0]=2;
   dialog.noofradios=1;
   dialog.noofboxes=3;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////

   if(g.getout)return;
   operation = dialog.radioset[0]; 
   mean = atof(dialog.answer[1]);
   observed = atof(dialog.answer[2]);

   //// probably of exactly k
   prob = exp(-mean) * pow(mean, observed) / (double )factorial(observed);
  
   probless = 0.0;
   for(k=0; k<(int)observed; k++)
         probless += exp(-mean) * pow(mean, k) / (double )factorial(k);
   probmore = 1.0 - probless;

   sprintf(tempstring,"\
   Mean rate         \t%g\n\
   Observed number   \t%g\n\
   Prob. of exactly %g         \t%g\n\
   Prob. of between 0 and <%g  \t%g\n\
   Prob. of between %g and inf \t%g", 
   mean, observed, 
   observed, prob, 
   observed, probless,
   observed, probmore   );
   
   edit((char*)"Poisson results",(char*)"Poisson results",tempstring,8,30,1024,0);

}





//--------------------------------------------------------------------//
//  rowstats                                                          //
//--------------------------------------------------------------------//
void rowstats(void)
{
  int row, col; 
  double  total=0.0, ave=0.0;
  row = s.current_row;
 
  for(col=1; col<=s.maxcol; col++)
      if(s.number[col][row]) total += s.x[col][row];
  ave = total / (double )s.maxcol;
  printf("row %d  total %g  average %g\n",row,total,ave);
}


//--------------------------------------------------------------------------//
//  combinese - combine standard errors of two means + se values            //
//--------------------------------------------------------------------------//
void combinese(void)
{
   Dialog dialog;
   int j,k;
   static char tempstring[1024];
   static double  mean1=0.0, mean2=0.0, mean3, se1=0.0, se2=0.0, se3=0.0, x;
   static int n1=0, n2=0, n=0;
   static int hit = 0;

   dialog.needrepeat = 0;  
   strcpy(dialog.title,"Combine 2 SE values");
   strcpy(dialog.boxes[0],"Group 1");
   strcpy(dialog.boxes[1],"Mean");
   strcpy(dialog.boxes[2],"Std Error.");
   strcpy(dialog.boxes[3],"N");
   strcpy(dialog.boxes[4],"Group 2");
   strcpy(dialog.boxes[5],"Mean");
   strcpy(dialog.boxes[6],"Std Error.");
   strcpy(dialog.boxes[7],"N");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=STRING; if(hit) sprintf(dialog.answer[1], "%g", mean1);
   dialog.boxtype[2]=STRING; if(hit) sprintf(dialog.answer[2], "%g", se1);
   dialog.boxtype[3]=STRING; if(hit) sprintf(dialog.answer[3], "%d", n1);
   dialog.boxtype[4]=LABEL;
   dialog.boxtype[5]=STRING; if(hit) sprintf(dialog.answer[5], "%g", mean2);
   dialog.boxtype[6]=STRING; if(hit) sprintf(dialog.answer[6], "%g", se2);
   dialog.boxtype[7]=STRING; if(hit) sprintf(dialog.answer[7], "%d", n2);

   dialog.radiono[0]=0;
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=0;
   dialog.noofboxes=8;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   mean1 = atof(dialog.answer[1]);
   se1   = atof(dialog.answer[2]);
   n1    = atoi(dialog.answer[3]);
   mean2 = atof(dialog.answer[5]);
   se2   = atof(dialog.answer[6]);
   n2    = atoi(dialog.answer[7]);

   tempstring[0] = 0;
 
   n = n1 + n2;
   mean3 = ((double )n1*mean1 + (double )n2*mean2)/(double )n;
   x = ((double )n1*(double )n2)/(double )n;
   x *= (mean1 - mean2) * (mean1 - mean2);
   x += (double )n2*((double )n2-1)*se2*se2;
   x += (double )n1*((double )n1-1)*se1*se1;
   x *= 1 / ((double )n*((double )n-1));
   se3 = sqrt(x);

   sprintf(tempstring,"\
   mean #1       \t%g\t n #1  \t %d\n\
   mean #2       \t%g\t n #2  \t %d\n\
   SE #1         \t%g\n\
   SE #2         \t%g\n\
   Total n       \t%d\n\
   Combined mean \t%g\n\
   Combined SE   \t%g", mean1,n1,mean2,n2,se1,se2,n,mean3,se3);
   
   edit((char*)"Combined standard error",
        (char*)"Combined standard error",
        tempstring,8,30,1024,0);
   hit = 1;
}


//--------------------------------------------------------------------------//
//  combinese3 - combine standard errors of 3 means + se values             //
//--------------------------------------------------------------------------//
void combinese3(void)
{
   Dialog dialog;
   int j,k;
   static char tempstring[1024];
   static double  mean1=0.0, mean2=0.0, mean3=0.0, mean12, mean123,
                 se1=0.0,   se2=0.0,   se3=0.0,   se12,   se123, x;
   static int n1=0, n2=0, n3=0, n123=0, n12=0;
   static int hit = 0;

   dialog.needrepeat = 0;  
   strcpy(dialog.title,"Combine 3 SE values");
   strcpy(dialog.boxes[0],"Group 1");
   strcpy(dialog.boxes[1],"Mean");
   strcpy(dialog.boxes[2],"Std Error.");
   strcpy(dialog.boxes[3],"N");
   strcpy(dialog.boxes[4],"Group 2");
   strcpy(dialog.boxes[5],"Mean");
   strcpy(dialog.boxes[6],"Std Error.");
   strcpy(dialog.boxes[7],"N");
   strcpy(dialog.boxes[8],"Group 3");
   strcpy(dialog.boxes[9],"Mean");
   strcpy(dialog.boxes[10],"Std Error.");
   strcpy(dialog.boxes[11],"N");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=STRING; if(hit) sprintf(dialog.answer[1], "%g", mean1);
   dialog.boxtype[2]=STRING; if(hit) sprintf(dialog.answer[2], "%g", se1);
   dialog.boxtype[3]=STRING; if(hit) sprintf(dialog.answer[3], "%d", n1);
   dialog.boxtype[4]=LABEL;
   dialog.boxtype[5]=STRING; if(hit) sprintf(dialog.answer[5], "%g", mean2);
   dialog.boxtype[6]=STRING; if(hit) sprintf(dialog.answer[6], "%g", se2);
   dialog.boxtype[7]=STRING; if(hit) sprintf(dialog.answer[7], "%d", n2);
   dialog.boxtype[8]=LABEL;
   dialog.boxtype[9]=STRING;  if(hit) sprintf(dialog.answer[9],  "%g", mean3);
   dialog.boxtype[10]=STRING; if(hit) sprintf(dialog.answer[10], "%g", se3);
   dialog.boxtype[11]=STRING; if(hit) sprintf(dialog.answer[11], "%d", n3);

   dialog.radiono[0]=0;
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=0;
   dialog.noofboxes=12;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   mean1 = atof(dialog.answer[1]);
   se1   = atof(dialog.answer[2]);
   n1    = atoi(dialog.answer[3]);
   mean2 = atof(dialog.answer[5]);
   se2   = atof(dialog.answer[6]);
   n2    = atoi(dialog.answer[7]);
   mean3 = atof(dialog.answer[9]);
   se3   = atof(dialog.answer[10]);
   n3    = atoi(dialog.answer[11]);

   tempstring[0] = 0;
 
   /// combine 1 and 2 to give 12
   n12 = n1 + n2;
   mean12 = ((double )n1*mean1 + (double )n2*mean2)/(double )n12;

   x = ((double )n1*(double )n2)/(double )n12;
   x *= (mean1 - mean2) * (mean1 - mean2);
   x += (double )n2*((double )n2-1.0)*se2*se2;
   x += (double )n1*((double )n1-1.0)*se1*se1;
   x *= 1.0 / ((double )n12*((double )n12-1.0));
   se12 = sqrt(x);

   /// combine 12 and 3 to give 123
   n123 = n1 + n2 + n3;
   mean123 = ((double )n1*mean1 + (double )n2*mean2 + (double )n3*mean3)/(double )n123;

   x = ((double )n12*(double )n3)/(double )n123;
   x *= (mean12 - mean3) * (mean12 - mean3);
   x += (double )n12*((double )n12-1.0)*se12*se12;
   x += (double )n3 *((double )n3 -1.0)*se3 *se3;
   x *= 1.0 / ((double )n123*((double )n123-1.0));
   se123 = sqrt(x);


   sprintf(tempstring,"\
   mean #1       \t%g\t n #1  \t %d\n\
   mean #2       \t%g\t n #2  \t %d\n\
   mean #3       \t%g\t n #3  \t %d\n\
   SE #1         \t%g\n\
   SE #2         \t%g\n\
   SE #3         \t%g\n\
   Total n       \t%d\n\
   Combined mean \t%g\n\
   Combined SE   \t%g", mean1,n1,mean2,n2,mean3,n3,se1,se2,se3,n123,mean123,se123);
   
   edit((char*)"Combined standard error",
        (char*)"Combined standard error",
        tempstring,8,30,1024,0);
   hit = 1;
}


//--------------------------------------------------------------------------//
//  error_propagation                                                       //
//--------------------------------------------------------------------------//
void error_propagation(void)
{
   Dialog dialog;
   int j,k;
   static int set1=1,set2=2,set3=3,set4=4,set5=5,set6=6, operation=1;
   int col1,col2,col3,col4,col5,col6;
   enum { NUL, ADD, SUB, MUL, DIV };
   double  x,y,ex,ey;

   strcpy(dialog.radio[0][0],"Operator"); 
   strcpy(dialog.radio[0][1],"Add");
   strcpy(dialog.radio[0][2],"Subtract");             
   strcpy(dialog.radio[0][3],"Multiply");             
   strcpy(dialog.radio[0][4],"Divide");             
   dialog.radiono[0]=5;
   dialog.radioset[0] = operation;

   strcpy(dialog.title,"Error propagation operation on two sets");
   strcpy(dialog.boxes[0],"Data sets");
   strcpy(dialog.boxes[1],"Data set 1");
   strcpy(dialog.boxes[2],"Error set 1");
   strcpy(dialog.boxes[3],"Data set 2");
   strcpy(dialog.boxes[4],"Error set 2");
   strcpy(dialog.boxes[5],"Result data set");
   strcpy(dialog.boxes[6],"Result error set");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=INTCLICKBOX;
           sprintf(dialog.answer[1],"%d",set1);
           dialog.boxmin[1]=1; 
           dialog.boxmax[1]=s.cols;
   dialog.boxtype[2]=INTCLICKBOX;
           sprintf(dialog.answer[2],"%d",set2);
           dialog.boxmin[2]=1; 
           dialog.boxmax[2]=s.cols;
   dialog.boxtype[3]=INTCLICKBOX;
           sprintf(dialog.answer[3],"%d",set3);
           dialog.boxmin[3]=1; 
           dialog.boxmax[3]=s.cols;
   dialog.boxtype[4]=INTCLICKBOX;
           sprintf(dialog.answer[4],"%d",set4);
           dialog.boxmin[4]=1; 
           dialog.boxmax[4]=s.cols;
   dialog.boxtype[5]=INTCLICKBOX;
           sprintf(dialog.answer[5],"%d",set5);
           dialog.boxmin[5]=1; 
           dialog.boxmax[5]=s.cols;
   dialog.boxtype[6]=INTCLICKBOX;
           sprintf(dialog.answer[6],"%d",set6);
           dialog.boxmin[6]=1; 
           dialog.boxmax[6]=s.cols;
   
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=1;
   dialog.noofboxes=7;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   set1  = atoi(dialog.answer[1]);
   set2  = atoi(dialog.answer[2]);
   set3  = atoi(dialog.answer[3]);
   set4  = atoi(dialog.answer[4]);
   set5  = atoi(dialog.answer[5]);
   set6  = atoi(dialog.answer[6]);
   operation = dialog.radioset[0]; 

   col1 = set1;
   col2 = set2;
   col3 = set3;
   col4 = set4;
   col5 = set5;
   col6 = set6;

  //// If new column, initialize it
  s.maxcol = max(s.maxcol, col6+1);
  s.maxrow[col5] = s.maxrow[col1];
  s.maxrow[col6] = s.maxrow[col1];

  for(k=0; k<s.maxrow[col5]; k++)
  {    s.x[col5][k] = 0;
       s.x[col6][k] = 0;
       s.string[col5][k] = 0;
       s.number[col5][k] = 1;
       s.hex[col5][k] = 0;
       s.string[col6][k] = 0;
       s.number[col6][k] = 1;
       s.hex[col6][k] = 0;
       strcpy(s.text[col5][k], "0");
       strcpy(s.disp[col5][k], "0");
       strcpy(s.text[col6][k], "0");
       strcpy(s.disp[col6][k], "0");
  }

  for(k=0; k<s.maxrow[col5]; k++)
  {
       x = s.x[col1][k];
       y = s.x[col3][k];
       ex = s.x[col2][k];
       ey = s.x[col4][k];     
       switch(operation)
       {    case ADD:
                s.x[col5][k] = x + y;
                s.x[col6][k] = sqrt(ex*ex + ey*ey);
                break;
            case SUB:
                s.x[col5][k] = x - y;
                s.x[col6][k] = sqrt(ex*ex + ey*ey);
                break;
            case MUL:
                s.x[col5][k] = x * y;
                s.x[col6][k] = x * y * sqrt( (ex*ex)/(x*x) + (ey*ey)/(y*y) );
                break;
            case DIV:
                if(y==0) { message((char*)"Division by zero",ERROR); return;} 
                s.x[col5][k] = x / y;
                s.x[col6][k] = (x / y) * sqrt( (ex*ex)/(x*x) + (ey*ey)/(y*y) );
                break;
       }
  
  }
  setcoltitle(col5, (char*)"Result");
  setcoltitle(col6, (char*)"Error");
  refreshspreadsheet();
}

//--------------------------------------------------------------------------//
//  errorpropagationfrommean                                                //
//  can be sd or sem                                                        //
//--------------------------------------------------------------------------//
void errorpropagationfrommean(void)
{
   Dialog dialog;
   static double  mean1=0.0, mean2=0.0, mean3=0.0, sd1=0.0, sd2=0.0, sd3=0.0;
   static int hit=0;
   static int op=1;
   enum { NUL, ADD, SUB, MUL, DIV };
   double  a,b;
   static char tempstring[1024];
   char opstring[64];
   int j,k;

   strcpy(dialog.radio[0][0],"Operator"); 
   strcpy(dialog.radio[0][1],"Add");
   strcpy(dialog.radio[0][2],"Subtract");             
   strcpy(dialog.radio[0][3],"Multiply");             
   strcpy(dialog.radio[0][4],"Divide");             
   dialog.radiono[0]=5;
   dialog.radioset[0] = op;

   strcpy(dialog.title,"Error propagation from mean and SD or SEM"); 
   strcpy(dialog.boxes[0],"Add");
   strcpy(dialog.title,"Error propagation from mean and SD or SEM");
   strcpy(dialog.boxes[0],"Group 1 (x)");
   strcpy(dialog.boxes[1],"Mean x");
   strcpy(dialog.boxes[2],"Error x");
   strcpy(dialog.boxes[3],"Group 2 (y)");
   strcpy(dialog.boxes[4],"Mean y");
   strcpy(dialog.boxes[5],"Error y");

   strcpy(dialog.answer[1],"0");
   strcpy(dialog.answer[2],"0");
   strcpy(dialog.answer[4],"0");
   strcpy(dialog.answer[5],"0");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=STRING; if(hit) sprintf(dialog.answer[1], "%g", mean1);
   dialog.boxtype[2]=STRING; if(hit) sprintf(dialog.answer[2], "%g", sd1);
   dialog.boxtype[3]=LABEL;
   dialog.boxtype[4]=STRING; if(hit) sprintf(dialog.answer[4], "%g", mean2);
   dialog.boxtype[5]=STRING; if(hit) sprintf(dialog.answer[5], "%g", sd2);

   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=1;
   dialog.noofboxes=6;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   mean1 = atof(dialog.answer[1]);
   sd1   = atof(dialog.answer[2]);
   mean2 = atof(dialog.answer[4]);
   sd2   = atof(dialog.answer[5]);
   op = dialog.radioset[0];

   tempstring[0] = 0;
   opstring[0] = 0;
   switch(op)
   {        case ADD:
                strcpy(opstring, "x+y");
                mean3 = mean1+mean2;
                sd3 = sqrt(sd1*sd1 + sd2*sd2);
                break;
            case SUB:
                strcpy(opstring, "x-y");
                mean3 = mean1-mean2;
                sd3 = sqrt(sd1*sd1 + sd2*sd2);
                break;
            case MUL:
                strcpy(opstring, "x*y");
                mean3 = mean1*mean2;
                sd3 = mean3 * sqrt((sd1*sd1)/(mean1*mean1) + (sd2*sd2)/(mean2*mean2));
                break;
            case DIV:
                strcpy(opstring, "x/y");
                if(mean2!=0.0) mean3 = mean1/mean2;
                else { message((char*)"Division by zero",ERROR); return;} 
                a = (sd1 * sd1) / (mean2 * mean2);
                b = (mean1 / (mean2 * mean2)) * (mean1 / (mean2 * mean2)) * sd2 * sd2;
                sd3 = sqrt(a + b);
                break;
   }
   sprintf(tempstring, "\
   x                    \t%g\n\
   Error x              \t%g\n\n\
   y                    \t%g\n\
   Error y              \t%g\n\n\
   %s                   \t%g\n\
   Error %s \t%g", mean1,sd1,mean2,sd2,opstring,mean3,opstring,sd3);
   edit((char*)"Error propagation results",(char*)"Error propagation results",tempstring,8,30,1024,0);
   hit = 1;
}


//--------------------------------------------------------------------------//
//  errorpropagationratio                                                   //
//  (x - a) / ((x - a) + (y - b))                                           //
//  where each variable has a standard deviation (can be SEM)               //
//--------------------------------------------------------------------------//
void errorpropagationratio(void)
{
   Dialog dialog;
   static double  mean1=0.0, mean2=0.0, mean3=0.0, mean4=0.0;
   static double  sd1=0.0,   sd2=0.0,   sd3=0.0,   sd4=0.0;
   static int hit=0;
   enum { NUL, ADD, SUB, MUL, DIV };
   int j;

   strcpy(dialog.title,"Error propagation of ratio of two numbers from mean and  SD or SEM"); 
   strcpy(dialog.boxes[0],"Add");
   strcpy(dialog.title,"Error propagation of ratio of two numbers  from mean and SD or SEM");

   strcpy(dialog.boxes[0],"x Raw value");
   strcpy(dialog.boxes[1],"x Mean");
   strcpy(dialog.boxes[2],"x Error");

   strcpy(dialog.boxes[3],"x Blank");
   strcpy(dialog.boxes[4],"x Blank Mean");
   strcpy(dialog.boxes[5],"x Blank Std Dev.");

   strcpy(dialog.boxes[6],"y Raw value");
   strcpy(dialog.boxes[7],"y Mean");
   strcpy(dialog.boxes[8],"y Error.");

   strcpy(dialog.boxes[9], "y Blank");
   strcpy(dialog.boxes[10],"y Blank Mean");
   strcpy(dialog.boxes[11],"y Blank Error");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=STRING; if(hit) sprintf(dialog.answer[1], "%g", mean1);
   dialog.boxtype[2]=STRING; if(hit) sprintf(dialog.answer[2], "%g", sd1);

   dialog.boxtype[3]=LABEL;
   dialog.boxtype[4]=STRING; if(hit) sprintf(dialog.answer[4], "%g", mean2);
   dialog.boxtype[5]=STRING; if(hit) sprintf(dialog.answer[5], "%g", sd2);

   dialog.boxtype[6]=LABEL;
   dialog.boxtype[7]=STRING; if(hit) sprintf(dialog.answer[7], "%g", mean3);
   dialog.boxtype[8]=STRING; if(hit) sprintf(dialog.answer[8], "%g", sd3);

   dialog.boxtype[9]=LABEL;
   dialog.boxtype[10]=STRING; if(hit) sprintf(dialog.answer[10], "%g", mean4);
   dialog.boxtype[11]=STRING; if(hit) sprintf(dialog.answer[11], "%g", sd4);

   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=0;
   dialog.noofboxes=12;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   mean1 = atof(dialog.answer[1]);
   sd1   = atof(dialog.answer[2]);
   mean2 = atof(dialog.answer[4]);
   sd2   = atof(dialog.answer[5]);
   mean3 = atof(dialog.answer[7]);
   sd3   = atof(dialog.answer[8]);
   mean4 = atof(dialog.answer[10]);
   sd4   = atof(dialog.answer[11]);
   errorpropagation_showresults(mean1,mean2,mean3,mean4,sd1,sd2,sd3,sd4);
}


//--------------------------------------------------------------------------//
//  errorpropagation_showresults                                            //
//  1 = membrane                                                            //
//  2 = membrane sd                                                         //
//  3 = cytosol                                                             //
//  4 = cytosol sd                                                          //
//--------------------------------------------------------------------------//
void errorpropagation_showresults(double  mean1, double  mean2, double  mean3,
     double  mean4, double  sd1, double  sd2, double  sd3, double  sd4)
{
   double  mean1minus2, mean3minus4, sd1minus2, sd3minus4;
   double  total, ratio, sdtotal, sdratio;
   int gotnegative = 0;
   char note[256];
   enum { NUL, ADD, SUB, MUL, DIV };
   double  a,b;
   static char tempstring[4*1024];
   char opstring[64];
   tempstring[0] = 0;
   opstring[0] = 0;
   strcpy(opstring, "(x1-x2)/((x1-x2)+(x3-x4))");
 
   //// membrane - blank
   mean1minus2 = mean1-mean2;
   sd1minus2 = sqrt(sd1*sd1 + sd2*sd2);
   if(mean1minus2 < 0) { gotnegative++; mean1minus2=0; }

   //// cytosol - blank
   mean3minus4 = mean3-mean4;
   sd3minus4 = sqrt(sd3*sd3 + sd4*sd4);
   if(mean3minus4 < 0) { gotnegative++; mean3minus4=0; }

   //// total
   total = mean1minus2 + mean3minus4;
   sdtotal = sqrt(sd1minus2*sd1minus2 + sd3minus4*sd3minus4);

   //// membrane / total
   if(total != 0.0)
        ratio = mean1minus2 / total;
   else{ message((char*)"Division by zero",ERROR); return;} 

   a = (sd1minus2 * sd1minus2) / (total * total);
   b = (mean1minus2 / (total * total)) * (mean1minus2 / (total * total)) * sdtotal * sdtotal;
   sdratio = sqrt(a + b);

   if(gotnegative)
      sprintf(note, "Warning: %d negative numbers truncated to 0", gotnegative);
   else
      strcpy(note, "OK");

   sprintf(tempstring, "\
   Filename                     \t%s\n\
   Membrane Raw                 \t%g\n\
   Membrane Raw SD              \t%g\n\n\
   Membrane Blank               \t%g\n\
   Membrane SD                  \t%g\n\n\
   Membrane - Blank             \t%g\n\
   Membrane - Blank SD          \t%g\n\n\
   Cytosol Raw                  \t%g\n\
   Cytosol Raw SD               \t%g\n\n\
   Cytosol Blank                \t%g\n\
   Cytosol Blank SD             \t%g\n\n\
   Cytosol - Cytosol Blank      \t%g\n\
   Cytosol - Cytosol Blank SD   \t%g\n\n\
   Note:                        \t%s\n\n\
   Total                        \t%g\n\
   Total SD                     \t%g\n\n\
   Membrane / Total             \t%g\n\
   Membrane / Total SD          \t%g\n\n", 
   g.filename,
   mean1, sd1,
   mean2, sd2,
   mean1minus2, sd1minus2,
   mean3, sd3,
   mean4, sd4,
   mean3minus4, sd3minus4,
   note,
   total, sdtotal,
   ratio, sdratio);

   edit((char*)"Error propagation results",(char*)"Error propagation results",
      tempstring,29,30,1024,0);
}


//--------------------------------------------------------------------------//
//  error_propagationsetvalue                                               //
//  column +- sd versus a single number +- sd                               //
//--------------------------------------------------------------------------//
void error_propagationsetvalue(void)
{
   Dialog dialog;
   int j,k;
   static int set1=1,set2=2,set5=5,set6=6, operation=1;
   int col1,col2,col5,col6;
   enum { NUL, ADD, SUB, MUL, DIV };
   double  x,y,ex,ey;
   static double  value=0.0, valuesd=0.0;

   strcpy(dialog.radio[0][0],"Operator"); 
   strcpy(dialog.radio[0][1],"Add");
   strcpy(dialog.radio[0][2],"Subtract");             
   strcpy(dialog.radio[0][3],"Multiply");             
   strcpy(dialog.radio[0][4],"Divide");             
   dialog.radiono[0]=5;
   dialog.radioset[0] = operation;

   strcpy(dialog.title,"Error propagation operation on one set and a value");
   strcpy(dialog.boxes[0],"Data sets");
   strcpy(dialog.boxes[1],"Data set");
   strcpy(dialog.boxes[2],"Error set 1");
   strcpy(dialog.boxes[3],"Value     ");
   strcpy(dialog.boxes[4],"Error of value");
   strcpy(dialog.boxes[5],"Result data set");
   strcpy(dialog.boxes[6],"Result error set");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=INTCLICKBOX;
           sprintf(dialog.answer[1],"%d",set1);
           dialog.boxmin[1]=1; 
           dialog.boxmax[1]=s.cols;
   dialog.boxtype[2]=INTCLICKBOX;
           sprintf(dialog.answer[2],"%d",set2);
           dialog.boxmin[2]=1; 
           dialog.boxmax[2]=s.cols;
   dialog.boxtype[3]=STRING;
           sprintf(dialog.answer[3],"%g",value);
   dialog.boxtype[4]=STRING;
           sprintf(dialog.answer[4],"%g",valuesd);
   dialog.boxtype[5]=INTCLICKBOX;
           sprintf(dialog.answer[5],"%d",set5);
           dialog.boxmin[5]=1; 
           dialog.boxmax[5]=s.cols;
   dialog.boxtype[6]=INTCLICKBOX;
           sprintf(dialog.answer[6],"%d",set6);
           dialog.boxmin[6]=1; 
           dialog.boxmax[6]=s.cols;
   
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=1;
   dialog.noofboxes=7;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   set1  = atoi(dialog.answer[1]);
   set2  = atoi(dialog.answer[2]);
   value = atof(dialog.answer[3]);
   valuesd= atof(dialog.answer[4]);
   set5  = atoi(dialog.answer[5]);
   set6  = atoi(dialog.answer[6]);
   operation = dialog.radioset[0]; 

   col1 = set1;
   col2 = set2;
   col5 = set5;
   col6 = set6;

  //// If new column, initialize it
  s.maxcol = max(s.maxcol, col6+1);
  s.maxrow[col5] = s.maxrow[col1];
  s.maxrow[col6] = s.maxrow[col1];

  for(k=0; k<s.maxrow[col5]; k++)
  {    s.x[col5][k] = 0;
       s.x[col6][k] = 0;
       s.string[col5][k] = 0;
       s.number[col5][k] = 1;
       s.hex[col5][k] = 0;
       s.string[col6][k] = 0;
       s.number[col6][k] = 1;
       s.hex[col6][k] = 1;
       strcpy(s.text[col5][k], "0");
       strcpy(s.disp[col5][k], "0");
       strcpy(s.text[col6][k], "0");
       strcpy(s.disp[col6][k], "0");
  }

  for(k=0; k<s.maxrow[col5]; k++)
  {
       x = s.x[col1][k];
       y = value;
       ex = s.x[col2][k];
       ey = valuesd;     
       switch(operation)
       {    case ADD:
                s.x[col5][k] = x + y;
                s.x[col6][k] = sqrt(ex*ex + ey*ey);
                break;
            case SUB:
                s.x[col5][k] = x - y;
                s.x[col6][k] = sqrt(ex*ex + ey*ey);
                break;
            case MUL:
                s.x[col5][k] = x * y;
                s.x[col6][k] = x * y * sqrt( (ex*ex)/(x*x) + (ey*ey)/(y*y) );
                break;
            case DIV:
                if(y==0) { message((char*)"Division by zero",ERROR); return;} 
                s.x[col5][k] = x / y;
                s.x[col6][k] = (x / y) * sqrt( (ex*ex)/(x*x) + (ey*ey)/(y*y) );
                break;
       }
  
  }
  setcoltitle(col5, (char*)"Result");
  setcoltitle(col6, (char*)"Error");
  refreshspreadsheet();
}

//--------------------------------------------------------------------------//
//  errorsets                                                               //
//  mean, sd of one set with second set representing error term             //
//--------------------------------------------------------------------------//
void errorsets(void)
{
   Dialog dialog;
   char tempstring[1024];
   char sss[512];
   int j,k,n;
   static int set1=1,set2=2,count=1;
   int col1,col2;
   enum { SEM, SD };
   static int datatype=SD;
   double  xtot=0.0, extot=0.0;
   double  mean, sd, sem, ex, variance;
   double  x,nsd=0.0,nse=0.0,esd=0.0,ese=0.0,adev=0.0,xx=0.0,p=0.0;

   strcpy(dialog.radio[0][0],"Type of data"); 
   strcpy(dialog.radio[0][1],"Number +/- SEM");
   strcpy(dialog.radio[0][2],"Number +/- SD");             
   dialog.radiono[0]=3;
   dialog.radioset[0] = 1 + datatype;

   strcpy(dialog.title,"Statistics on one column");
   strcpy(dialog.boxes[0],"Data sets");
   strcpy(dialog.boxes[1],"Column for data");
   strcpy(dialog.boxes[2],"Column for error term");
   strcpy(dialog.boxes[3],"N (if error = SEM)");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=INTCLICKBOX;
           sprintf(dialog.answer[1],"%d",set1);
           dialog.boxmin[1]=1; 
           dialog.boxmax[1]=s.cols;
   dialog.boxtype[2]=INTCLICKBOX;
           sprintf(dialog.answer[2],"%d",set2);
           dialog.boxmin[2]=1; 
           dialog.boxmax[2]=s.cols;
   dialog.boxtype[3]=INTCLICKBOX;
           sprintf(dialog.answer[3],"%d",count);
           dialog.boxmin[3]=1; 
           dialog.boxmax[3]=100;
   
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=1;
   dialog.noofboxes=4;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   set1  = atoi(dialog.answer[1]);
   set2  = atoi(dialog.answer[2]);
   count = atoi(dialog.answer[3]);
   datatype = dialog.radioset[0] - 1; 

   col1 = set1;
   col2 = set2;
   n = s.maxrow[col1];
   if(n==0) return;
   
   for(k=1; k<=n; k++)
   {
       xtot += s.x[col1][k];
       ex = s.x[col2][k]; 
       if(datatype==SEM) ex *= sqrt(count); // convert to sd
       extot += ex * ex;
   }

   mean = xtot / (double )n;
   esd = sqrt(extot);                // combined sd terms
   ese = esd / sqrt((double )n);      // combined sem terms
   variance = 0.0;

   for(k=1; k<=n; k++)
   { 
        x = s.x[col1][k];
        adev += fabs((xx=x-mean));          
        variance += (p=xx*xx);
   }

   adev /= (double )n;
   variance /= ((double )n-1.0);
   nsd = sqrt(variance);
   nse = nsd / sqrt((double )n);

   sd  = sqrt(nsd*nsd + esd*esd);
   sem = sqrt(nse*nse + ese*ese);

   if(datatype==SEM)strcpy(sss,"Std.Error");
   if(datatype==SD) strcpy(sss,"Std.Deviation");
   
   sprintf(tempstring,
   "    Column for data        \t%d\n\
    Column for error           \t%d\n\
    Data type for error        \t%s\n\
    Number of points           \t%d\n\n\
    Mean                       \t%g\n\
    SD                         \t%g\n\
    SEM                        \t%g",
         set1, set2, sss, n, mean, sd, sem);  
   edit((char*)"Combined statistics on column",(char*)"Combined statistics on column",tempstring,8,30,1024,0);   
}


//--------------------------------------------------------------------------//
//  pkc                                                                     //
//  analyze pkc data from elisa plates                                      //
//  col 1 must be cytosol blanks                                            //
//  col 2 must be membrane blanks                                           //
//  col 3 must be cytosol pkc                                               //
//  col 4 must be membrane pkc                                              //
//  result = membrane / total +- sd                                         //
//--------------------------------------------------------------------------//
void pkc(void)
{
  double  mean1, sum1, sd1, se1, variance1;
  double  mean2, sum2, sd2, se2, variance2;
  double  mean3, sum3, sd3, se3, variance3;
  double  mean4, sum4, sd4, se4, variance4;
  statistics2(1, 0, mean1, sum1, sd1, se1, variance1);
  statistics2(2, 0, mean2, sum2, sd2, se2, variance2);
  statistics2(3, 0, mean3, sum3, sd3, se3, variance3);
  statistics2(4, 0, mean4, sum4, sd4, se4, variance4);
  errorpropagation_showresults(mean4,mean2,mean3,mean1,sd4,sd2,sd3,sd1);
}

//--------------------------------------------------------------------------//
//  areaundercurve                                                          //
//--------------------------------------------------------------------------//
void areaundercurve(void)
{
   Dialog dialog;
   int j,k,n;
   static int col1=1,col2=2;
   double  area = 0.0, total = 0.0;
   double  x1,y1,x2,y2,ymin,ymax;
   char tempstring[1000];
   dialog.radiono[0]=0;

   strcpy(dialog.title,"Area under curve");
   strcpy(dialog.boxes[0],"Data sets");
   strcpy(dialog.boxes[1],"Data set for x");
   strcpy(dialog.boxes[2],"Data set for y");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=INTCLICKBOX;
           sprintf(dialog.answer[1],"%d",col1);
           dialog.boxmin[1]=1; 
           dialog.boxmax[1]=s.cols;
   dialog.boxtype[2]=INTCLICKBOX;
           sprintf(dialog.answer[2],"%d",col2);
           dialog.boxmin[2]=1; 
           dialog.boxmax[2]=s.cols;
   
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=0;
   dialog.noofboxes=3;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   col1  = atoi(dialog.answer[1]);
   col2  = atoi(dialog.answer[2]);
   n = s.maxrow[col1];

   area = 0.0;
   for(k=0; k<n-1; k++)
   {   total = 0.0;
       x1 = s.x[col1][k];
       x2 = s.x[col1][k+1];
       y1 = s.x[col2][k];
       y2 = s.x[col2][k+1];
       ymin = min(y1, y2);
       ymax = max(y1, y2);
       total += ymin * (x2-x1);  // rectangle portion
       total += (ymax - ymin) * (x2-x1) / 2;
       area += total;
   }
   sprintf(tempstring, "AUC = %g", area);
   message(tempstring);
}

//--------------------------------------------------------------------//
//  normalizemax                                                      //
//--------------------------------------------------------------------//
void normalizemax(char **arg, int noofargs)
{
   int col,k,n,errcol;
   char tempstring[128];
   double  x1, highest=0.0, factor=1.0;
   arg=arg; noofargs=noofargs;
   itoa(s.current_col, tempstring, 10);
   if(noofargs<2) 
   {     message((char*)"Column no.", tempstring, PROMPT, 1023);
         col = atoi(tempstring);
   }else col = atoi(arg[1]);
   itoa(col + 2, tempstring, 10);
   if(noofargs<3) 
   {     message((char*)"Column no. for error", tempstring, PROMPT, 1023);
         errcol = atoi(tempstring);
   }else errcol = atoi(arg[2]);

   n = s.maxrow[col];
   highest = s.x[col][0];
   for(k=0; k<n; k++)
   {   x1 = s.x[col][k];
       highest = max(highest, x1);
   }
   if(highest==0.0){ message((char*)"Highest number is 0"); return;}
   if(highest<0.0){ message((char*)"All numbers are less than 0"); return;}
   factor = 100.0 / highest;
   for(k=0; k<n; k++)
   {   x1 = factor * s.x[col][k];
       s.x[col][k] = x1;
       x1 = factor * s.x[errcol][k];
       s.x[errcol][k] = x1;
   }
   refreshspreadsheet();
}

//--------------------------------------------------------------------//
//  normalize                                                         //
//  normalize single column to a set maximum value                    //
//--------------------------------------------------------------------//
void normalize(char **arg, int noofargs)
{
   int col,k,n;
   char tempstring[128];
   double  x1, highest=0.0, factor=1.0, refval=0.0;
   arg=arg; noofargs=noofargs;
   itoa(s.current_col, tempstring, 10);
   if(noofargs<2) 
   {     message((char*)"Column no.", tempstring, PROMPT, 1023);
         col = atoi(tempstring);
   }else col = atoi(arg[1]);
   itoa(col + 2, tempstring, 10);
   if(noofargs<3) 
   {     message((char*)"Reference value to maximize to", tempstring, PROMPT, 1023);
         refval = atof(tempstring);
   }else refval = atof(arg[2]);

   n = s.maxrow[col];
   highest = s.x[col][0];
   for(k=1; k<=n; k++)
   {   x1 = s.x[col][k];
       highest = max(highest, x1);
   }
   if(highest==0.0){ message((char*)"Highest number is 0"); return;}
   if(highest<0.0){ message((char*)"All numbers are less than 0"); return;}
   factor = refval / highest;
   for(k=1; k<=n; k++)
   {   x1 = factor * s.x[col][k];
       s.x[col][k] = x1;
   }
   refreshspreadsheet();
}


//--------------------------------------------------------------------//
//  normalize2                                                        //
//  normalize 2 columns to a set maximum value in one column          //
//  useful if one column is numbers and other is a SD.                //
//--------------------------------------------------------------------//
void normalize2(char **arg, int noofargs)
{
/*
   int col,k,n,errcol;
   char tempstring[128];
   double  x1, highest=0.0, factor=1.0, refval=0.0;
   arg=arg; noofargs=noofargs;
   itoa(s.current_col, tempstring, 10);
   if(noofargs<2) 
   {     message((char*)"Column no.", tempstring, PROMPT, 1023);
         col = atoi(tempstring);
   }else col = atoi(arg[1]);
   itoa(col + 2, tempstring, 10);
   if(noofargs<3) 
   {     message((char*)"Column no. for error", tempstring, PROMPT, 1023);
         errcol = atoi(tempstring);
   }else errcol = atoi(arg[2]);
   if(noofargs<4) 
   {     message((char*)"Reference value", tempstring, PROMPT, 1023);
         refval = atof(tempstring);
   }else refval = atof(arg[2]);

   n = s.maxrow[col];
   highest = s.x[col][0];
   for(k=0; k<n; k++)
   {   x1 = s.x[col][k];
       highest = max(highest, x1);
   }
   if(highest==0.0){ message((char*)"Highest number is 0"); return;}
   if(highest<0.0){ message((char*)"All numbers are less than 0"); return;}
   factor = 100.0 / refval;
   for(k=0; k<n; k++)
   {   x1 = factor * s.x[col][k];
       s.x[col][k] = x1;
       x1 = factor * s.x[errcol][k];
       s.x[errcol][k] = x1;
   }
   refreshspreadsheet();
*/
}


//--------------------------------------------------------------------------//
//  anova simple 1-way anova                                                //
//--------------------------------------------------------------------------//
void anova(void)
{
   double  ave[100], var[100];
   double  total=0.0, mean=0.0, sb=0.0, msb=0.0, sw=0.0, msw=0.0;
   double  ff;           // f ratio statistic
   double  prob;         // Probability of a greater t
   int fb,fw;           // degrees of freedom between groups
   int nsets=0;         // no of sets
   static char tempstring[1024];
   int j,k, datacount=0, setcount=0;

                              // count the no of sets first, must be at least 3
                              // _then_ declare the arrays
   for(k=0; k<=s.maxcol; k++) if(s.column_selected[k+1]) nsets++;
   if(nsets <=1){ message((char*)"Please select two or more columns",ERROR); return;}

   double  *data[nsets];      // array of data
   int n[nsets];             // which column numbers

   for(k=0; k<=s.maxcol; k++) n[k] = 0;
   for(k=0; k<=s.maxcol; k++)
   {
       if(s.column_selected[k+1])
       {  
          n[setcount++]=k+1; 
       }
   }

   for(k=0; k<nsets; k++)
   {  data[k] = new double [s.maxrow[n[k]]];  if(data[k]==NULL) return; 
   }
   for(k=0; k<nsets; k++)
   {  for(j=0; j<s.maxrow[n[k]]; j++)
      {      data[k][j] = s.x[n[k]][j+1];  //// j+1 because columns start at 1
             datacount++;
      }
   }
  
   for(k=0; k<nsets; k++)
   {  avevar(data[k], s.maxrow[n[k]], &ave[k], &var[k]);
   }

   //// overall mean
   for(k=0; k<nsets; k++) total += ave[k];
   mean = total / nsets;

   //// Between group sum of squares
   for(k=0; k<nsets; k++)
      sb += s.maxrow[n[k]] * (ave[k] - mean) * (ave[k] - mean);
   
   //// fb Between-group degrees of freedom
   fb = nsets - 1;
   //// Between-group mean square
   msb = sb / (double )fb;

   //// Center the data in each group
   for(k=0; k<nsets; k++)
        for(j=0; j<s.maxrow[n[k]]; j++)
            data[k][j] -= ave[k];

   //// Within-group 
   for(k=0; k<nsets; k++)
        for(j=0; j<s.maxrow[n[k]]; j++)
            sw += data[k][j] * data[k][j];

   //// fw within-group deg of freedom
   fw = 0;
   for(k=0; k<nsets; k++)
        fw += s.maxrow[n[k]] - 1;

   //// within-group mean square
   msw = sw / fw;
   ff = msb / msw;
   prob = betai(0.5*fw, 0.5*fb, fw/(fw + fb*ff));
   if(prob > 1.0) prob = 2.0 - prob;
   
   sprintf(tempstring,"\
   number of data sets        \t%d\n\
   number of data points      \t%d\n\
   mean                       \t%g\n\
   between group deg. freedom \t%d\n\
   between group sum squares  \t%g\n\
   within group deg. freedom  \t%d\n\
   within group sum squares   \t%g\n\
   f ratio                    \t%g\n\
   probability                \t%g", nsets,datacount,mean,fb,msb,fw,msw,ff,prob);  

   edit((char*)"1-way anova",(char*)"1-way anova",tempstring,8,30,1024,0);

   for(k=0; k<nsets; k++)
       delete[] data[k];  
}

//--------------------------------------------------------------------------//
//  mann whitney u-test                                                     //
//  always uses selected data                                               //
//--------------------------------------------------------------------------//
void mannwhitney(void)
{
/*
test data
https://epilab.ich.ucl.ac.uk/coursematerial/statistics/non_parametric/wilcox_mann_whitney.html
group 1            group 1
 -5.5              -4.5             
 -3.2              -4.2             
 -3.2              -0.7             
 -2.4              -0.6             
 -1.4              -0.3             
 -0.9              -0.1             
 -0.7              0.5              
 -0.6              1.4              
 -0.3              2                
 -0.2              4.7              
 -0.1              0                
  0.1              0    
 
after 2nd quicksort
1+rank[k] data[rank[k]] fixed
1          -5.5          1  
4          -3.2          4.5  
5          -3.2          4.5  
6          -2.4          6  
7          -1.4          7  
8          -0.9          8  
10         -0.7          9.5
11         -0.6          11.5   
14         -0.3          13.5
15         -0.2          15 
17         -0.1          16.5
18          0.1          18 
2          -4.5          2  
3          -4.2          3  
9          -0.7          9.5
12         -0.6          11.5
13         -0.3          13.5
16         -0.1          16.5 
19          0.5          19 
20          1.4          20 
21          2            21 
22          4.7          22 

uses fuzzy p value, 10000 iterations to correct for ties
geyer cj j stat software
www.stat.umn.edu/geyer/fuzz/ties.pdf 
rtot0 115 rtot1 138 u0 37 u1 83

                          Exact             Corrected for ties
           U             W      2-Tailed P           Z      2-Tailed P
          98.0         218.0       .5668          -.6020       .5472


*/
   int nsets = 2;    // no of sets
   int ITERATIONS = 100000;
   static char tempstring[2048];
   static char tempstring2[2048];
   int j,k,l,nn,total,n0=0,n1=0,pass,ties;
   double  e;
   int count = 0;
   int counttot=0;
   double  u0tot =0.0; 
   double  u1tot =0.0; 
   double  umintot =0.0;   
   double  ztot =0.0; 
   double  permutationstot =0.0; 
   double  probtot =0.0; 
   double  pexacttot =0.0; 

   double  prob,t,r,pexact,pctties;
   double  mean, u0, u1, rtot0, rtot1, z;
   double  p=0;
   double  permutations;
   double  umin;
   int n[nsets];             // which column numbers
   int selectedcol = 0;
   for(k=0; k<nsets; k++) n[k]=0; 
   for(k=0; k<s.maxcol; k++) 
      if(s.column_selected[k]) n[selectedcol++]=k; 
   for(k=0; k<nsets; k++)
      for(j=0; j<s.maxrow[n[k]]; j++) if(s.selected[n[k]][j]) count++;
 
   for(j=0; j<s.maxrow[n[0]]; j++)
         if(s.selected[n[0]][j]) n0++;
   for(j=0; j<s.maxrow[n[1]]; j++)
         if(s.selected[n[1]][j]) n1++;
   printf("Select two columns for U test. Numbers should be integers.\n");
   printf("count %d n0 %d n1 %d\n",count,n0,n1);
   if((count<4) || (n0<1) || (n1<1)){ message((char*)"Insufficient cells selected",ERROR); return;}

 
   double  data[count+1];

   int *index;
   index = new int[count+1];

   double  *dindex;
   dindex = new double [count+1];

   int *rank;
   rank = new int[count+1];

   double  *drank;
   drank = new double [count+1];

   int *flag;  // flag of drank to prevent overwrites
   flag = new int[count+1];
   srand(time(NULL));


   for(pass=0; pass<ITERATIONS; pass++)
   {
           ties = 0;
           mean=0.0; u0=0.0; u1=0.0; rtot0=0.0; rtot1=0.0; z=0.0;
           for(k=0;k<count;k++) index[k]=k;
           for(k=0;k<count;k++) rank[k]=k;
           for(k=0;k<count;k++) drank[k]=(double )k;
           for(k=0;k<count;k++) flag[k]=0;

           //// put all the data together in one dataset
           //// do 1 normal and 100000 fuzzy p-values and calculate average
           
           l = 0;
           for(k=0; k<nsets; k++)
                for(j=0; j<s.maxrow[n[k]]; j++)
                {   
                    if(!s.selected[n[k]][j]) continue;
                    if(pass==0) data[l++] = s.x[n[k]][j];
                    if(pass>0) 
                    {    e = (double )rand()/(double )RAND_MAX /1000.0-0.0005;

                         //// Add small number to eliminate ties as recommended in
                         //// Geyer et al. Must add, not multiply to eliminate zeros

                         data[l] = s.x[n[k]][j] + e;  
                         l++;
                    }            
                }

           quicksort(data, count, index);
           for(k=0; k<count;k++) dindex[k] = (double )index[k];
           quicksort(dindex, count, rank);

           //// at this point rank is 1+rank[k]
           //// measure is data[rank[k]]
           //// now we have to take care of ties
           //// ties receive the average of their ranks

           for(k=0;k<count;k++) drank[k] = (double )rank[k];

           //for(k=0;k<count;k++) printf("k %d value %g rank %d\n",k,data[rank[k]],rank[k]);

           for(k=0; k<count;k++)
           {   j = 1;
               nn = 1;
               if(flag[k]) continue;
               total = rank[k];
               for(j=1; j+k<count; j++)
               {
                   if(data[rank[k]] == data[rank[k+j]])
                   { 
                        total += rank[k+j]; nn++;
                   }
               }
               if(nn>1)
               {   mean = (double )total / (double )nn;
                   for(j=0; j+k<count; j++)
                   {   if(data[rank[k]] == data[rank[k+j]]) 
                       {
                         drank[k+j] = mean;
                         flag[k+j] = 1;  // don't check again
                       }
                   }
               }
           }

           for(k=0; k<count;k++) if(flag[k]) ties++;

           //// at this point rank is 1+drank[k]
           //// measure is still data[rank[k]]


           for(k=0; k<n0; k++) rtot0 += 1.0 + drank[k];
           for(k=n0; k<count; k++) rtot1 += 1.0 + drank[k];

           //// u0 = rtot0 - (n0*(n0+1))/2.0;
           //// u1 = rtot1 - (n1*(n1+1))/2.0;

           //// these are equivalent, but switch u0 and u1
           //// u0 = n0*n1 + n0*(n0+1)/2.0 - rtot0;
           //// u1 = n1*n0 + n1*(n1+1)/2.0 - rtot1;

           u0 = u(n0,n1,rtot0);
           u1 = u(n1,n0,rtot1);

           umin = min(u0, u1);

           z = (umin - ((double )n0*(double )n1/2.0)) / 
               (sqrt((double )n0*(double )n1*((double )n0+(double )n1+1.0)/12.0));

           r = fabs(z) / sqrt(count);  // effect size, must be > 0.5 for exact to be ok

           t = z;   //// t value = z value for infinite no.of samples (normal approx.)

           prob = betai(0.5*10000, 0.5, 10000/(10000+t*t));
           pctties = 100.0 * (double )ties/(double )count;

           //// Calculate permutations sum of Ho possible outcomes
           //// Us a loop instead of the crazy recursive algorithm
           //// everybody else gets from Mann-Whitney's paper
          
           p = 0;
           for(j=0;j<=n0;j++)
           for(k=0;k<=n1;k++) 
               p += u(j,k,umin);
           pexact = 4.0 * umin / p;

           if(pass==0)
           {
                   sprintf(tempstring,"\%s\n\
                   Sample count               \t%d\n\
                   U value group 0            \t%g\n\
                   U value group 1            \t%g\n\
                   U value                    \t%g\n\
                   Z score                    \t%g\n\
                   Effect size                \t%g\n\
                   Pct. Ties (%d found)       \t%g\n\
                   Permutations               \t%g\n\
                   prob. (2-tailed normal)    \t%g\n\
                   prob. (2-tailed exact )    \t%g",
                       "Uncorrected for ties",
                       count,u0,u1,umin,z,r,ties,pctties,p,prob,pexact);  

                   if((n0<5 || n1<5)|| (count<20))
                       strcat(tempstring,"\nWarning: Insufficient number of samples, statistics may not be accurate");   

                   //// Print out what is being tested
                   printf("\nStatistics were performed on the following data:\n");
                   printf("Group 0 (n=%d)\n",s.maxrow[n[0]]);
                   for(k=0;k<s.maxrow[n[0]];k++) if(s.selected[n[0]][k]) printf("%d %g\n",k+1, s.x[n[0]][k]);
                   printf("\nGroup 1 (n=%d)\n",s.maxrow[n[1]]);
                   for(k=0;k<s.maxrow[n[1]];k++) if(s.selected[n[1]][k]) printf("%d %g\n",k+1, s.x[n[1]][k]);

//                   printf("\nIndex\n");
//                   for(k=0;k<count;k++) printf("%d %g index %d rank %d\n",k+1, data[k], index[k],rank[k]);
//
//                   printf("count %d n0 %d n1 %d\n",count,n0,n1);

           }
           counttot += count;
           u0tot += u0;
           u1tot += u1;
           umintot += umin;
           ztot += z;
           permutationstot += permutations;
           probtot += prob;
           pexacttot += pexact;           

           if(pass==ITERATIONS-1)
           {
                   count = counttot/(double (ITERATIONS));
                   u0 = u0tot/(double (ITERATIONS)); 
                   u1 = u1tot/(double (ITERATIONS));
                   umin = umintot/(double (ITERATIONS));
                   z = ztot/(double (ITERATIONS));
                   permutations = permutationstot/(double (ITERATIONS));
                   prob = probtot/(double (ITERATIONS));
                   pexact = pexacttot/(double (ITERATIONS));           

                   sprintf(tempstring2,"\n\n%s, Geyer & Meeden fuzzy p-values, %d iterations \n\
                   Sample count               \t%d\n\
                   U value group 0            \t%g\n\
                   U value group 1            \t%g\n\
                   U value                    \t%g\n\
                   Z score                    \t%g\n\
                   Effect size                \t%g\n\
                   Permutations               \t%g\n\
                   prob. (2-tailed normal)    \t%g\n\
                   prob. (2-tailed exact )    \t%g",
                       "Corrected for ties",
                       ITERATIONS,count,u0,u1,umin,z,r,p,prob,pexact);  
           }
   }
   strcat(tempstring, tempstring2);
   edit((char*)"Mann-Whitney U test, normal approximation",(char*)"Mann-Whitney U test, normal approximation",tempstring,22,30,1024,0);   

   delete[] flag;
   delete[] drank;
   delete[] dindex;
   delete[] index;
   delete[] rank;
}

double  u(int m, int n, double  t)
{
   return  (double )m*n + (double )m*((double )m+1.0)/2.0 - t;
}

double  randomnumber(double  d) // for calculator
{
    return (double )rand() / (double )RAND_MAX;
}


//--------------------------------------------------------------------------//
//  replacewithmean - make all selected numbers equal to their mean value   //
//  operation = 1 mean, 2 max, 3 min                                        //
//--------------------------------------------------------------------------//
void replacewithmean(char **arg, int noofargs, int operation)
{
  int j,k, count=0;
  double  mean=0.0, sum=0.0, lowest=0.0, highest=0.0;
 
  count = 0;
  if(operation==1)
  {  sum = 0.0;
     for(j=1; j<=s.maxcol; j++) 
     for(k=1; k<=s.maxrow[j]; k++) 
     {  if(s.number[j][k] && s.selected[j][k])
        {   count++;
            sum += s.x[j][k];
        }  
     }
     mean = sum / (double )count;
  }else{
     for(j=1; j<=s.maxcol; j++) 
     for(k=1; k<=s.maxrow[j]; k++) 
     {  if(s.number[j][k] && s.selected[j][k])
        {   lowest = highest = s.x[j][k];  //// get a starting value
            count++;
            break;
        }  
     }
     for(j=1; j<=s.maxcol; j++) 
     for(k=1; k<=s.maxrow[j]; k++) 
     {  if(s.number[j][k] && s.selected[j][k])
        {   
            count++;
            if(s.x[j][k] > highest) highest = s.x[j][k];
            if(s.x[j][k] < lowest) lowest = s.x[j][k];
        }  
     }
  }
  if(count<=1){ message((char*)"No cells selected", ERROR); return;}
  
  for(j=1; j<=s.maxcol; j++) 
  for(k=1; k<=s.maxrow[j]; k++) 
  {  if(s.number[j][k] && s.selected[j][k])
     {
         switch(operation)
         {   case 1: s.x[j][k] = mean; break;
             case 2: s.x[j][k] = highest; break;
             case 3: s.x[j][k] = lowest; break;
         } 
     }
  }
  refreshspreadsheet();
}

//--------------------------------------------------------------------------//
//  hex                                                                     //
//--------------------------------------------------------------------------//
void hex(char **arg, int noofargs)
{
   char tempstring[2*1024];
   char tempstring2[1024];
   static char number[1024];
   double  answer = 0.0;
   int ianswer = 0;
   char *endptr;
   int k;
   int base=10;
   if(noofargs>1) 
       strcpy(number, arg[1]);
   else       
       message((char*)"Operand", number, PROMPT, 1023);
 
   if(strchr(number, 'b'))      base=2;
   else if(strchr(number, 'o')) base=8;
   else if(strchr(number, 'h')) base=16;
   else                         base=10;

   for(k=0;k<(int)strlen(number);k++) if ( !isxdigit(number[k]) ) number[k]=0;   
   answer = strtol(number, &endptr, base);
   ianswer = (int) answer;
   binval(sizeof(ianswer), &ianswer, tempstring2);

   sprintf(tempstring, "base  2 = %s\n\
base  8 = %o\n\
base 10 = %d\n\
base 16 = %x\n",
tempstring2, ianswer, ianswer, ianswer);
   edit((char*)"Hexadecimal conversion results",
      (char*)"Hexadecimal conversion results",tempstring,8,30,1024,0);
}

//--------------------------------------------------------------------------//
//  binval - convert string to binary integer                               //
//--------------------------------------------------------------------------//
void binval(size_t const size, void const * const ptr, char *tempstring2)
{
    unsigned char *b = (unsigned char*) ptr;
    unsigned char byte;
    int pos = 0;
    int i, j;

    for (i=size-1;i>=0;i--)
    {
        for (j=7;j>=0;j--)
        {
            byte = b[i] & (1<<j);
            byte >>= j;
            tempstring2[pos++] = byte + '0';
        }
    }
}

//--------------------------------------------------------------------------//
//  errorallsets                                                            //
//--------------------------------------------------------------------------//
void errorallsets(void)
{
   Dialog dialog;
   int j,k,n,nrows,row,col;
   static int set1=1,set2=2,set3=4,set4=5,set5=6;
   double  xtot=0.0;
   double  mean, sd, sem;
   double  sumxminusmean=0.0;

   strcpy(dialog.title,"Statistics on many columns");
   strcpy(dialog.boxes[0],"Data sets");
   strcpy(dialog.boxes[1],"Starting Column");
   strcpy(dialog.boxes[2],"Ending Column");
   strcpy(dialog.boxes[3],"Columns for results");
   strcpy(dialog.boxes[4],"Column to put means");
   strcpy(dialog.boxes[5],"Column to put SDs");
   strcpy(dialog.boxes[6],"Column to put SEMs");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=INTCLICKBOX;
           sprintf(dialog.answer[1],"%d",set1);
           dialog.boxmin[1]=1; 
           dialog.boxmax[1]=s.cols;
   dialog.boxtype[2]=INTCLICKBOX;
           sprintf(dialog.answer[2],"%d",set2);
           dialog.boxmin[2]=1; 
           dialog.boxmax[2]=s.cols;
   dialog.boxtype[3]=LABEL;
   dialog.boxtype[4]=INTCLICKBOX;
           sprintf(dialog.answer[4],"%d",set3);
           dialog.boxmin[4]=1; 
           dialog.boxmax[4]=s.cols;
   dialog.boxtype[5]=INTCLICKBOX;
           sprintf(dialog.answer[5],"%d",set4);
           dialog.boxmin[5]=1; 
           dialog.boxmax[5]=s.cols;
   dialog.boxtype[6]=INTCLICKBOX;
           sprintf(dialog.answer[6],"%d",set5);
           dialog.boxmin[6]=1; 
           dialog.boxmax[6]=s.cols;
   
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=0;
   dialog.noofboxes=7;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   set1  = atoi(dialog.answer[1]);
   set2  = atoi(dialog.answer[2]);
   set3  = atoi(dialog.answer[4]);
   set4  = atoi(dialog.answer[5]);
   set5  = atoi(dialog.answer[6]);

   n     = set2 - set1 + 1;
   nrows = s.maxrow[set1];
   s.maxrow[set3] = nrows;
   s.maxrow[set4] = nrows;
   s.maxrow[set5] = nrows;
   s.maxcol = max(s.maxcol, set5);
   setcoltitle(set3,(char*)"Mean");
   setcoltitle(set4,(char*)"SD");
   setcoltitle(set5,(char*)"SEM");

   if(n==0) return;
   
   for(row=1; row<=nrows; row++)             // data points
   {
       xtot = 0.0;
       for(col=set1; col<=set2; col++)     // mean across columns
       {
           xtot += s.x[col][row];
       }
       mean = xtot / (double )n;
       sumxminusmean = 0.0;
       for(col=set1; col<=set2; col++)     // columns
       {
           sumxminusmean += (s.x[col][row] - mean) * (s.x[col][row] - mean);
       }
       sd = sqrt(sumxminusmean / (n-1));    // sd       
       sem = sd / sqrt(n);
       
       s.x[set3][row] = mean;
       s.x[set4][row] = sd;
       s.x[set5][row] = sem;       
       s.number[set3][row] = 1;
       s.number[set4][row] = 1;
       s.number[set5][row] = 1;
  }
  refreshspreadsheet();  
}

//--------------------------------------------------------------------------//
//  ttestallvalues                                                          //
//  do a t test of 2 columns for each row                                   //
//--------------------------------------------------------------------------//
void ttestallvalues(void)
{
   Dialog dialog;
   int j,k,nrows,row;
   static int set1=1,set2=2,set3=3,set4=4,set5=5,set6=6;
   double  mean1, sd1, mean2, sd2;
   int n1, n2;

   n1 = s.maxrow[set1];
   n2 = s.maxrow[set1];

   strcpy(dialog.title,"Statistics on many columns");
   strcpy(dialog.boxes[0],"Data sets");
   strcpy(dialog.boxes[1],"Column for means Set 1");
   strcpy(dialog.boxes[2],"Column for SD of Set 1");
   strcpy(dialog.boxes[3],"Column for means Set 2");
   strcpy(dialog.boxes[4],"Column for SD of Set 2");

   strcpy(dialog.boxes[5],"N (population size)");
   strcpy(dialog.boxes[6],"N set 1               ");
   strcpy(dialog.boxes[7],"N set 2               ");

   strcpy(dialog.boxes[8],"Columns for results   ");
   strcpy(dialog.boxes[9],"Column to put t values");
   strcpy(dialog.boxes[10],"Column to put p values");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=INTCLICKBOX;
           sprintf(dialog.answer[1],"%d",set1);
           dialog.boxmin[1]=1; 
           dialog.boxmax[1]=s.cols;
   dialog.boxtype[2]=INTCLICKBOX;
           sprintf(dialog.answer[2],"%d",set2);
           dialog.boxmin[2]=1; 
           dialog.boxmax[2]=s.cols;
   dialog.boxtype[3]=INTCLICKBOX;
           sprintf(dialog.answer[3],"%d",set3);
           dialog.boxmin[3]=1; 
           dialog.boxmax[3]=s.cols;
   dialog.boxtype[4]=INTCLICKBOX;
           sprintf(dialog.answer[4],"%d",set4);
           dialog.boxmin[4]=1; 
           dialog.boxmax[4]=s.cols;

   dialog.boxtype[5]=LABEL;
 
   dialog.boxtype[6]=INTCLICKBOX;
           sprintf(dialog.answer[6],"%d",n1);
           dialog.boxmin[6]=1; 
           dialog.boxmax[6]=1000;
   dialog.boxtype[7]=INTCLICKBOX;
           sprintf(dialog.answer[7],"%d",n2);
           dialog.boxmin[7]=1; 
           dialog.boxmax[7]=1000;

   dialog.boxtype[8]=LABEL;

   dialog.boxtype[9]=INTCLICKBOX;
           sprintf(dialog.answer[9],"%d",set5);
           dialog.boxmin[9]=1; 
           dialog.boxmax[9]=s.cols;
   dialog.boxtype[10]=INTCLICKBOX;
           sprintf(dialog.answer[10],"%d",set6);
           dialog.boxmin[10]=1; 
           dialog.boxmax[10]=s.cols;
   
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=0;
   dialog.noofboxes=11;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout)return;
   set1  = atoi(dialog.answer[1]);
   set2  = atoi(dialog.answer[2]);
   set3  = atoi(dialog.answer[3]);
   set4  = atoi(dialog.answer[4]);
   n1    = atoi(dialog.answer[6]);
   n2    = atoi(dialog.answer[7]);
   set5  = atoi(dialog.answer[9]);
   set6  = atoi(dialog.answer[10]);

   nrows = s.maxrow[set1];
   s.maxrow[set5] = nrows;
   s.maxrow[set6] = nrows;
   s.maxcol = max(s.maxcol, set6);
   setcoltitle(set5, (char*)"t value");
   setcoltitle(set6, (char*)"p value");

   
   for(row=1; row<=nrows; row++)             // data points
   {
       mean1 = s.x[set1][row];
       sd1   = s.x[set2][row];
       mean2 = s.x[set3][row];
       sd2   = s.x[set4][row];
       s.x[set5][row] = tvalue(mean1, sd1, n1, mean2, sd2, n2);
       s.x[set6][row] = pvalue(mean1, sd1, n1, mean2, sd2, n2);
       s.number[set5][row] = 1;
       s.number[set6][row] = 1;
  }
  refreshspreadsheet();  
}

double  tvalue(double  mean1, double  sd1, int n1, double  mean2, double  sd2, int n2)
{
   //// welch's t test
   double  t,sed;

   //// satterthwaite eq
   //df = (((sd1*sd1/n1) + (sd2*sd2/n2)) * ((sd1*sd1/n1) + (sd2*sd2/n2))) /
   //     (((sd1*sd1*sd1*sd1) / (n1*n1*(n1-1)))
   //     +((sd2*sd2*sd2*sd2) / (n2*n2*(n2-1))));

   sed = sqrt((sd1 * sd1 / n1) + (sd2 * sd2 / n2));
   t = fabs(mean1 - mean2) / sed;
   // p = betai(0.5*df, 0.5, df/(df+t*t));
   return t;
}

double  pvalue(double  mean1, double  sd1, int n1, double  mean2, double  sd2, int n2)
{
   double  p,t,df,sed;
 
   //// satterthwaite eq
   df = (((sd1*sd1/n1) + (sd2*sd2/n2)) * ((sd1*sd1/n1) + (sd2*sd2/n2))) /
        (((sd1*sd1*sd1*sd1) / (n1*n1*(n1-1)))
        +((sd2*sd2*sd2*sd2) / (n2*n2*(n2-1))));

   sed = sqrt((sd1 * sd1 / n1) + (sd2 * sd2 / n2));
   t = fabs(mean1 - mean2) / sed;
   p = betai(0.5*df, 0.5, df/(df+t*t));
   return p;
}

//--------------------------------------------------------------------------//
//  pvaluefromr                                                             //
//  calculates p value from pearson's correlation coefficient given the     //
//  n and r                                                                 //
//--------------------------------------------------------------------------//
void pvaluefromr(void)
{
  int n=0,df=0;
  double  r=0.0, p=1.0, t = 0;
  char tempstring[128];

  tempstring[0] = 0;
  message((char*)"Correlation coefficient (r)", tempstring, PROMPT, 1023);
  r = atof(tempstring);

  tempstring[0] = 0;
  message((char*)"N (number of samples)", tempstring, PROMPT, 1023);
  n = atoi(tempstring);
  t = r * sqrt(n-2) / sqrt(1-r*r); 
  df = n - 2;
  p = betai(0.5*df, 0.5, df/(df+t*t));
  sprintf(tempstring, "p-value is %g", p);
  message((char*)tempstring, INFO);
}


//--------------------------------------------------------------------------//
//  integraterange                                                          //
//  calculates integral of a column between 2 indexes w/o changing data     //
//  args = column, start x, end x                                           //
//--------------------------------------------------------------------------//
void integraterange(char **arg, int noofargs)
{
  int k;
  double  total=0.0;
  static int col, start=1, end=1;
  char tempstring[128];
  arg=arg; noofargs=noofargs;
  itoa(s.current_col, tempstring, 10);
  if(noofargs<2) 
  {     message((char*)"Column no.", tempstring, PROMPT, 1023);
        col = atoi(tempstring);
  }else col = atoi(arg[1]);

  itoa(start, tempstring, 10);
  if(noofargs<3) 
  {     message((char*)"Starting x", tempstring, PROMPT, 1023);
        start = atoi(tempstring);
  }else start = atoi(arg[2]);

  itoa(end, tempstring, 10);
  if(noofargs<4) 
  {     message((char*)"Ending x", tempstring, PROMPT, 1023);
        end = atoi(tempstring);
  }else end = atoi(arg[3]);
  
  for(k=start; k<=end; k++) total += s.x[col][k];

  sprintf(tempstring, "Total is %g", total);
  message((char*)tempstring, INFO);
}

