//--------------------------------------------------------------------------//
// xdata6.cc                                                                //
// Latest revision: 10-07-2023                                              //
// Math functions                                                           //
//--------------------------------------------------------------------------//

#include "xdata.h"
extern Globals  g;
extern Spreadsheet s;
static int infft = 0;

//--------------------------------------------------------------------------//
//  math                                                                    //
//  Handle commands such as   + 4 5   (add 5 to column 4)                   //
//     arg0 = operator                                                      //
//     arg1 = destination column                                            //
//     arg2 = equation or number                                            //
//--------------------------------------------------------------------------//
void math(char **arg, int noofargs, int function)
{
  static int col=0, rows_guess=0;
  static double aaa = 0.0;   //// a parameter, can be column no or operand
  static double bbb = 0.0;   //// a parameter, can be column no or operand
  int k, row, divisionbyzero=0, negativelogs=0, helptopic=0, backwards=0;
  int gotnumber = 0;
  static char tempstring[1024]="";
  char valuestring[1024]="";

  //// Parameter is concatenation of arg[1] +...
  if(atoi(arg[1])) col=atoi(arg[1]); else col = s.current_col;
  for(k=2; k<noofargs; k++) strcat(valuestring, arg[k]);
  valuestring[strlen(valuestring)]=0;

//printf("noofargs %d\n",noofargs);
//printf("valuestring >>>%s<<<\n",valuestring);

  //// If new column, initialize it
  s.maxcol = max(s.maxcol, 1+col);

  if(s.maxrow[col]<1) 
  {    
       s.maxrow[col] = clickbox((char*)"No. of rows", rows_guess, 0, s.rows-1, null, helptopic);
       rows_guess = s.maxrow[col];
       for(k=1; k<=s.maxrow[col]; k++) s.number[col][k]=1;
  }

  if(noofargs > 2)  //// called from editor, all params on command line
  {
       aaa = strtod(arg[2], NULL);
       bbb = strtod(arg[3], NULL);
  }else           //// single column, need operands
  {
       for(row=1; row<=s.maxrow[col];  row++)
       {      if(s.number[col][row])
              {  gotnumber++;
                 break;
              }
       }
       if(gotnumber)
       {      switch(function)
              {   case ADD:
                  case DIVIDE:
                  case SUBTRACT:
                  case MULTIPLY:
                         message((char*)"Operand", tempstring, PROMPT, 1023);
                         aaa = strtod(tempstring, NULL);
                         break;
                  case MOD:
                  case MODINVERSE:
                         message((char*)"Modulus", tempstring, PROMPT, 1023);
                         aaa = strtod(tempstring, NULL);
                         break;
                  case MODPWR:
                         message((char*)"Power", tempstring, PROMPT, 1023);
                         aaa = strtod(tempstring, NULL);
                         message((char*)"Modulus", tempstring, PROMPT, 1023);
                         bbb = strtod(tempstring, NULL);
                         break;
                  default:
                         aaa = eval(valuestring);
                         break;
              }
       }
  }



  for(row=1;  row<=s.maxrow[col];  row++)
  {    
        if(s.string[col][row])
        {  switch(function)
           {
               case SETTONUMBER: 
                               s.string[col][row] = 0;
                               s.hex[col][row] = 0;
                               s.number[col][row] = 1; break;
           }  
        }
        if(s.number[col][row])
        {  
           switch(function)
           {   
               case ADD:       s.x[col][row] += aaa; break;
                             
#ifdef HAVE_CBRT
               case CBRT:      s.x[col][row] = cbrt(s.x[col][row]); break;
#endif
               case COS:       s.x[col][row] = cos(s.x[col][row]); break;
               case DIFFERENTIATE:
                               backwards = 1;
                               if(row>1) s.x[0][row] = s.x[col][row] - s.x[col][row-1]; 
                               else s.x[0][row] = s.x[col][row];
                               break;
               case DIVIDE:    if(aaa !=0 ) s.x[col][row] /= aaa;
                                   else divisionbyzero++;
                               break;
               case EXPONENTIAL:  
                               s.x[col][row] = pow(2.718281828, s.x[col][row]); break;
                               break;
               case INT:       s.x[col][row] = int(s.x[col][row]); 
                               break;
               case INTEGRATE:
                               if(row>1) s.x[col][row] += s.x[col][row-1]; 
                               break;
               case INVERSE:   if(s.x[col][row]) s.x[col][row] = 1/s.x[col][row]; 
                               else divisionbyzero++;
                               break;
               case LOG10:     if(s.x[col][row]>0) s.x[col][row] = log10(s.x[col][row]); 
                               else negativelogs++;
                               break;
               case LN:       if(s.x[col][row]>0) s.x[col][row] = log(s.x[col][row]); 
                               else negativelogs++;
                               break;
               case MOD:       s.x[col][row] = (int)s.x[col][row] % (int)aaa; 
                               break;
               case MODPWR:  
                               s.x[col][row] = modpwr(s.x[col][row], aaa, bbb); 
                               break;
               case MODINVERSE:  
                               s.x[col][row] = modinverse(s.x[col][row], aaa); 
                               break;
               case MULTIPLY:  
                               s.x[col][row] *= aaa; 
                               break;
               case NEGATIVETOZERO:  if(s.x[col][row]<0.0) s.x[col][row] = 0; break;
               case RAMP:      s.x[col][row] = row; break; 
               case RANDOM:    s.x[col][row] = rand(); break;
               case CONVERTTOHEX: 
                               s.hex[col][row] = 1;
                               s.string[col][row] = 0;
                               s.number[col][row] = 0; break;
               case CONVERTTODEC: s.x[col][row] = aaa; break; 
               case SETTOSTRING: 
                               s.hex[col][row] = 0;
                               s.string[col][row] = 1;
                               s.number[col][row] = 0; break;
               case SIN:       s.x[col][row] = sin(s.x[col][row]); break;
               case SQRT:      s.x[col][row] = sqrt(s.x[col][row]); break;
               case SUBTRACT:  
                               s.x[col][row] -= aaa; 
                               break;
               case TAN:       
                               s.x[col][row] = tan(s.x[col][row]); break;
               case TENTOX:    s.x[col][row] = pow(10, s.x[col][row]); break;
               case TWOTOX:    s.x[col][row] = pow(2, s.x[col][row]); break;

           }
        }
  }
  if(backwards)
  {
       for(row=0; row<=s.maxrow[col];  row++)
       s.x[col][row] = s.x[0][row];
  }
  if(divisionbyzero)
  {    sprintf(tempstring,"%d divisions by\nzero detected",divisionbyzero);
       message(tempstring,ERROR);
  }
  if(negativelogs)
  {    sprintf(tempstring,"%d negative logs detected",negativelogs);
       message(tempstring,ERROR);
  }
  refreshspreadsheet();
}


//--------------------------------------------------------------------//
// math                                                               //
//--------------------------------------------------------------------//
int macro_math(int n, char **arg)
{
   n=n;arg=arg;
printf("oh no, math!/n");
return 0;
}


//--------------------------------------------------------------------//
// whatis                                                             //
//--------------------------------------------------------------------//
void whatis(char **arg, int noofargs)
{
   char tempstring[1024];
   sprintf(tempstring, "%g", eval(arg[1]));
   message(tempstring);
}


//--------------------------------------------------------------------//
// set                                                                //
//--------------------------------------------------------------------//
void set(char **arg, int noofargs)
{
   define_variable(arg[1], atof(arg[2]));
}


//--------------------------------------------------------------------//
// evaluate_formula - call eval() for entire column                   //
//--------------------------------------------------------------------//
int evaluate_formula(char *equation)
{
   int k,row,start,end,startofequal,endofequal,helptopic=0;
   char tempstring[1024];
   char formula[1024];
   char temp_formula[4*1024];
   char equals_sign[1024];
   int dcol, scol;
   int error_count = 0;

   start = string_index(equation,1,0,'y');
   startofequal = 1;
   endofequal = 1;
   //// Look for =, +=, etc
   for(k=1;k<(int)strlen(equation);k++) 
   {    if(ispunct(equation[k])) 
        {    startofequal = k;
             while(ispunct(equation[k])) endofequal=k++; 
             break; 
        }
   }
   end = strlen(equation);

   if(endofequal >= (int)strlen(equation))
   {    sprintf(tempstring,"Invalid formula\n%s",equation);
        message(tempstring,ERROR);
        return 0;
   }


   ////  Destination column
   if(startofequal <= 1)
       dcol = s.current_col;
   else
   {   strncpy(tempstring, equation+start+1, startofequal-start-1);
       tempstring[endofequal-start]=0;
       dcol = atoi(tempstring);
   }
   if(dcol >= s.cols)
   {    sprintf(tempstring,(char*)"Destination column exceeds array size\ndcol=%d, s.cols=%d",dcol,s.cols);
        message(tempstring, ERROR);
        return 0;
   }

   ////  Substitute temp variable name into formula
   strncpy(equals_sign, equation+startofequal, 1+endofequal-startofequal);
   equals_sign[ 1+endofequal-startofequal ] = 0;
   strncpy(formula, equation+endofequal+1, end-endofequal);
   formula[end-endofequal]=0;
   sprintf(temp_formula,"YYtemp%s%s", equals_sign, formula);

   ////  Evaluate entire column
   s.maxcol = max(s.maxcol, dcol);

   for(row=1; row<=s.rows; row++)
   { 
           s.number[dcol][row] = 1;
           s.string[dcol][row] = 0;

           //// s.number can be changed in eval()
           add_variable((char*)"YYtemp", s.x[dcol][row]);
            
           //// don't pop up message if dividing by zero 
           g.want_messages = 0;
           eval(temp_formula, dcol, row);
           g.want_messages = 1;
           
           read_variable((char*)"YYtemp");
           scol = g.cc;

           if(s.number[scol][row]) s.x[dcol][row] = read_variable((char*)"YYtemp");
           if(g.cc == dcol && s.maxrow[dcol]==0)  //// not reading from another column
           {    s.maxrow[dcol] = clickbox((char*)"No. of rows", s.maxrow[dcol], 0, s.rows, null, helptopic);
           }else if(row >= s.maxrow[scol])         //// reading from scol
           {    s.maxrow[dcol] = row;
                break; 
           }
           if(g.getout){ error_count++; g.getout = 0; }
           if(row>=s.rows){ g.getout = 0; break; }
   } 


   if(error_count >0)
   {
        sprintf(tempstring, "%d math errors occurred", error_count);
        message(tempstring, ERROR);
   }
   
   if(s.maxrow[dcol]<=1) 
        s.maxrow[dcol] = s.maxrow[max(1, dcol-1)];
   for(row=1; row<=s.maxrow[dcol]; row++) 
   {    s.number[dcol][row] = 1;
        s.string[dcol][row] = 0;
        s.hex[dcol][row] = 0;
   }
   g.getout = 0;

   refreshspreadsheet();
   return 0;
} 


//--------------------------------------------------------------------//
// evaluate_conditional if(y==something) for entire column            //
// The following types of formulas are handled:                       //
//    if(y==123) y=1-y                                                //
//    if(y12==123) y34=1-y45                                          //
//    if(x>4) y=-y                                                    //
//--------------------------------------------------------------------//
int evaluate_conditional(char *string)
{
   int row, col;
   char paren[128], equals[128], arg[128], colno[128];
   int in_col=1, in_paren=1, in_operator=1;
   int k, parenstate=1, lparen=0, rparen=0, lequals=0, requals=0, lcol=0, rcol=0;
   int rarg=0, larg=0;
   double answer=0;
   lparen = 2;         // if(y12 == 34) y=1-y
                       //    ^start here
                       // if(x>4) y2=y3
//printf("in evaluate conditional string=%s\n",string);
   for(k=3; k<(int)strlen(string); k++)
   {    //// location of column no.
        if(isdigit(string[k]) && in_col && lcol==0) lcol=k;
        if(isdigit(string[k]) && in_col && lcol!=0) rcol=k;
        if(rcol && rcol==lcol) rcol++;

        //// location of operator
        if(strchr("!=><", string[k]) && lequals==0 && in_operator){ in_col=0; lequals=k; }
        if(strchr("!=><", string[k]) && lequals!=0 && in_operator) requals=k;

        //// location of parentheses
        if(string[k] == '(' && in_paren) { parenstate++; in_col = 0; }
        if(string[k] == ')' && in_paren) parenstate--;
        if(parenstate==0 && in_paren){ rparen = k; in_paren=0; break; }
   }
        //// location of argument
   larg = rparen+1;
   rarg = strlen(string)-1;

   if(rparen!=lparen)
   {    strncpy(paren, string+lparen+1, rparen-lparen);
        paren[rparen-lparen-1] = 0;
   }
//printf("paren %s\n",paren);

   if(strchr(paren, 'y') && rcol>=lcol)
   {    strncpy(colno, string+lcol, rcol-lcol+1);
        colno[rcol-lcol+1] = 0;
        col = max(0,atoi(colno) - 1);
   }else col = s.current_col;

//printf("column %s %d %d %d\n",colno, lcol,rcol,col);

   if(requals!=lequals)           
   {    strncpy(equals, string+lequals, requals-lequals+1);
        equals[requals-lequals+1] = 0;
   }else equals[0] = 0;
   
   if(rarg!=larg)
   {    strncpy(arg, string+larg, rarg-larg+1);
        arg[rarg-larg+1] = 0;
   }

   for(row=0; row<s.maxrow[col]; row++)
   {   g.rr = row;
       g.cc = col;
//printf("row %d col %d  paren %s\narg %s\n",row,col,paren,arg);
       answer = eval(paren);

//printf("answer %g\n",answer);
//if(answer) printf("arg %s\n");
       if(answer && arg!=NULL) eval(arg);
   }
   refreshspreadsheet();
   return answer;
}


//--------------------------------------------------------------------//
// set_length                                                         //
//--------------------------------------------------------------------//
void set_length(char **arg, int noofargs, int dummy)
{   
    static int newlen = 0;
    dummy=dummy;
    int row,col,olen,helptopic=0;
    char tempstring[1024];
    itoa(s.current_col, tempstring, 10);
    if(noofargs<2) 
    {     message((char*)"Column no.", tempstring, PROMPT, 1023);
          col = atoi(tempstring);
    }else col = atoi(arg[1]);

    olen = s.maxrow[col];
    itoa(s.maxrow[col], tempstring, 10);
    if(noofargs<3) newlen = clickbox((char*)"New no. of rows", newlen, 0, s.rows, null, helptopic);
    else newlen = atoi(arg[2]);

    if(newlen >= g.rows)
    {  message((char*)"Too many rows, resize array in .xdata/xdata.ini");
       return;
    }
    s.maxcol = max(s.maxcol, col);
    if(newlen>olen)
    {  for(row=olen; row<=newlen; row++)        // If making bigger
       {   s.number[col][row] = 1;
           s.string[col][row] = 0;
           s.hex[col][row] = 0;
           s.x[col][row] = 0.0;
       }
       strcpy(s.text[col][row], "0");
       strcpy(s.disp[col][row], "0");
       s.maxrow[col] = newlen;
       refreshspreadsheet();
    }else
    {  
       for(row=newlen+1; row<=olen; row++)        // If making smaller, get rid of 
       {   s.number[col][row] = 0;                // numbers after the end
           s.string[col][row] = 0;
           s.hex[col][row] = 0;
           s.x[col][row] = 0.0;
           s.text[col][row][0]=0;
           s.disp[col][row][0]=0;
       }
       refreshspreadsheet();
       s.maxrow[col] = newlen;
    }
}    


//--------------------------------------------------------------------//
// shrink                                                             //
// n = no.of data points to shrink or expand into                     //
//--------------------------------------------------------------------//
void shrink(char **arg, int noofargs)
{  
  static int n=s.maxrow[s.current_col], col=s.current_col;
  int kk, oldn, newn, index;
  char tempstring[1024];
  itoa(s.current_col, tempstring, 10);
  double aa, cc, factor;
  double lower, upper;

  if(noofargs<2) 
  {     message((char*)"Column no.", tempstring, PROMPT, 1023);
        col = atoi(tempstring);
  }else col = atoi(arg[1]);
  if(noofargs<3) 
  {     sprintf(tempstring, "%d", n); 
        message((char*)"No. of data points to expand/squeeze into", tempstring, PROMPT, 1023);
        newn = atoi(tempstring);
  }else newn = atoi(arg[2]);

  if(newn>=s.rows){ message((char*)"Too large\nResize matrix first",ERROR); return; }
  if(newn < 1) newn = 1;
  double temp[newn + 2];
 
  oldn = s.maxrow[col];
  if(oldn<=1) return;

  factor = (double)(newn)/ (double)(oldn);
  for(kk=1; kk<=newn; kk++)
  {   
      //// fraction of the distance between adjacent original numbers
      index = (int)((double)(kk)/factor);  //// integer part 0, 0,    0,    1, 1, 
      aa = (double)kk / factor - index;  //// decimal part 0, 0.33, 0.66, 0, 0.33, ... if 3x

      lower = s.x[col][max(1,index)];
      upper = s.x[col][min(newn,index+1)];


      //// linear interpolation between adjacent numbers       
      cc =  (aa*upper) + (1.0-aa)*lower;

      temp[kk] = cc;
  }

  if(newn<oldn) 
  {   for(kk=oldn+1; kk>newn; kk--)
      {    
         erase_cell(kk,col);
      }
  }
  s.maxrow[col] = newn;
  for(kk=1; kk<=newn; kk++)
  {   s.x[col][kk] = temp[kk];  // substitute answer back
      s.number[col][kk] = 1;
      s.string[col][kk] = 0;
      s.hex[col][kk] = 0;
  }
  refreshspreadsheet();
  return;
}



//--------------------------------------------------------------------//
// smooth                                                             //
// n = no.of data points to shrink or expand into                     //
//--------------------------------------------------------------------//
void smooth(char **arg, int noofargs)
{   
  int index, j, k, col, n;
  static int npoint=3;
  char tempstring[1024];
  double sum;
  itoa(s.current_col, tempstring, 10);
  if(noofargs<2) 
  {     message((char*)"Column no.", tempstring, PROMPT, 1023);
        col = atoi(tempstring);
  }else col = atoi(arg[1]);
  itoa(npoint, tempstring, 10);
  if(noofargs<3) message((char*)"No. of smoothing points", tempstring, PROMPT, 1023);
  else strcpy(tempstring, arg[2]);

  npoint = atoi(tempstring);
  npoint = max(npoint,3);
  double bi[npoint];
  int total=0;
  n = s.maxrow[col];
  double temp[n];
  
  for(k=0;k<npoint;k++)
  {    bi[k]=binom(npoint-1,k);
       total += (int)bi[k]; 
  }
  for(j=0;j<=n;j++)
  {   sum = 0.0;
      for(k=0; k<npoint; k++)
      {    index = j + k-(int)(npoint/2);
           if(index>=0 && index<=s.maxrow[col]) sum+=bi[k]*s.x[col][index];
      }
      temp[j] = sum / total;
  }

  for(k=0;k<n;k++)
  {   s.x[col][k] = temp[k];  // substitute answer back
      s.number[col][k] = 1;
      s.string[col][k] = 0;
      s.hex[col][k] = 0;
  }
  refreshspreadsheet();
}



//--------------------------------------------------------------------//
// fft                                                                //
//--------------------------------------------------------------------//
void fft(char **arg, int noofargs)
{
   if(infft)return;
   infft = 1;
   static int  realcol=1, imagcol=2, realresult=3, imagresult=4,  direction=1;
   arg=arg; noofargs=noofargs; 
   int j, k, ss, nn[2], ndim=1,  size = 0;
   static complex *data;

   ////------------------------////

   Dialog dialog;
   dialog.needrepeat = 0;  
   strcpy(dialog.radio[0][0],"Direction"); 
   strcpy(dialog.radio[0][1],"Forward");             
   strcpy(dialog.radio[0][2],"Reverse");
   if(direction==1) dialog.radioset[0] = 1; else dialog.radioset[0] = 2;

   strcpy(dialog.title,"Fourier transform");
   strcpy(dialog.boxes[0],"Source");
   strcpy(dialog.boxes[1],"Column for real   ");
   strcpy(dialog.boxes[2],"Column for imaginary ");
   strcpy(dialog.boxes[3],"Results");
   strcpy(dialog.boxes[4],"Column for real   ");
   strcpy(dialog.boxes[5],"Column for imaginary ");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[3]=LABEL;

   dialog.boxtype[1]=INTCLICKBOX;
   sprintf(dialog.answer[1],"%d",realcol);
   dialog.boxmin[1]=1; 
   dialog.boxmax[1]=s.cols;

   dialog.boxtype[2]=INTCLICKBOX;
   sprintf(dialog.answer[2],"%d",imagcol);
   dialog.boxmin[2]=1; 
   dialog.boxmax[2]=s.cols;

   dialog.boxtype[4]=INTCLICKBOX;
   sprintf(dialog.answer[4],"%d",realresult);
   dialog.boxmin[4]=1; 
   dialog.boxmax[4]=s.cols;

   dialog.boxtype[5]=INTCLICKBOX;
   sprintf(dialog.answer[5],"%d",imagresult);
   dialog.boxmin[5]=1; 
   dialog.boxmax[5]=s.cols;

   dialog.radiono[0]=3;
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=1;
   dialog.noofboxes=6;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

 
   ///////---///////
   dialogbox(dialog);
   ///////---///////

   if(g.getout){ g.getout=0; infft=0; return; }
   realcol = atoi(dialog.answer[1]);
   imagcol = atoi(dialog.answer[2]);
   realresult = atoi(dialog.answer[4]);
   imagresult = atoi(dialog.answer[5]);
   if(dialog.radioset[0] == 1) direction=1; else direction = -1;

   ////------------------------////

   size = max(s.maxrow[realcol], s.maxrow[imagcol]);

   //// Calculate next higher power of 2
   for(ss=1; ss<=size*2; ss*=2) if(ss>=size) break; 
   data = new complex[2*ss];  
   for(k=1; k<=s.maxrow[realcol]; k++) data[k-1].real() = s.x[realcol][k];
   for(k=1; k<=s.maxrow[imagcol]; k++) data[k-1].imag() = s.x[imagcol][k];
   for(k=s.maxrow[realcol]+1; k<=ss; k++) data[k-1].real() = 0;
   for(k=s.maxrow[imagcol]+1; k<=ss; k++) data[k-1].imag() = 0;
   nn[1] = ss;  // no of points in 1st dimension

   fourn(data, nn, ndim, direction);

   for(k=1; k<=ss; k++) s.x[realresult][k] = data[k-1].real();
   for(k=1; k<=ss; k++) s.x[imagresult][k] = data[k-1].imag();
   if(direction==-1)
   {  for(k=1; k<=ss; k++) s.x[realresult][k] /= ss;
      for(k=1; k<=ss; k++) s.x[imagresult][k] /= ss;
   }
   for(k=1; k<=ss; k++) 
   {     s.number[realresult][k] = 1; 
         s.number[imagresult][k] = 1; 
         s.string[realresult][k] = 0; 
         s.string[imagresult][k] = 0; 
         s.hex[realresult][k] = 0; 
         s.hex[imagresult][k] = 0; 
   }
   s.maxrow[realresult] = ss;
   s.maxrow[imagresult] = ss;
   s.maxcol=max(1+realresult, max(s.maxcol, 1+imagresult));
   delete[] data;
   setcoltitle(realresult, (char*)"Real FFT");
   setcoltitle(imagresult, (char*)"Imag FFT");
   refreshspreadsheet();
   infft = 0;
}
     
 
//--------------------------------------------------------------------------//
// fourn                                                                    //
// Replaces data by its 2-dimensional discrete Fourier transform, if        //
//  isign is 1.  The lengths of each dimension (no.of complex values)       //
//  must be powers of 2.                                                    //
// If isign is -1, data are replaced by the inverse transform times the     //
//  product of the lengths of all dimensions.                               //
// Data contains all the dimensions in 1 array and starts at data[1] not    //
//  data[0]. Lowest dimensions change fastest.                              //
// nn[] contains the no.of elements in each dimension.                      //
// Modified to use complex numbers, and to make array indices start with 0  //
//  instead of 1.                                                           //
//--------------------------------------------------------------------------//
void fourn(complex data[], int nn[], int ndim, int isign)
{  
  int idim;
  uint i1,i2,i3,i2rev,i3rev,ip1,ip2,ip3,ifp1,ifp2;
  uint ibit,k1,k2,n,nprev,nrem,ntot;
  complex temp, w, wp;
  double theta,wtemp;

  ntot=1;
  for(idim=1;idim<=ndim;idim++) ntot *= nn[idim];
  nprev=1;
  
  for(idim=ndim;idim>=1;idim--)            // main loop over the dimensions
  { 
     n=nn[idim];
     nrem=ntot/(n*nprev);
     ip1=nprev<<1;
     ip2=ip1*n;
     ip3=ip2*nrem;
     i2rev=1;
     for(i2=1;i2<=ip2;i2+=ip1)             // Bit-reversal section
     {  if(i2<i2rev)
        {  for(i1=i2;i1<=i2+ip1-2;i1+=2)
           {  for(i3=i1/2;i3<=ip3/2;i3+=ip2/2)
              {  i3rev=i3+(i2rev-i2)/2;
                 cswap(data[i3], data[i3rev]);
              }
           }
        }
        ibit=ip2>>1;
        while((ibit>=ip1)&&(i2rev>ibit))
        {  i2rev-=ibit;
           ibit>>=1;
        }
        i2rev += ibit;
     }
     ifp1=ip1;
     while(ifp1<ip2)                       // Danielson-Lanczos section
     {  ifp2=ifp1<<1;
        theta = (double)isign*6.28318530717959/(double)(ifp2/ip1);
        wtemp = sin(0.5*theta);
        wp.real() = -2.0*wtemp*wtemp;
        wp.imag() = sin(theta);
        w.real() = 1.0;
        w.imag() = 0.0;
        for(i3=1;i3<=ifp1;i3+=ip1)
        {  for(i1=i3;i1<=i3+ip1-2;i1+=2)
           {  for(i2=i1;i2<=ip3;i2+=ifp2)
              {  k1=i2/2;                    // Danielson-Lanczos formula
                 k2=k1+ifp1/2;
                 temp = w * data[k2];        // Complex multiplication
                 data[k2] = data[k1] - temp;
                 data[k1] = data[k1] + temp;
              }
           }
           w = w + wp * w;
        }
        ifp1=ifp2;
     }
     nprev *= n;
  } 
}



//--------------------------------------------------------------------------//
//  cswap                                                                   //
//--------------------------------------------------------------------------//
void cswap(complex &a, complex &b)
{
     dswap(a.real(), b.real());
     dswap(a.imag(), b.imag());
}


//--------------------------------------------------------------------------//
// linreg  linear regression by least squares                               //
// If points are vertical,slope is set to 10000                             //
// Returns error code (OK if successful).                                   //
//--------------------------------------------------------------------------//
int linreg(char **arg, int noofargs)
{
   static int regorder=1;
   arg=arg; noofargs=noofargs;
   int i,i2=0,j,k,order=0,t=0,df;
   double bb,cc=0, ff= 0.0,nn,xx,yy,r2,ss=0,see = 0.0,tt=0,xj2=0,xk2=0,xj3=0,chisq=0,prob=0,sig2i,dy;
   double sx=0.0, sy=0.0, sxoss, st2=0.0, siga=0, sigb=0, sss, ttt, aaa=0, bbb=0;
   double sumx, sumy, ppx, ppy, meanx, meany, xxx, yyy, xvariance, yvariance, 
          sdx, sdy, sex=0.0, sey=0.0, sb=0.0;
   double sigdat, chisqfit=0, tvalue=0, coeffcorr;
   char qinv0[1000];
   char qinv1[1000];
   int ok;
   double *xraa = NULL;
   double *ee = NULL;
   double q[31];
   int n;
   static char tempstring[65536];
   char tempstring2[1024];

   double *data[2];
   int count[2] = {0,0};
   static int sets[2];  
   sets[0]=0;
   sets[1]=0;

   if(count_two_datasets(sets, count)!=OK) return ERROR; 

   ////------------------------////

   Dialog dialog;
   dialog.needrepeat = 0;  
   strcpy(dialog.title,"Linear regression");
   strcpy(dialog.boxes[0],"Column for x");
   strcpy(dialog.boxes[1],"Column for y");
   strcpy(dialog.boxes[2],"Regression order");

   dialog.boxtype[0]=INTCLICKBOX;
   sprintf(dialog.answer[0],"%d",sets[0]);
   dialog.boxmin[0]=1; 
   dialog.boxmax[0]=s.cols;

   dialog.boxtype[1]=INTCLICKBOX;
   sprintf(dialog.answer[1],"%d",sets[1]);
   dialog.boxmin[1]=1; 
   dialog.boxmax[1]=s.cols;

   dialog.boxtype[2]=INTCLICKBOX;
   sprintf(dialog.answer[2],"%d",regorder);
   dialog.boxmin[2]=0; 
   dialog.boxmax[2]=30;

   dialog.radiono[0]=0;
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=0;
   dialog.noofboxes=3;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////
 
   if(g.getout){ g.getout=0; return ABORT; }
   sets[0] = atoi(dialog.answer[0]);
   sets[1] = atoi(dialog.answer[1]);
   regorder = atoi(dialog.answer[2]);


   ////------------------------////

   data[0] = new double[count[0]];  if(data[0]==NULL) return NOMEM;  // x
   data[1] = new double[count[1]];  if(data[1]==NULL) return NOMEM;  // y
   n = count[0];
   if(count[0] != count[1]){ message((char*)"Selected columns must be same length"); return ERROR; }


   get_two_datasets(data, sets);

   for(k=0;k<n;k++)
   {  
      if(data[0][k]==0) data[0][k] = 1e-25;
      if(data[1][k]==0) data[1][k] = 1e-25;
   }

   nn = (double)n;
   if(n < 2)
   {  message((char*)"Insufficient data selected",ERROR);
      return ERROR;
   }   
   if((regorder >= n-1)||(regorder < 0))
   {  message((char*)"Bad regression order",ERROR);
      return ERROR;
   }   


   i = 2 * regorder + 3;
   xraa = new double[i];
   ee = new double[4 * regorder];
   array<double> r(regorder+3,regorder+3); 
   if(!r.allocated) goto linregend;

   for(i=2; i<=2*regorder+1; i++) xraa[i] = 0;
   for(i=1; i<=regorder+2; i++) ee[i] = 0;
   xraa[1] = nn;
   for(i=1;i<=n;i++)
   {  yy = data[1][i-1];
      xx = data[0][i-1];
      for(j=2;j<=2*regorder+1;j++) xraa[j] += pow(xx,(double)j - 1.0);
      for(j=1;j<=regorder+1;j++)
      {  ee[j] += yy * pow(xx,(double)j - 1.0);
         r.p[j][regorder+2] = ee[j];
      }
      ee[regorder + 2] += yy*yy;
   }
   for(i=1;i<=regorder+1;i++)
   {  for(j=1;j<=regorder+1;j++)
        r.p[i][j] = xraa[i+j-1];
   }     
   ok=0;

   for(order=1;order<=regorder+1;order++)
   {
      for(t=order;t<=regorder+1;t++) if(r.p[t][order] != 0.0){ ok=1; break; } 
      if(!ok){ message((char*)"No unique solution",ERROR); goto linregend; }
      for(j=1;j<=regorder+2;j++)
      {   bb = r.p[order][j];
          r.p[order][j] = r.p[t][j];
          r.p[t][j] = bb;
      }

      if(fabs(r.p[order][order])>1e-30)  cc=1.0/r.p[order][order]; //check for underflow
      else 
      {  message((char*)"Division by zero",ERROR); 
         goto linregend; 
      }
      for(j=1;j<=regorder+2;j++) r.p[order][j] = cc * r.p[order][j];


      for(t=1;t<=regorder+1;t++)
      {  if(t!=order) 
         {  cc = -r.p[t][order];
            for(j=1;j<=regorder+2;j++) r.p[t][j] += cc * r.p[order][j];
         }
      }   
   }                                       


    //---------Part 2--------//

   ss = 0.0;                                   // sum of squares of regression

   for(i=2;i<=regorder+1;i++) ss += r.p[i][regorder+2]*(ee[i]-xraa[i]*ee[1]/nn);
   tt = ee[regorder+2] - ee[1]*ee[1]/nn;        // total sum of squares
   cc = tt - ss;                               // sum of squares of residual
   i2 = n - regorder - 1;                      // deg.freedom of residual
   if(regorder!= 0) xj2 = ss/(double)regorder; // mean square of regression
   xk2 = cc / (double)i2;                      // mean square of residual
   if(xk2!=0.0) ff = xj2/xk2;                  // f-value
   if(tt!=0.0) xj3=ss/tt;                      // coeff. of determination
                                               // k-degree coefficient
                                               // q[0] = 0-degree coeff
   for(i=1;i<=regorder+1;i++) q[i-1]=r.p[i][regorder+2]; 
   r2 = sqrt(xj3);                             // coeff.of correlation
   if(i2) see = sqrt(cc / i2);                 // std.error of estimate
   see=see + ff - r2;                          // shut up compiler

   //--------calculate chi squared and goodness of fit-------//

   chisqfit = 0.0;
   for(i=1; i<=n; i++)
   {   xx = data[0][i-1];
       yy = q[0];           // yy is estimated y value
       for(j=1; j<=regorder; j++) yy += q[j] * pow(xx, j);
       dy = data[1][i-1] - yy;
       sig2i = divide(1, data[0][i-1] * data[0][i-1]);
       chisqfit += dy * dy * sig2i;
   }

   
   // Probability value for goodness of fit. Only meaningful if each
   // data point has a SD value, otherwise it always gives 1, don't use.   
   // prob = 1 - gammq(0.5*(n-2), 0.5*chisqfit); 

   //------Calculate standard deviation of estimates---------//

   st2 = 0.0; sx = 0.0; sy = 0.0; aaa=0.0; bbb=0.0;
   for(i=1; i<=n; i++)
   {   sx += data[0][i-1]; // x
       sy += data[1][i-1]; // y
   }
   sss = n;
   sxoss = sx / sss;   
   for(i=1; i<=n; i++)
   {   ttt = data[0][i-1] - sxoss;
       st2 += ttt * ttt;
       bbb += ttt * data[1][i-1];
   }
   bbb /= st2;
   aaa = (sy - sx * bbb)/sss;

   chisq = 0.0;
   for(i=1; i<=n; i++)
       chisq += square( data[1][i-1] - aaa - bbb * data[0][i-1] );

   siga = sqrt((1.0 + sx * sx / (sss * st2))/sss);
   sigb = sqrt(1.0 / st2);
   sigdat = sqrt(chisq / (n-2));   
   siga *= sigdat;
   sigb *= sigdat;

   //------Calculate t and p(>t)-----------------------------//


   if(tt==0.0) xj3 = 0;
   else        xj3 = ss/tt; 
   if(xj3>0) coeffcorr = sqrt(xj3); else coeffcorr=0;
   df = n-2;
   tvalue = coeffcorr * sqrt( (n-2) / (1-square(coeffcorr))) ;
   prob = betai(0.5*df, 0.5, df/(df+tvalue*tvalue));


linregend:

   sprintf(tempstring,"Regression table\nRegression order \t%d\n",regorder);
   sprintf(tempstring2, "x data set %d  y data set %d\n", sets[0], sets[1]); strcat(tempstring,tempstring2);
   sprintf(tempstring2, "Source       Sum of Squares  Deg. Freedom  Mean Sq.\n");strcat(tempstring,tempstring2);
   sprintf(tempstring2, "-------   ------------ --------  --------\n");strcat(tempstring,tempstring2);
   sprintf(tempstring2, "Regression \t%g      \t    %d      \t %g\n", ss, regorder, xj2);strcat(tempstring,tempstring2);
   sprintf(tempstring2, "Residual   \t%g      \t    %d      \t %g\n", cc, i2, xk2);strcat(tempstring,tempstring2);
   sprintf(tempstring2, "Total      \t%g      \t    %d      \t   \n", tt, n-1);strcat(tempstring,tempstring2);
   if(xk2==0){ sprintf(tempstring2, " F = infinity\n");strcat(tempstring,tempstring2); }
   else      { sprintf(tempstring2, " F = %g\n", xj2 / xk2);strcat(tempstring,tempstring2); }
   sprintf(tempstring2, " Coeff. of determination = %g\n",xj3);strcat(tempstring,tempstring2);
   for(i=1;i<=regorder+1;i++) 
   {   sprintf(tempstring2, " %d deg coeff =\t%g\n", i-1, q[i-1]);
       strcat(tempstring,tempstring2);
   }
   if(xj3>0){ sprintf(tempstring2, " Coeff. of correlation = %g\n",sqrt(xj3));strcat(tempstring,tempstring2);}
   if(i2!=0) see = sqrt(cc/i2);
   if(cc>0) { sprintf(tempstring2, " Std. err. of estimate = %g\n",see);strcat(tempstring,tempstring2);}
   sprintf(tempstring2, "t value         = %g\n", tvalue  ); strcat(tempstring,tempstring2);
   sprintf(tempstring2, "p(>t)           = %g\n", prob    ); strcat(tempstring,tempstring2);
   sprintf(tempstring2, "Chi sq. of fit  = %g\n", chisqfit); strcat(tempstring,tempstring2);


   if(regorder == 1)
   {  sprintf(tempstring2, "Chi sq. of est. = %g\n", chisq); strcat(tempstring,tempstring2);
      sprintf(tempstring2, "param 0 +/- SD    = %g +/- %g\n", aaa, siga); strcat(tempstring,tempstring2);
      sprintf(tempstring2, "param 1 +/- SD    = %g +/- %g\n", bbb, sigb); strcat(tempstring,tempstring2);

      //// inverse
      sprintf(tempstring2, "\nInverse parameters (x = my + intercept) +/- SD\n"); strcat(tempstring,tempstring2);
      if(bbb == 0.0) strcpy(qinv0, "Infinity"); else sprintf(qinv0, "%g +/- %g", -aaa/bbb, fabs(siga/bbb));
      if(bbb == 0.0) strcpy(qinv1, "Infinity"); else sprintf(qinv1, "%g +/- %g", 1/bbb, fabs(sigb/(bbb*bbb)));
      sprintf(tempstring2, "  0 degree coeff %s\n", qinv0); strcat(tempstring,tempstring2);
      sprintf(tempstring2, "  1 degree coeff %s\n", qinv1); strcat(tempstring,tempstring2);
   }



   //// stuff needed to compare two regression lines
   ////  x = data[0], y = data[1]

   sumx = 0.0; xvariance = 0.0;
   sumy = 0.0; yvariance = 0.0;
   ppx  = 0.0; 
   ppy  = 0.0;   
   for(j=0; j<nn; j++){ sumx += data[0][j]; sumy += data[1][j]; }
   meanx = sumx / nn;
   meany = sumy / nn;
   for(j=0; j<nn; j++) 
   {     xxx = data[0][j] - meanx;          
         yyy = data[1][j] - meany;
         xvariance += ( ppx = xxx * xxx);
         yvariance += ( ppy = yyy * yyy);
   }
   xvariance /= (nn - 1);
   sdx = sqrt(xvariance);
   sdy = sqrt(yvariance);
   sex = sdx / sqrt(nn);
   sey = sdy / sqrt(nn);
   sex = sex; //// shut up compiler
   sey = sey; //// shut up compiler
   sb = see / (sdx * sqrt(nn - 1));

   sprintf(tempstring2, "x variance %g\n", xvariance); strcat(tempstring,tempstring2);
   sprintf(tempstring2, "y variance %g\n", yvariance); strcat(tempstring,tempstring2);
   sprintf(tempstring2, "SDx (std dev x)   %g\n", sdx); strcat(tempstring,tempstring2);
   sprintf(tempstring2, "SDy (std dev y)   %g\n", sdy); strcat(tempstring,tempstring2);
   sprintf(tempstring2, "SEx (std err x)   %g\n", sex); strcat(tempstring,tempstring2);
   sprintf(tempstring2, "SEy (std err y)   %g\n", sey); strcat(tempstring,tempstring2);
   sprintf(tempstring2, "SEE (std err y/x) %g\n", see); strcat(tempstring,tempstring2);
   sprintf(tempstring2, "SB                %g\n", sb); strcat(tempstring,tempstring2);
   sprintf(tempstring2, "B1 (slope)        %g\n", bbb); strcat(tempstring,tempstring2);

   //// 
   //// to compare two lines, combine the numbers from two linear regressions
   //// sb1b2 = sqrt(sb1^2 + sb2^2)
   //// t = (b1-b2) /(sb1b2)
   //// df = n1+ n2 - 4
   ////////////
   
   edit((char*)"Linear regression",(char*)"Linear regression results",tempstring,21,30,1024,0);

   delete[] xraa;
   delete[] ee;
   delete[] data[1];  
   delete[] data[0];
   return OK;
}

//--------------------------------------------------------------------------//
// pvaluelinreg                                                             //
//--------------------------------------------------------------------------//
void pvaluelinreg(void)
{
   //// from: https://www.real-statistics.com/regression/hypothesis-testing-significance-regression-line-slope/comparing-slopes-two-independent-samples/

   Dialog dialog;
   int j,k;
   static double sb1=0.0, sb2=0.0, b1=0.0, b2=0.0;
   static int n1=0, n2=0;

   strcpy(dialog.title,"Compare slope of two lines using values from linear regression");
   strcpy(dialog.boxes[0],"Values from lin reg 1");
   strcpy(dialog.boxes[1],"SB 1");
   strcpy(dialog.boxes[2],"Slope 1");
   strcpy(dialog.boxes[3],"n 1");
   strcpy(dialog.boxes[4],"Values from lin reg 2");
   strcpy(dialog.boxes[5],"SB 2");
   strcpy(dialog.boxes[6],"Slope 2");
   strcpy(dialog.boxes[7],"n 2");

   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=STRING; sprintf(dialog.answer[1],"%g",sb1);
   dialog.boxtype[2]=STRING; sprintf(dialog.answer[2],"%g",b1);
   dialog.boxtype[3]=STRING; sprintf(dialog.answer[3],"%d",n1);
   dialog.boxtype[4]=LABEL;
   dialog.boxtype[5]=STRING; sprintf(dialog.answer[5],"%g",sb2);
   dialog.boxtype[6]=STRING; sprintf(dialog.answer[6],"%g",b2);
   dialog.boxtype[7]=STRING; sprintf(dialog.answer[7],"%d",n2);
   
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=0;
   dialog.noofboxes=8;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path, ".");

   ///////---///////
   dialogbox(dialog);
   ///////---///////

   sb1  = atof(dialog.answer[1]);
   b1   = atof(dialog.answer[2]);
   n1   = atoi(dialog.answer[3]);
   sb2  = atof(dialog.answer[5]);
   b2   = atof(dialog.answer[6]);
   n2   = atoi(dialog.answer[7]);
   pvaluelinregcheck(sb1,b1,n1,sb2,b2,n2);
 
}

//--------------------------------------------------------------------------//
//  pvaluelinregcheck                                                       //
//--------------------------------------------------------------------------//
void pvaluelinregcheck(double sb1, double b1, int n1, double sb2, double b2, int n2)
{

   double  tt, prob=0.0, sb1b2;
   char tempstring[1024];
   int df;

   df = n1 + n2 - 4;
   sb1b2 = sqrt(sb1*sb1 + sb2*sb2);
   tt = (b1 - b2) / sb1b2;  //// t value
   if(tt !=0)
   {
      prob = betai(0.5*df, 0.5, df/(df + tt * tt));
      sprintf(tempstring, (char*)"P value is %g", prob);
      message(tempstring, INFO);
   }else
      message((char*)"t value is 0", ERROR);

}



//--------------------------------------------------------------------------//
// lr    1st order only experimental                                        //
// aaa = intercept  bbb = slope                                             //
//--------------------------------------------------------------------------//
void lr(double *xxx, double *yyy, int n, int &i2, double &aaa, double &siga, double &bbb, double &sigb, 
        double &cc,  double &ss, double &tt, double &xj2, double &xj3, double &xk2)
{
return;
   static int regorder=1;
   int i, j, order=1, t=0, alloc1=0, alloc2=0;
   double bb,  f, nn, xx, yy, r2, see, chisq=0, sig2i, dy;
   double sx=0.0, sy=0.0, sxoss, st2=0.0, sss, ttt;
   double sigdat, chisqfit=0;

   int ok;
   double *xraa = NULL;
   double *e = NULL;
   double q[31];

   nn = n;
   i = 2 * regorder + 3;
   xraa = new double[i];
   e = new double[regorder + 3];
   array<double> r(regorder+3,regorder+3); 
   if(!r.allocated) goto linregend;

   for(i=2;i<=2*regorder+1;i++) xraa[i] = 0;
   for(i=1;i<=regorder+2;i++) e[i] = 0;
   xraa[1] = nn;
   for(i=1;i<=n;i++)
   {  yy = yyy[i-1];
      xx = xxx[i-1];
      for(j=2;j<=2*regorder+1;j++) xraa[j] += pow(xx,(double)j - 1.0);
      for(j=1;j<=regorder+1;j++)
      {  e[j] += yy * pow(xx,(double)j - 1.0);
         r.p[j][regorder+2] = e[j];
      }
      e[regorder + 2] += yy*yy;
   }
   for(i=1;i<=regorder+1;i++)
   {  for(j=1;j<=regorder+1;j++)
        r.p[i][j] = xraa[i+j-1];
   }     
   ok=0;

   for(order=1;order<=regorder+1;order++)
   {
      for(t=order;t<=regorder+1;t++) if(r.p[t][order] != 0.0){ ok=1; break; } 
      if(!ok){ message((char*)"No unique solution",ERROR); goto linregend; }
      for(j=1;j<=regorder+2;j++)
      {   bb = r.p[order][j];
          r.p[order][j] = r.p[t][j];
          r.p[t][j] = bb;
      }

      if(fabs(r.p[order][order])>1e-30)  cc=1.0/r.p[order][order]; //check for underflow
      else 
      {  message((char*)"Division by zero",ERROR); 
         goto linregend; 
      }
      for(j=1;j<=regorder+2;j++) r.p[order][j] = cc * r.p[order][j];


      for(t=1;t<=regorder+1;t++)
      {  if(t!=order) 
         {  cc = -r.p[t][order];
            for(j=1;j<=regorder+2;j++) r.p[t][j] += cc * r.p[order][j];
         }
      }   
   }                                       

    //---------Part 2--------//

   ss = 0.0;                               // sum of squares of regression

   for(i=2;i<=regorder+1;i++) ss += r.p[i][regorder+2]*(e[i]-xraa[i]*e[1]/nn);
   tt = e[regorder+2] - e[1]*e[1]/nn;          // total sum of squares
   cc = tt - ss;                               // sum of squares of residual
   i2 = n - regorder - 1;                      // deg.freedom of residual
   if(regorder!= 0) xj2 = ss/(double)regorder; // mean square of regression
   xk2 = cc / (double)i2;                      // mean square of residual
   if(xk2!=0.0) f=xj2/xk2;                     // f-value
   if(tt!=0.0) xj3=ss/tt;                      // coeff. of determination
                                               // k-degree coefficient
                                               // q[0] = 0-degree coeff
   for(i=1;i<=regorder+1;i++) q[i-1]=r.p[i][regorder+2]; 
   r2 = sqrt(xj3);                             // coeff.of correlation
   if(i2) see=sqrt(cc / i2);                   // std.error of estimate
   see=see+f-r2;                               // shut up compiler

   //--------calculate chi squared and goodness of fit-------//

   chisqfit = 0.0;
   for(i=1; i<=n; i++)
   {   xx = xxx[i-1];
       yy = q[0];           // yy is estimated y value
       for(j=1; j<=regorder; j++) yy += q[j] * pow(xx, j);
       dy = yyy[i-1] - yy;
       sig2i = divide(1, xxx[i-1] * xxx[i-1]);
       chisqfit += dy * dy * sig2i;
   }
   
   // Probability value for goodness of fit. Only meaningful if each
   // data point has a SD value, otherwise it always gives 1, don't use.   
   // prob = 1 - gammq(0.5*(n-2), 0.5*chisqfit); 

   //------Calculate standard deviation of estimates---------//

   st2 = 0.0; sx = 0.0; sy = 0.0; aaa=0.0; bbb=0.0;
   for(i=1; i<=n; i++)
   {   sx += xxx[i-1]; // x
       sy += yyy[i-1]; // y
   }
   sss = n;
   sxoss = sx / sss;   
   for(i=1; i<=n; i++)
   {   ttt = xxx[i-1] - sxoss;
       st2 += ttt * ttt;
       bbb += ttt * yyy[i-1];
   }
   bbb /= st2;                 /// slope
   aaa = (sy - sx * bbb)/sss;  /// intercept

   chisq = 0.0;
   for(i=1; i<=n; i++)
       chisq += square( yyy[i-1] - aaa - bbb * xxx[i-1] );

   siga = sqrt((1.0 + sx * sx / (sss * st2))/sss);
   sigb = sqrt(1.0 / st2);
   sigdat = sqrt(chisq / (n-2));   
   siga *= sigdat;
   sigb *= sigdat;

   //------Calculate t and p(>t)-----------------------------//

   if(tt==0.0) xj3 = 0;
   else        xj3 = ss/tt; 

linregend:
   if(alloc1) delete[] xraa;
   if(alloc2) delete[] e;
}


//-------------------------------------------------------------------------//
// cint                                                                    //
//-------------------------------------------------------------------------//
int cint(double a)
{   return((int)(0.49999*sgn(a) + a)); }


//--------------------------------------------------------------------------//
// subtract_baseline                                                        //
//--------------------------------------------------------------------------//
void subtract_baseline(char **arg, int noofargs)
{
  static int blcount = 50;
  int col;
  char tempstring[1024];
  itoa(s.current_col, tempstring, 10);

  if(noofargs<2) col = s.current_col;
  else           col = atoi(arg[1]);

  if(noofargs<3)
         blcount = clickbox((char*)"Baseline smoothness", blcount, 0, 1000, null, 0);
  else   blcount = atoi(arg[2]);

  if(!g.getout)
  {      baseline(s.x[col], s.maxrow[col]+1, blcount);
         refreshspreadsheet();
  }
}


//--------------------------------------------------------------------------//
//  baseline                                                                //
//  Automatically subtracts baseline from data set of count points          //
//--------------------------------------------------------------------------//
void baseline(double *pdata, int count, int iterations)
{  
   int h, i, j, k;
   int decimation_factor = 10;
   double *tdata;
   tdata = new double[count];
   double *udata;
   udata = new double[count];
   double tot, avg;
   char tempstring[1024];
   int answer=YES;

   pdata++;  //// this corrects for the fact that data now starts at
             //// index of 1 instead of 0. 

   //// If there are negative numbers, we can't subtract a baseline
   //// so let them opt out
   int noofnegatives=0;
   for(k=0; k<count; k++) if(pdata[k]<0.0) noofnegatives++;
   if(noofnegatives)
   {   sprintf(tempstring,"%d negative values found\nCannot subtract baseline from negative numbers.\nThese numbers will be ignored\nContinue?",noofnegatives);
       switch(message(tempstring,YESNOQUESTION))
       {  case YES: answer=YES;   break;  // yes, continue
          case NO:     answer=NO; break;  // No, dont do it
          case CANCEL: answer=NO; break;  // No, dont do it
       }
   }
   if(answer==NO) return;

   for(k=0; k<count; k++) tdata[k] = 0.0;
   for(k=0; k<count; k++) udata[k] = 0.0;
   if(count<50) decimation_factor = 3;
   if(count<15) decimation_factor = 1;
   h = 0;
   for(k=0; k<count-decimation_factor; k+=decimation_factor) 
   {    tot = 0;
        for(i=0; i<decimation_factor; i++) 
            tot += pdata[k+i];
        avg = tot / decimation_factor;
        tdata[h++] = avg;
   }
   for(j=0; j<iterations; j++)
   {   
        smooth_baseline(tdata, count/decimation_factor, 21);

        for(k=0; k<int(count/decimation_factor); k++)
        {   if(tdata[k] > pdata[k*decimation_factor]) tdata[k]=pdata[k*decimation_factor];
            if(tdata[k] < 0) tdata[k]=0;
        }    
   }
   for(k=0, h=0; k<count-decimation_factor; h++,k+=decimation_factor) 
        for(i=0; i<decimation_factor; i++) 
            udata[k+i] = 0.96 * tdata[h];

   smooth_baseline(udata, count, 21);
   for(k=1; k<count; k++)
   {    pdata[k] -= udata[k];
   }
   pdata[0]=pdata[1];
   delete[] udata;
   delete[] tdata;

}


//--------------------------------------------------------------------------//
//  smooth_baseline                                                         //
//--------------------------------------------------------------------------//
void smooth_baseline(double* pdata, int count, int width)
{
    int j,k,tot,w;
    static int gau3[ 4] = {  1, 2,  1                     };
    static int gau5[ 6] = {  1, 4,  6,  4,   1            };
    static int gau7[ 8] = {  1, 6, 15, 20,  15,   6,   1  };
    static int gau9[10] = {  1, 8, 28, 56,  70,  56,  28,   8,   1         };
    static int gau11[12]= {  1,10, 45,120, 210, 252, 210, 120,  45,  10, 1 };
    static int gau15[16]= {  1,14, 91,364,1001,2002,3003,3432,3003,2002,1001,
                           364,91, 14,  1 }; 
    static int gau19[20]= {  1,18,153,816,3060,8568,18564,31824,43758,48620,
                           43758,31824,18564,8568,3060,816,153,18,1 };
    static int gau21[22]= {  1,20,190, 1140, 4845, 15504,  38760, 77520, 
                            125970, 167960, 184756, 167960, 125970, 77520,
                             38760,  15504,   4845,   1140, 190, 20, 1 }; 
    double testval;
    double *tdata;
    tdata = new double[count+10];
    if(tdata==NULL){ message((char*)"No memory"); return; }
    for(k=0;k<count;k++) tdata[k]=pdata[k];
    switch(width)
    {    case 3: { w=1; tot=4;  break;  }
         case 5: { w=2; tot=16; break;  }
         case 7: { w=3; tot=64; break;  }
         case 9:  { w=4; tot=256;     break; }
         case 11: { w=5; tot=1024;    break; }
         case 15: { w=7; tot=16384;   break; }
         case 19: { w=9; tot=262144;  break; }
         case 21: { w=10;tot=1048576; break; }
         default: { message((char*)"Error", ERROR); delete[] tdata; return; }
    }
    for(k=0;k<count;k++)
    {  tdata[k]=0;
       for(j=-w;j<=w;j++)
       {  
          if(k+j<0) testval = pdata[0];
          else if(k+j>=count) 
              testval = pdata[count-1];
          else 
              testval = pdata[k+j];
          switch(width)
          {    case 3: { tdata[k] += testval*gau3[w+j]; break; }
               case 5: { tdata[k] += testval*gau5[w+j]; break; }
               case 7: { tdata[k] += testval*gau7[w+j]; break; }
               case 9: { tdata[k] += testval*gau9[w+j]; break; }
               case 11:{ tdata[k] += testval*gau11[w+j]; break; }
               case 15:{ tdata[k] += testval*gau15[w+j]; break; }
               case 19:{ tdata[k] += testval*gau19[w+j]; break; }
               case 21:{ tdata[k] += testval*gau21[w+j]; break; }
          }
       }
       if(tot!=0) tdata[k] /= tot;
    }      
    for(k=0;k<count;k++) pdata[k]=tdata[k];
    delete[] tdata;
    return;
}

//--------------------------------------------------------------------------//
//  decimate                                                                //
//  remove values close to 0 in thermo total ion chromatogram               //
//--------------------------------------------------------------------------//
void decimate(char **arg, int noofargs)
{
  static double minpct = 25.00;
  int col = s.current_col;
  int n = s.rows;
  int j, k, pos;
  double amax;
  double minfrac=0.0;
  char tempstring[1024];
  char returnstring[1024];
  if(noofargs<2) 
  {     sprintf(tempstring, "Minimum percent to include [%g]", minpct);
        sprintf(returnstring, "%2.4g", minpct);
        message(tempstring, returnstring, PROMPT, 64);
        minpct = atof(returnstring);
  }else minpct = atof(arg[1]);
  minfrac = minpct * 0.01;
  
  pos = 0;
  for(k=3; k<n-3; k++)
  {
        amax = max(s.x[col][k-3], max(s.x[col][k-2], max(s.x[col][k-1], max(s.x[col][k+0], max(s.x[col][k+1], s.x[col][k+2])))));
        if(s.x[col][k] / amax >= minfrac)
        {
             for(j=0; j<s.cols; j++)
                 s.x[j][pos] = s.x[j][k];
             pos++;           
        }
  }  
  for(j=0; j<s.maxcol; j++) s.maxrow[j] = pos;
  s.rows = pos;
  refreshspreadsheet();
}


//--------------------------------------------------------------------------//
//  remove spikes                                                           //
//  remove values >10x higher than neigbbors in thermo total ion chromatogram//
//--------------------------------------------------------------------------//
void remove_spikes(char **arg, int noofargs)
{
  static double spikefac = 1.50;
  int col = s.current_col;
  int n = s.maxrow[col];

  int k;
  double a1,a2;
  char tempstring[1024];
  char returnstring[1024];
  if(noofargs<2) 
  {     sprintf(tempstring, "Factor to count as spike [%g]", spikefac);
        sprintf(returnstring, "%3.3g", spikefac);
        message(tempstring, returnstring, PROMPT, 64);
        spikefac = atof(returnstring);
  }else spikefac = atof(arg[1]);

  for(k=2; k<n-1; k++)
  {
        a1 = s.x[col][k-1];
        a2 = s.x[col][k+1];
        if((s.x[col][k] / a1 > spikefac) &&
           (s.x[col][k] / a2 > spikefac))
        {
            s.x[col][k] = 0.5 * (s.x[col][k-1] + s.x[col][k+1]);
        }
  }  
  refreshspreadsheet();
}


//--------------------------------------------------------------------------//
//  msmstotic                                                               //
//  remove values >5x lower than neigbbors in thermo ms/ms chromatogram     //
//--------------------------------------------------------------------------//
void msmstotic(char **arg, int noofargs)
{
  static double spikefac = 1.50;
  int col = s.current_col;
  int n = s.maxrow[1];
  int oldn = n;
  int j, k, total;
  double a1,a2;
  spikefac= 0.5;  // values less than this factor are dropped
  if(col+1 !=2)
  {   message((char*)"Retention time must be in col.1\ndata must be in col.2");
      return;
  }
  total = 1;    
  k = 1;
  a2 = 0.0;
  a1 = max(s.x[col][k], s.x[col][k+1]);
  if(s.x[col][k] > s.x[col][k+1]) s.x[col-1][k]=s.x[col-1][k+1];
  while( k < n-1 )
  {     k++;
        a2 = s.x[col][k];
        if(a2 / a1 <= spikefac) continue;
        s.x[col][total] = a2;   //  value
        s.x[col-1][total] = s.x[col-1][k]; // retention time
        total++;
        a1 = a2;
  }  
  for(j=total; j<oldn; j++) 
  {  
      XbaeMatrixSetCell(s.spreadsheet_widget,j,col-1, (char*)"");
      XbaeMatrixSetCell(s.spreadsheet_widget,j,col, (char*)"");
  }
  for(j=0; j<s.maxcol; j++) s.maxrow[j] = total;
  refreshspreadsheet();
}


//--------------------------------------------------------------------//
// bin - average n adjacent data points                               //
// if want_average = 1 it averages otherwise it totals                //
//--------------------------------------------------------------------//
void bin(char **arg, int noofargs, int want_average)
{  
  int col = s.current_col;
  int j,k, oldn, newn;
  static int binfac=2;
  char tempstring[1024];
  double total;
  itoa(s.current_col, tempstring, 10);

  if(noofargs<2) 
  {     message((char*)"Column no.", tempstring, PROMPT, 1023);
        col = atoi(tempstring);
  }else col = atoi(arg[1]);
  if(noofargs<3) 
  {     sprintf(tempstring, "%d", binfac); 
        message((char*)"Bin factor", tempstring, PROMPT, 1023);
        binfac = atoi(tempstring);
  }else binfac = atoi(arg[2]);

  oldn = s.maxrow[col];
  if(binfac<2) return;
  newn = int(oldn / binfac);
  if(newn<=1 || newn>=s.rows){ message((char*)"Invalid bin factor",ERROR); return; }


  for(k=1; k<=newn+1; k++)
  {   total = 0.0;
      for(j=k*binfac-1; j<(k+1)*binfac-1; j++) 
      {    total += s.x[col][j];
      }
      if(want_average) total /= (double)binfac;
      s.x[col][k] = total;
  }

  s.maxrow[col] = newn;

  //// get rid of extra points at the end
  for(k=newn+1; k<=oldn; k++)
  {     s.x[col][k]=0.0;
        s.number[col][k] = 0;
        s.string[col][k] = 0;
        s.hex[col][k] = 0;
        s.text[col][k][0]=0;
        s.disp[col][k][0]=0;
        XbaeMatrixSetCell(s.spreadsheet_widget,k,col,(char*)"");

  }
  refreshspreadsheet();
  return;
}


//--------------------------------------------------------------------//
// reduplicate - repeat n adjacent data points                        //
//--------------------------------------------------------------------//
void reduplicate(char **arg, int noofargs, int want_average)
{  
}


//--------------------------------------------------------------------------//
//  fact  factorial                                                         //
//--------------------------------------------------------------------------//
void fact(char **arg, int noofargs, int mode)
{
  static int n,k;
  double answer=0.0;
  char tempstring[65536];
  char heading[65536];
  if(noofargs<2) 
  {     sprintf(tempstring, "%d", n);
        message((char*)"N (number of items)", tempstring, PROMPT, 1023);
        n = atoi(tempstring);
  }else n = atoi(arg[1]);

  if(noofargs<3 && (mode != FACT)) 
  {     sprintf(tempstring, "%d", k); 
        message((char*)"K (k at a time)", tempstring, PROMPT, 1023);
        k = atoi(tempstring);
  }else k = atoi(arg[2]);
  switch(mode)
  {   case FACT:
           answer = factorial(n, tempstring);
           sprintf(heading, "Factorial of %d\n", n);
      break;
      case PERM:
           //// answer = factorial(n) / factorial(n - k);
           answer = permutations(n, k, tempstring);
           sprintf(heading, "Permutations of n=%d k=%d\n", n, k);
      break;
      case COMB:
           //// answer = factorial(n) / ( factorial(n - k) * factorial(k));
           answer = binom(n, k, tempstring);
           sprintf(heading, "Combinations of n=%d k=%d\n", n, k);
      break;
  }
  answer = answer; //// shut up compiler
  strcat(heading, tempstring);
  message((char*)heading, tempstring, INFO, 1023);  
}





//--------------------------------------------------------------------//
// gmp_series_term                                                    //
// calculate a^b / c! using libgmp                                    //
// call using series(a,b,c) in math formula                           //
// becomes (aa^bb)/cc! where bb and cc are ulong int                  //
//--------------------------------------------------------------------//
double gmp_series_term(double a, int b, int c)
{   
   mpz_t fac;         // declare int
   mpf_t fac2;        // declare float 
   mpf_t pow;         // declare float
   mpf_t aa;          // declare float, needed for mpf_pow_ui
   mpf_t quo;         // declare float
   unsigned long int bb;
   char tempstring[2048];
   double x;
   
   bb = (unsigned long int) b;
   mpz_init(fac);      // int  
   mpf_init(aa);       // float
   mpf_init(pow);      // float
   mpf_init(quo);      // float
   mpf_set_d (aa, a);  // convert a to a multiprecision float  

   mpz_fac_ui(fac, c);         // fac = c factorial  

   //// convert it to a string, then convert the string to a mpf_t
   mpz_get_str(tempstring, 10, fac);
   mpf_init_set_str(fac2, tempstring, 10);
   //// This method crashes when you free fac2
   //mpf_set_z(fac2, fac);       // convert int to float

   //// aa must be mpf_t, bb must be ulong int 
   mpf_pow_ui (pow, aa, bb);   // pow = aa^bb  

   //// quo = pow/fac, 
   //// quo must be mpf_t
   //// pow must be const mpf_t
   //// fac must be const mpf_t
   mpf_div(quo, pow, fac2);       

gmp_printf("\nfac   %Zd\n",fac);  // print big integer
gmp_printf("fac2 %Ff\n",fac2);    // print big float
gmp_printf("pow  %Ff\n",pow);     // print big float
gmp_printf("quo  %Ff\n",quo);     // print big float

   x = mpf_get_d(quo);

////printf("result %g\n",x);
   mpz_clear (fac);
   mpf_clear (fac2);
   mpf_clear (pow);
   mpf_clear (aa);
   mpf_clear (quo);
   return x;
}

//--------------------------------------------------------------------//
// binom  binomial coeff (n k) = combinations                         //
// must appear after #include<gmp.h>                                  //
// factorial(n)/(factorial(k)*factorial(n - k));                      //
// tempstring returns answer in string format, must be allocated      //
//--------------------------------------------------------------------//
double binom(double n, double k)
{
   char tempstring[65536];
   return binom((int)n, (int)k, tempstring);
}
double binom(int n, int k, char* tempstring)
{
//   double answer = 1;
//   if(between(n,0,17)) answer = factorial(n)/(factorial(k)*factorial(n - k));
//   return floor(answer+.001);

   double answer = 1.0;
   if(n>17235) { strcpy(tempstring, "Too frickin big"); return 0;}
   if(k>17235) { strcpy(tempstring, "Too frickin big"); return 0;}

   mpz_t facn;          // declare int
   mpz_t fack;          // declare int
   mpz_t facnk;         // declare int
   mpz_t facktimesnk;   // declare int
   mpz_t quo;           // quotient

   mpz_init(facn);       // must initialize it
   mpz_init(fack);       // must initialize it
   mpz_init(facnk);      // must initialize it
   mpz_init(facktimesnk);// must initialize it
   mpz_init(quo);        // must initialize it
 
   mpz_fac_ui(facn, n);         // facn = n factorial  
   mpz_fac_ui(fack, k);         // fack = n factorial  
   mpz_fac_ui(facnk, (n-k));     // facnk = (n-k) factorial  
   mpz_mul(facktimesnk, fack, facnk);  // integer mult of 2 mpz_t's
   mpz_fdiv_q(quo, facn, facktimesnk); // integer div, floor, return quotient
   answer = mpz_get_d(quo);            // convert to a double

   mpz_clear (facktimesnk);
   mpz_get_str(tempstring, 10, quo);
   mpz_clear (quo);
   mpz_clear (facnk);
   mpz_clear (fack);
   mpz_clear (facn);

   convert_mpz_to_scinot_string(quo, tempstring);

   //// will crash if answer is infinity, so set to 1e300
   if(answer>1e300) answer=1e300; 
   return answer;
}

//--------------------------------------------------------------------//
// permutations                                                       //
// must appear after #include<gmp.h>                                  //
// factorial(n) / factorial(n - k);                                   //
// tempstring returns answer in string format, must be allocated      //
//--------------------------------------------------------------------//
double permutations(double n, double k)
{
   return permutations((int)n, (int)k);
}
double permutations(int n, int k, char* tempstring)
{
//   double answer = 1;
//   if(between(n,0,17)) answer = factorial(n)/(factorial(k)*factorial(n - k));
//   return floor(answer+.001);

   double answer = 1.0;
   if(n>17235) { strcpy(tempstring, "Too frickin big"); return 0;}
   if(k>17235) { strcpy(tempstring, "Too frickin big"); return 0;}
   mpz_t facn;          // declare int
   mpz_t facnk;         // declare int
   mpz_t quo;           // quotient

   mpz_init(facn);       // must initialize it
   mpz_init(facnk);      // must initialize it
   mpz_init(quo);        // must initialize it
 
   mpz_fac_ui(facn, n);         // facn = n factorial  
   mpz_fac_ui(facnk, (n-k));     // facnk = (n-k) factorial  
   mpz_fdiv_q(quo, facn, facnk);       // integer div, floor, return quotient
   answer = mpz_get_d(quo);            // convert to a double
   mpz_get_str(tempstring, 10, quo);

   mpz_clear (quo);
   mpz_clear (facnk);
   mpz_clear (facn);

   convert_mpz_to_scinot_string(quo, tempstring);

   return answer;
}


//--------------------------------------------------------------------//
// factorial                                                          //
// tempstring returns answer in string format, must be allocated      //
//--------------------------------------------------------------------//
double factorial(double n)
{   return factorial((int)n);
}
double factorial(int n)
{   char tempstring[65536];
    return factorial(n, tempstring);
}
double factorial(int n, char* tempstring)
{   
//    int k;
//    double x=1;
//    for(k=1;k<=n;k++) x*=k;
//    return x;
   double answer = 1.0;
   if(n>17235) { strcpy(tempstring, "Too frickin big"); return 0;}

   mpz_t facn;                  // declare int
   mpf_t aa;                    // declare float, so can print result
   mpz_init(facn);              // must initialize it
   mpf_init(aa);                // must initialize it

   mpz_fac_ui(facn, n);         // facn = n factorial  

   //// convert it to a string, then convert the string to a mpf_t big float
   mpz_get_str(tempstring, 10, facn);
   mpf_set_str(aa, tempstring, 10);

   answer = mpz_get_d(facn);    // convert to a double

// print out the result in case it overflows double
//gmp_printf("\nfac %d = %Zd\n",n,facn);  // print big integer
//gmp_printf(" = %Ff\n",aa);            // print big float

   //// Can't get this into a long int, so print it out on screen.
   gmp_printf("fac %d = %Fg\n",n, aa);     // print big float in sci.not.

   //// print big float in sci.not.
   gmp_sprintf(tempstring, "%Fg\n", aa);    

   convert_mpf_to_scinot_string(aa, tempstring);

   mpf_clear (aa);
   mpz_clear (facn);
   return answer;
}

//--------------------------------------------------------------------//
// convert_mpf_to_scinot_string                                       //
//--------------------------------------------------------------------//
void convert_mpf_to_scinot_string(mpf_t aa, char *tempstring) 
{
   //// put float into a string, base 10, 15 decimal places 
   //// all this just to get the result in a reasonable size
   //// tempstring can contain hundreds of digits
   //// making it useless

   char tempstring2[65536];
   char tempstring3[10];
   mp_exp_t expptr;  // this is just a long int
   
   mpf_get_str(tempstring, &expptr, 10, 15, aa);
   tempstring3[0] = tempstring[0];
   tempstring3[1] = 0;
   sprintf(tempstring2,"%s.%se+%d",tempstring3,tempstring+1,(int)expptr-1);
   strcpy(tempstring, tempstring2);
}

//--------------------------------------------------------------------//
// convert_mpz_to_scinot_string                                       //
//--------------------------------------------------------------------//
void convert_mpz_to_scinot_string(mpz_t aa, char *tempstring) 
{
   //// put int into a string, base 10, 15 decimal places 
   //// all this just to get the result in a reasonable size
   //// tempstring can contain hundreds of digits
   //// making it useless

   char tempstring2[65536];
   char tempstring3[10];
   int expptr=0;

   //// convert it to a string, then count the length
   mpz_get_str(tempstring, 10, aa);
   expptr = strlen(tempstring);
   tempstring3[0] = tempstring[0];
   tempstring3[1] = 0;
   tempstring[16] = 0;
   sprintf(tempstring2,"%s.%se+%d",tempstring3,tempstring+1, expptr-1);
   strcpy(tempstring, tempstring2);
}


//--------------------------------------------------------------------//
// sum - for calculator, example: sum(x[r1][c1]:x[r2][c2])            //
//       or              example: sum(r1 c1 : r2 c2)                  //
//--------------------------------------------------------------------//
double sum(double a, double b, double c, double d) 
{
   int row, col;
   double answer = 0;
   for(col=(int)(b); col<=(int)(d); col++)
   for(row=(int)(a); row<=(int)(c); row++)
        answer += s.x[col][row];
   return answer;
}

//--------------------------------------------------------------------//
// sumsq - for calculator example: sumsq(r1 c1 : r2 c2)               //
//--------------------------------------------------------------------//
double sumsq(double a, double b, double c, double d) 
{
   int row, col;
   double answer = 0;
   for(col=(int)(b); col<=(int)(d); col++)
   for(row=(int)(a); row<=(int)(c); row++)
        answer += s.x[col][row] * s.x[col][row];
   return answer;
}

//--------------------------------------------------------------------//
// reverse                                                            //
// reverse order of rows in a column                                  //
//--------------------------------------------------------------------//
void reverse(char **arg, int noofargs)
{  
  int i;
  int rowmax, col;
  char tempstring[1024]="";
  col = s.current_col;
  rowmax = s.maxrow[col];

  for(i=1; i<=rowmax/2; i++)
  {
          dswap(s.x[col][i], s.x[col][1+rowmax-i]);
          dswap(s.number[col][i], s.number[col][1+rowmax-i]);
          dswap(s.string[col][i], s.string[col][1+rowmax-i]);
          dswap(s.hex[col][i], s.hex[col][1+rowmax-i]);

          strcpy(tempstring,s.text[col][1+rowmax-i]);
          strcpy(s.text[col][1+rowmax-i], s.text[col][i]);
          strcpy(s.text[col][i], tempstring);

          strcpy(tempstring,s.disp[col][1+rowmax-i]);
          strcpy(s.disp[col][rowmax-i], s.disp[col][i]);
          strcpy(s.disp[col][i], tempstring);

          strcpy(tempstring,s.form[col][1+rowmax-i]);
          strcpy(s.form[col][1+rowmax-i], s.form[col][i]);
          strcpy(s.form[col][i], tempstring);
  }
  refreshspreadsheet();  
  return;
}

//--------------------------------------------------------------------------//
//  binomialprobability                                                     //
//  arg 1=n arg2=k arg3=p                                                   //
// test data  n=20 k=4 p=0.5                                                //
// exactly 4 out of 20 0.00462055                                           //
// 4 or fewer of 20 0.00590897                                              //
// 4 or more 0.998712                                                       //
//--------------------------------------------------------------------------//
void binomialprobability(char **arg, int noofargs)
{
  static double pp=1.0, prob=0.0;
  static int kk=1.0, nn=1.0;
  double pfewer=0.0;
  double pmore=0.0;
  int i, errflag = 0;
  char tempstring[1026];
  char tempstring2[1026];
  arg=arg; noofargs=noofargs;

  sprintf(tempstring,"%d", nn);
  if(noofargs<2) 
  {     message((char*)"N Number of opportunities", tempstring, PROMPT, 1023);
        nn = atoi(tempstring);
  }else nn = atoi(arg[1]);

  sprintf(tempstring,"%d", kk);
  if(noofargs<3) 
  {     message((char*)"k Number of times stipulated to occur", tempstring, PROMPT, 1023);
        kk = atoi(tempstring);
  }else kk = atoi(arg[2]);

  sprintf(tempstring,"%g", pp);
  if(noofargs<4) 
  {     message((char*)"P Probability of an event", tempstring, PROMPT, 1023);
        pp = atof(tempstring);
  }else pp = atof(arg[3]);
  
  prob = binprob_gmp(nn, kk, pp, &errflag);
  for(i=0; i<=kk; i++) pfewer += binprob_gmp(nn, i, pp, &errflag);
  pmore = 1.0 - pfewer + prob;

  sprintf(tempstring, "Binomial Probability for N=%d k=%d p=%g q=%g\n",nn, kk, pp, 1.0-pp);
  if(errflag) 
  {
      strcat(tempstring, "\nOverflow occurred, number too big");
  }else
  {
      sprintf(tempstring2,"\nProbability of exactly %d out of %d is %g\n\
Probability of %d or fewer out of %d is %g\n\
Probability of %d or more  out of %d is %g\n",
kk, nn, prob, kk, nn, pfewer, kk, nn, pmore);
      strcat(tempstring, tempstring2);
  }
  message((char*)tempstring, INFO);
}

double binprob(int n, int k, double p, int *errflag)
{
  double answer;
  double q = 1.0 - p;
  answer =  pow(p,k) * pow(q, n-k) * binom(n,k);
  return answer;
}

double binprob_gmp(ulong n, ulong k, double p, int *errflag)
{
  double answer = 0.0;
  double q = 1.0 - p;
  *errflag = 0;                 // gmp puts inf into a double, then crashes
 
  //// floats
  mpf_t pp;                     // declare float
  mpf_t qq;                     // declare float
  mpf_t pk;                     // declare float
  mpf_t qnk;                    // declare float
  mpf_t ans;                    // declare float
  mpf_t anstemp;                // declare float

  mpf_init (pk);                // must initialize 
  mpf_init (qnk);               // must initialize 
  mpf_init (pp);                // must initialize 
  mpf_init (qq);                // must initialize 
  mpf_init (ans);               // must initialize 
  mpf_init (anstemp);           // must initialize 

  mpf_set_d (pp, p);            // convert a to a multiprecision float  
  mpf_set_d (qq, q);            // convert a to a multiprecision float  

  mpf_pow_ui(pk, pp, k);        // pk = pp^k
  mpf_pow_ui(qnk, qq, n-k);     // pk = qq^k

  //// ans = p^k * q^k * binom(n,k)
  mpf_mul (ans, pk, qnk);       // ans = pp x qq
  mpf_set_d(anstemp, binom(n,k)); // anstemp = binom(n,k) 
  answer = mpf_get_d (anstemp);   // put the result back into a double
  if(answer >=1e300) *errflag=1;  // overflow
  mpf_mul (ans, ans, anstemp);  // ans = ans x anstemp

// answer =  (pow(p,(double)k) * pow(q, (double)n-(double)k) * (double)factorial(n)) / 
//         ((double)factorial(k) * (double)factorial(n-k));
// binom(n,k) =  factorial(n)/(factorial(k)*factorial(n - k));                     
// answer (pow(p,k) * pow(q, n-k) * binom(n,k)

  answer = mpf_get_d (ans);     // put the result back into a double
  //// must delete all the mp variables

  mpf_clear (anstemp);          // must delete the variable
  mpf_clear (ans);              // must delete the variable
  mpf_clear (qnk);              // must delete the variable
  mpf_clear (pk);               // must delete the variable
  mpf_clear (qq);               // must delete the variable
  mpf_clear (pp);               // must delete the variable

  return answer;
}



//--------------------------------------------------------------------//
// multiplicative inverse                                             //
// returns modulo inverse of aaa w.r.t. m using extended Euclid algo. //
//--------------------------------------------------------------------//
double modinverse(double a, double m)
{
    int aaa = (int)a; 
    int mmm = (int)m; 
    int m0 = mmm;
    int y = 0, x = 1;

    if (mmm == 1) return 0;
    if (mmm == 0) return 0;
 
    while (aaa > 1) 
    {
        if (mmm == 0) return 0;
        int q = aaa / mmm;
        int t = mmm;
 
        mmm = aaa % mmm;
        aaa = t;

        t = y;
        y = x - q * y;
        x = t;
    }
    if (x < 0) x += m0;
    return (double) x;
}


//--------------------------------------------------------------------//
// mod power                                                          //
// returns aa to power of bb modulo mm using fast powering algo.      //
// see Hoffstein et al, p.25 a, pwr, and m are treated as integers    //
// only calculates up to 2^32                                         //
// usage in editor                                                    //
//       modpwr 1 218 1000 = all of col.1 raised to the 218, mod 1000 //
// usage as called in math()                                          //
//       modpwr 3 218 1000 = 3 raised to the 218, mod 1000            //
//--------------------------------------------------------------------//
double modpwr(double a, double power, double m)
{
    unsigned long long int answer = 0;
    int kk;
    unsigned long long int aaa = (unsigned long long int)a; 
    unsigned long long int pwr = (unsigned long long int)power; 
    unsigned long long int mmm = (unsigned long long int)m; 
    unsigned long long int ppp[33];
    unsigned long long int hit[33];
    unsigned long long int remainder = pwr;

    //// step 1 binary expansion in powers of 2
    ppp[0] = 1;
    for(kk = 0; kk<=32; kk++) hit[kk] = 0;
    for(kk = 1; kk<=32; kk++) ppp[kk] = 2*ppp[kk-1];
    for(kk=32; kk>=0; kk--)
    {   if(remainder >= ppp[kk])
        {    hit[kk]=1;
             remainder -= ppp[kk];
        }
    }
    //// step 2 
    answer = 1;
    ppp[0]=aaa;

    //// calculate 3^2^i mod 1000
    for(kk = 1; kk<32; kk++) ppp[kk] = (ppp[kk-1] * ppp[kk-1]) % mmm;
    for(kk = 0; kk<=32; kk++) 
    {   if(hit[kk]) 
        {  answer = (answer * ppp[kk]) % mmm;
        }
    }
    return (double) answer;
}
