//--------------------------------------------------------------------------//
// xdata9.cc                                                                //
// Latest revision: 05-23-2022                                              //
// Wavelets                                                                 //
//--------------------------------------------------------------------------//

#include "xdata.h"

extern Globals g;
extern Spreadsheet s;

//--------------------------------------------------------------------------//
// wavelet                                                                  //
//--------------------------------------------------------------------------//
void wavelet(char **arg, int noofargs)
{
   static int  sourcecol=0, resultcol=1, direction=1, transform_type=1, reconcol=2;
   static char wavefilename[1024] = "none";
   static char index_range[80] = "all";
   static char coeff_range[80] = "all";
   static char ignore_range[80] = "none";
   int j, k;
   arg=arg; noofargs=noofargs; 
   sourcecol = s.current_col;
   resultcol = s.current_col+1;
   reconcol = s.current_col+2;

   ////------------------------////

   Dialog dialog;
   dialog.needrepeat = 0;  
   strcpy(dialog.radio[0][0],"Execute"); 
   strcpy(dialog.radio[0][1],"Forward");             
   strcpy(dialog.radio[0][2],"Reverse");
   dialog.radiono[0]=3;
   if(direction==1) dialog.radioset[0] = 1; 
   if(direction==0) dialog.radioset[0] = 2; 

   strcpy(dialog.radio[1][0],"Algorithm"); 
   strcpy(dialog.radio[1][1],"Pyramidal");             
   strcpy(dialog.radio[1][2],"Mallat");
   strcpy(dialog.radio[1][3],"Multiresolution");
   dialog.radiono[1]=4;
   dialog.radioset[1] = transform_type;

   strcpy(dialog.title,"Wavelet transform");
   strcpy(dialog.boxes[0],"Data Column");
   strcpy(dialog.boxes[1],"Source  ");
   strcpy(dialog.boxes[2],"Coefficients");
   strcpy(dialog.boxes[3],"Reconstruction");
   dialog.boxtype[0]=LABEL;

   dialog.boxtype[1]=INTCLICKBOX;
   sprintf(dialog.answer[1],"%d",sourcecol);
   dialog.boxmin[1]=1; 
   dialog.boxmax[1]=s.cols;

   dialog.boxtype[2]=INTCLICKBOX;
   sprintf(dialog.answer[2],"%d",resultcol);
   dialog.boxmin[2]=1; 
   dialog.boxmax[2]=s.cols;

   dialog.boxtype[3]=INTCLICKBOX;
   sprintf(dialog.answer[3],"%d",reconcol);
   dialog.boxmin[3]=1; 
   dialog.boxmax[3]=s.cols;

   strcpy(dialog.boxes[4],"Wavelet");
   strcpy(dialog.answer[4], wavefilename);
   dialog.boxtype[4]=FILENAME;

   strcpy(dialog.boxes[5],"Reconstitution coefficients");
   dialog.boxtype[5]=LABEL;

   strcpy(dialog.boxes[6],"Index range to use");
   dialog.boxtype[6]=STRING;
   strcpy(dialog.answer[6], index_range);

   strcpy(dialog.boxes[7],"Value range to use");
   dialog.boxtype[7]=STRING;
   strcpy(dialog.answer[7], coeff_range);

   strcpy(dialog.boxes[8],"Value range to ignore");
   dialog.boxtype[8]=STRING;
   strcpy(dialog.answer[8], ignore_range);


   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=2;
   dialog.noofboxes=9;
   dialog.helptopic=0;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = wavelet_check;
   dialog.transient = 1;
   strcpy(dialog.path, g.waveletpath);
  
   dialogbox(dialog);

   sourcecol = atoi(dialog.answer[1]);
   resultcol = atoi(dialog.answer[2]);
   reconcol  = atoi(dialog.answer[3]);
   strcpy(wavefilename, dialog.answer[4]);
   strcpy(index_range, dialog.answer[6]);
   strcpy(coeff_range, dialog.answer[7]);
   strcpy(ignore_range, dialog.answer[8]);
}


//--------------------------------------------------------------------------//
// read_wavelet                                                             //
// File Format:                                                             //
//   Lines starting with # or blank = comments                              //
//    [Range low] [Range high]                                              //
//    [Decomposition LP filter coefficients]                                //
//    . . .                                                                 // 
//    [Range low] [Range high]                                              //
//    . . .                                                                 // 
//    [Reconstitution LP filter coefficients]                               //
//    . . . (If absent, use decomposition coefficients)                     // 
//    Any characters after column 80 may be ignored.                        //
//--------------------------------------------------------------------------//
void read_wavelet(Wavelet &w)
{
  FILE *fp;
  int  k, count=0, low, high, sign = -1;
  char buffer[1024];
  if((fp=fopen(w.filename,"rt")) == NULL)
      error_message(w.filename, errno);
  else
  {   
      if(read_next_non_comment(fp, buffer)){ fclose(fp); return; }
      sscanf(buffer, "%d %d", &low, &high);
      w.ioffset = low;
      w.size = high - low + 1; 
      count = 0;
      ////  Decomposition filter
      for(k=low; k<=high; k++)
      {   if(read_next_non_comment(fp, buffer)){ fclose(fp); return; }
          if(count<1024) w.filter[0][count] = atof(buffer);
          w.filter[1][count] = w.filter[0][count];
          count++;
      }
      if(read_next_non_comment(fp, buffer)){ fclose(fp); return; }
      sscanf(buffer, "%d %d", &low, &high);
      w.joffset = low;
      count = 0;
      ////  Reconstitution filter
      for(k=low; k<=high; k++)
      {   if(read_next_non_comment(fp, buffer)){ fclose(fp); return; }
          if(count<1024) w.filter[1][count] = atof(buffer);
          count++;
      }
      ////  High-pass filters - reversed with alternating sign changes
      for(k=0; k<w.size; k++)
      {   w.rfilter[0][w.size-k-1] = sign * w.filter[0][k];
          w.rfilter[1][w.size-k-1] = sign * w.filter[1][k];
          sign = -sign;
      }
      if(w.ioffset == 0 && w.joffset == 0) w.ioffset = w.joffset = -w.size/2; 
  }
  fclose(fp); 
  return;
}

//--------------------------------------------------------------------------//
// read_next_non_comment                                                    //
//--------------------------------------------------------------------------//
int read_next_non_comment(FILE *fp, char *buffer)
{
   int eof = 0;
   do
   {   fgets(buffer, 1024, fp);
       if(feof(fp)){ eof=1; break; }
       remove_trailing_characters(buffer);
   } while( (buffer[0] == '#') || (!strlen(buffer)) );
   return eof;
}


//--------------------------------------------------------------------------//
//  remove_trailing_space - spaces only                                     //
//--------------------------------------------------------------------------//
void remove_trailing_space(char *s)
{
   int k;
   for(k=strlen(s)-1; k>=0; k--){ if(s[k]==' ') s[k] = 0; else break; }
}

//--------------------------------------------------------------------------//
//  remove_all_cr - cr only - some files have 2 cr's, get rid of all        //
//--------------------------------------------------------------------------//
void remove_all_cr(char *s)
{
   int k;
   for(k=strlen(s)-1; k>=0; k--){ if(s[k]==13) s[k] = ' '; }
}


//--------------------------------------------------------------------------//
//  remove_trailing_characters - including \f \n \r \t \v and space           //
//--------------------------------------------------------------------------//
void remove_trailing_characters(char *s)
{
   int k;
   for(k=strlen(s)-1; k>=0; k--){ if(isspace(s[k]) || s[k]==10 || s[k]==13) s[k] = 0; else break; }
}


//--------------------------------------------------------------------------//
// wavelet_pyramidal_calc                                                   //
// n-dimensional vavelet transform                                          //
//--------------------------------------------------------------------------//
void wavelet_pyramidal_calc(Wavelet &wave, int direction, int ndim, int *nn, 
     int col1, int col2, int imin, int imax, double cmin, double cmax, 
     double ignmin, double ignmax)
{
   int dim, k, i1,i2,i3,n,nnew,nprev=1,nt,ntot=1;
   double *wksp;
   double coef;
printf("column %d -> %d  direction %d\n",col1,col2,direction);
printf("Using wavelet of size %d ioffset %d joffset %d\n", wave.size, wave.ioffset, wave.joffset);
for(k=0;k<wave.size;k++) printf("%g\t%g\t%g\t%g\n",
wave.filter[0][k], wave.filter[1][k], wave.rfilter[0][k], wave.rfilter[1][k]);

printf("using index  %d to %d\nusing values %g to %g\nignoring     %g to %g\n",imin,imax,cmin,cmax,ignmin,ignmax);
   for(k=0; k<ndim; k++) ntot *= nn[k];
   wksp = new double[ntot];
   for(dim=0; dim<ndim; dim++)
   {  
      n = nn[dim];
      nnew = n*nprev;
      for(i2=0;i2<ntot;i2+=nnew)
      for(i1=0;i1<nprev;i1++)
      {   
          for(i3=i1+i2, k=0; k<n; k++,i3+=nprev) 
          {   coef = s.x[col1][i3];
              if(direction==1 || (between(i3, imin, imax) && between(coef, cmin, cmax) &&
                  !between(coef, ignmin, ignmax)))
                     wksp[k] = coef;
              else wksp[k] = 0;
          }
          if(direction==1)   // 1=forward
              for(nt=n; nt>=4; nt/=2) wavelet_transform(wave, wksp, nt, direction);
          else               // 0=reverse
              for(nt=4; nt<=n; nt*=2)  wavelet_transform(wave, wksp, nt, direction);
          for(i3=i1+i2, k=0; k<n; k++,i3+=nprev) s.x[col2][i3] = wksp[k];
      }
      nprev = nnew;
   }
   delete[] wksp;
}  


//--------------------------------------------------------------------------//
// wavelet_transform - called by wavelet_pyramidal_calc                     //
//--------------------------------------------------------------------------//
void wavelet_transform(Wavelet &wave, double *aa, int n, int direction)
{
   double ai, ai1, *wksp;
   int i,ii,jf,jr,k,n1,ni,nj,nh,nmod;
   wksp = (double*)malloc(n*sizeof(double));
   if(wksp==NULL) printf("Oh nooooooooo it was null!!!!!!!!\nAiieeeee!!!!!");
   for(k=0;k<n;k++) wksp[k]=0.0;
   nmod = wave.size * n;
   n1 = n-1;
   nh = n/2;
   if(direction == 1)   // forward
   {   for(ii=0, i=0; i<n; i+=2, ii++)
       {   ni = i + nmod + wave.ioffset;
           nj = i + nmod + wave.joffset;
           for(k=0; k<wave.size; k++)
           {   jf = n1 & (1 + ni + k);
               jr = n1 & (1 + nj + k);
               wksp[ii] += wave.filter[0][k] * aa[jf];
               wksp[ii+nh] += wave.rfilter[0][k] * aa[jr];
           }
       }
   }else
   {  
       for(ii=0, i=0; i<n; i+=2, ii++)
       {   ai = aa[ii];
           ai1 = aa[ii+nh];
           ni = i + nmod + wave.ioffset;
           nj = i + nmod + wave.joffset;
           for(k=0; k<wave.size; k++)
           {   jf = n1 & (1 + ni + k);
               jr = n1 & (1 + nj + k);
               wksp[jf] += wave.filter[1][k] * ai;
               wksp[jr] += wave.rfilter[1][k] * ai1;
           }
      }
   }
   for(k=0;k<n;k++) aa[k] = wksp[k];
   free(wksp);
   wksp = NULL;
}


//--------------------------------------------------------------------------//
//  wavelet_check                                                           //
//--------------------------------------------------------------------------//
void wavelet_check(dialoginfo *a, int radio, int radiobutton, int box, int boxbutton)
{
   static int incheck=0, hit=0;
   static int sourcecol=0, resultcol=1, reconcol=2, direction=-1, transform_type=1;
   if(incheck) return;            // XtSetValues->dialogstringcb->fftcheck
   static Wavelet wave;    
   static int imin, imax;         // index min to max
   static double cmin, cmax;      // coeff min to max
   static double ignmin, ignmax;  // coeff min to max to ignore
   int  k, ndim=1, size[10], noofargs, ss, len;
   char **buf;
   char tempstring[1024];

   sourcecol = atoi(a->answer[1]);
   resultcol = atoi(a->answer[2]);
   reconcol  = atoi(a->answer[3]);
                                  // Make it a power of 2 in length
   len = s.maxrow[sourcecol];
   for(ss=1;ss<=len*2;ss*=2) if(ss>=len) break; 
   if(s.maxrow[sourcecol] != ss) 
   {   if(ss > s.rows)
       {  sprintf(tempstring,"Spreadsheet is too small\nEdit xdata.ini to increase rows to %d\nand restart xdata",ss);
          message(tempstring, ERROR);
          return;
       }
       else
       {
          message((char*)"Enlarging data to nearest\npower of 2");
       }
   }
   if(g.getout){ g.getout=0; return; }
   for(k=s.maxrow[sourcecol];k<ss;k++)
   {    s.number[sourcecol][k]=1; 
        s.string[sourcecol][k]=0;
        s.x[sourcecol][k]=0.0; 
        s.hex[sourcecol][k]=0;
   }
   s.maxrow[sourcecol] = s.maxrow[resultcol] = s.maxrow[reconcol] = ss;


   buf = new char*[4];
   for(k=0;k<4;k++){ buf[k] = new char[32]; buf[k][0] = 0; }

   if(!hit)
   {   wave.filename   = new char[FILENAMELENGTH];
       wave.filter     = new double*[3];
       wave.filter[0]  = new double[1024];  // decomposition filter
       wave.filter[1]  = new double[1024];  // reconstruction filter
       wave.rfilter    = new double*[3];
       wave.rfilter[0] = new double[1024];  // reverse decomposition filter
       wave.rfilter[1] = new double[1024];  // reverse reconstruction filter
       strcpy(wave.filename, "none");
   }

   incheck=1;
   box=box; boxbutton=boxbutton;

   if(a->radioset[0] == 1) direction=1; else direction = 0;
   transform_type = a->radioset[1];
   strcpy(wave.filename, a->answer[4]);
   size[0] = s.maxrow[sourcecol]; // length of 1st dimension (must be power of 2)
   ndim = 1;                      // no. of dimensions

                                  // index range to use
   if(!hit || (radio==-1 && box==6))
   {   strip_extraneous_dashes(a->answer[6]);
       parse_string(a->answer[6], buf, NULL, noofargs, 0, 32, 99, ' ',0);
       if(!strcmp(buf[0],"all"))
       {   imin=0; imax=size[0];}
       else if(!strcmp(buf[0],"none"))
       {   imin=0; imax=0; }
       else if(noofargs>=2)
       {   imin = atoi(buf[0]);
           if(noofargs==2) imax = atoi(buf[1]);
           else imax = atoi(buf[2]);
       }else{ imin=0; imax=size[0];}
   }
                                  // values to use
   if(!hit || (radio==-1 && box==7))
   {   strip_extraneous_dashes(a->answer[7]);
       parse_string(a->answer[7], buf, NULL, noofargs, 0, 32, 99, ' ',0);
       if(!strcmp(buf[0],"all"))
       {   cmin = smallest(resultcol); cmax = largest(resultcol); }
       else if(!strcmp(buf[0],"none"))
       {   cmin=0.0; cmax=0.0; }
       else if(noofargs>=2)
       {   cmin = atof(buf[0]);
           if(noofargs==2) cmax = atof(buf[1]);
           else cmax = atof(buf[2]);
       }else{ cmin = smallest(resultcol); cmax = largest(resultcol); }
   }
                                  // values to ignore
   if(!hit || (radio==-1 && box==8))
   {   strip_extraneous_dashes(a->answer[8]);
       parse_string(a->answer[8], buf, NULL, noofargs, 0, 32, 99, ' ',0);
       if(!strcmp(buf[0],"all"))
       {   ignmin = smallest(resultcol); ignmax = largest(resultcol); }
       else if(!strcmp(buf[0],"none"))
       {   ignmin = 0.0; ignmax = 0.0; }
       else if(noofargs>=2)
       {   ignmin = atof(buf[0]);
           if(noofargs==2) ignmax = atof(buf[1]);
           else ignmax = atof(buf[2]);
       }else{ ignmin = 0.0; ignmax = 0.0; }
   }

   
   //// Radio buttons
   if(radio==0 && box==-1 && boxbutton==-1)
   {   if(!strcmp(wave.filename,"none")) 
          message((char*)"Select a, like, wavelet first");
       else
       {  read_wavelet(wave);
          switch(transform_type)
          {  case 1: 
                 if(direction==1)
                   wavelet_pyramidal_calc(wave, 1, ndim, size, sourcecol, resultcol,
                      imin, imax, cmin, cmax, ignmin, ignmax);
                 else
                   wavelet_pyramidal_calc(wave, 0, ndim, size, resultcol, reconcol,
                      imin, imax, cmin, cmax, ignmin, ignmax);
                 break;
             default: message((char*)"Nope! no good!\n",ERROR); break;
          }     
          if(direction==1)
          {     for(k=0; k<size[0]; k++){ s.number[resultcol][k] = 1; s.string[resultcol][k] = 0; s.hex[resultcol][k] = 0;} 
                s.maxrow[resultcol] = size[0];
                s.maxcol=max(s.maxcol, 1 + resultcol);
          }else
          {     for(k=0; k<size[0]; k++){ s.number[reconcol][k] = 1; s.string[reconcol][k] = 0; s.hex[reconcol][k] = 0;}
                s.maxrow[reconcol] = size[0];
                s.maxcol=max(s.maxcol, 1 + reconcol);
          }
          refreshspreadsheet();
       }
   }
   incheck=0;
   for(k=0;k<4;k++) delete[] buf[k];
   delete[] buf;
   hit = 1;
   return;
}


//--------------------------------------------------------------------------//
//  strip_extraneous_dashes                                                 //
//  get rid of any '-' not preceded by a whitespace.                        //
//--------------------------------------------------------------------------//
void strip_extraneous_dashes(char *s)
{
   int k, number=0;
   for(k=0;k<(int)strlen(s);k++)
   {   if(number && s[k]=='-') s[k]=' ';
       if(isdigit(s[k])) number=1;
       if(isspace(s[k])) number=0;
   }   
}


//--------------------------------------------------------------------------//
//  smallest                                                                //
//--------------------------------------------------------------------------//
double smallest(int col)
{
   int k;
   double d = 0;
   for(k=0;k<s.maxrow[col];k++)
      if(s.number[col][k]){ d = s.x[col][k]; break; }
   for(k=0;k<s.maxrow[col];k++)
      if(s.number[col][k] && s.x[col][k] < d) d = s.x[col][k]; 
   return d;
}

//--------------------------------------------------------------------------//
//  largest                                                                 //
//--------------------------------------------------------------------------//
double largest(int col)
{
   int k;
   double d = 0;
   for(k=0;k<s.maxrow[col];k++)
      if(s.number[col][k]){ d = s.x[col][k]; break; }
   for(k=0;k<s.maxrow[col];k++)
      if(s.number[col][k] && s.x[col][k] > d) d = s.x[col][k]; 
   return d;
}
